/* Ofront 1.4 -xtspkaG */
#include "SYSTEM.h"
#include "DialogClocks.h"
#include "DialogFrames.h"
#include "Dialogs.h"
#include "Display.h"
#include "In.h"
#include "Math.h"
#include "Oberon.h"
#include "Printer.h"

typedef
	struct DialogAnalogClocks_ItemDesc *DialogAnalogClocks_Item;

typedef
	struct DialogAnalogClocks_ItemDesc { /* DialogClocks_ItemDesc */
		LONGINT _prvt0;
		char _prvt1[16];
		CHAR name[16];
		CHAR cmd[32];
		CHAR par[32];
		BOOLEAN selected, overlapping;
		Dialogs_Panel panel;
		char _prvt2[1];
	} DialogAnalogClocks_ItemDesc;

export void DialogAnalogClocks_Item_Copy (DialogAnalogClocks_Item c, Dialogs_Object *dup);
export void DialogAnalogClocks_Item_Draw (DialogAnalogClocks_Item c, INTEGER x, INTEGER y, Display_Frame f);
export void DialogAnalogClocks_Item_Print (DialogAnalogClocks_Item c, INTEGER x, INTEGER y);
export void DialogAnalogClocks_Item_Redraw (DialogAnalogClocks_Item c, Display_Frame f, INTEGER x, INTEGER y, DialogClocks_Time old, DialogClocks_Time new);


static REAL DialogAnalogClocks_sin[60], DialogAnalogClocks_cos[60];
static void (*DialogAnalogClocks_Line)(Display_Frame, INTEGER, INTEGER, INTEGER, INTEGER, INTEGER, INTEGER);
static void (*DialogAnalogClocks_Circle)(Display_Frame, INTEGER, INTEGER, INTEGER, INTEGER, INTEGER);

export LONGINT *DialogAnalogClocks_ItemDesc__typ;

static void DialogAnalogClocks_Format (LONGINT time, INTEGER *sec, INTEGER *min, INTEGER *hour, INTEGER *hourm);
static void DialogAnalogClocks_Init (void);
export void DialogAnalogClocks_Insert (void);
static void DialogAnalogClocks_Line2 (Display_Frame f, INTEGER ang, INTEGER x0, INTEGER y0, INTEGER r1, INTEGER r2, INTEGER color);
static void DialogAnalogClocks_Line3 (Display_Frame f, INTEGER ang, INTEGER x0, INTEGER y0, INTEGER r1, INTEGER r2, INTEGER color);
static INTEGER DialogAnalogClocks_Min (INTEGER x, INTEGER y);
static void DialogAnalogClocks_PCircle (Display_Frame f, INTEGER x0, INTEGER y0, INTEGER r, INTEGER color, INTEGER mode);
static void DialogAnalogClocks_PLine (Display_Frame f, INTEGER x1, INTEGER y1, INTEGER x2, INTEGER y2, INTEGER color, INTEGER mode);
static void DialogAnalogClocks_SCircle (Display_Frame f, INTEGER x0, INTEGER y0, INTEGER r, INTEGER color, INTEGER mode);
static void DialogAnalogClocks_SLine (Display_Frame f, INTEGER x1, INTEGER y1, INTEGER x2, INTEGER y2, INTEGER color, INTEGER mode);


static INTEGER DialogAnalogClocks_Min (INTEGER x, INTEGER y)
{
	if (x < y) {
		return x;
	} else {
		return y;
	}
	__RETCHK;
}

static void DialogAnalogClocks_Init (void)
{
	INTEGER i = 0;
	i = 0;
	while (i < 60) {
		DialogAnalogClocks_sin[__X(i, 60)] = Math_sin(  1.0471976e-001 * i);
		DialogAnalogClocks_cos[__X(i, 60)] = Math_cos(  1.0471976e-001 * i);
		i += 1;
	}
}

static void DialogAnalogClocks_Format (LONGINT time, INTEGER *sec, INTEGER *min, INTEGER *hour, INTEGER *hourm)
{
	*hour = (int)__MASK(__ASHR(time, 12), -32);
	*min = (int)__MASK(__ASHR(time, 6), -64);
	*sec = (int)__MASK(time, -64);
	*hourm = (int)__MOD(*hour, 12) * 5 + __DIV(*min, 12);
}

static struct SCircle__13 {
	Display_Frame *f;
	struct SCircle__13 *lnk;
} *SCircle__13_s;

static void Dot4__14 (INTEGER x1, INTEGER x2, INTEGER y1, INTEGER y2, INTEGER color, INTEGER mode);

static void Dot4__14 (INTEGER x1, INTEGER x2, INTEGER y1, INTEGER y2, INTEGER color, INTEGER mode)
{
	Display_DotC(*SCircle__13_s->f, color, x1, y1, mode);
	Display_DotC(*SCircle__13_s->f, color, x1, y2, mode);
	Display_DotC(*SCircle__13_s->f, color, x2, y1, mode);
	Display_DotC(*SCircle__13_s->f, color, x2, y2, mode);
}

static void DialogAnalogClocks_SCircle (Display_Frame f, INTEGER x0, INTEGER y0, INTEGER r, INTEGER color, INTEGER mode)
{
	INTEGER x = 0, y = 0, dx = 0, dy = 0, d = 0;
	struct SCircle__13 _s;
	_s.f = &f;
	_s.lnk = SCircle__13_s;
	SCircle__13_s = &_s;
	x = r;
	y = 0;
	dx = __ASHL(x - 1, 3);
	dy = __ASHL(y, 3) + 4;
	d = 1 - __ASHL(r, 2);
	while (x > y) {
		Dot4__14(x0 - x, x0 + x, y0 - y, y0 + y, color, mode);
		Dot4__14(x0 - y, x0 + y, y0 - x, y0 + x, color, mode);
		d += dy;
		dy += 8;
		y += 1;
		if (d >= 0) {
			d -= dx;
			dx -= 8;
			x -= 1;
		}
	}
	if (x == y) {
		Dot4__14(x0 - x, x0 + x, y0 - y, y0 + y, color, mode);
	}
	SCircle__13_s = _s.lnk;
}

static void DialogAnalogClocks_SLine (Display_Frame f, INTEGER x1, INTEGER y1, INTEGER x2, INTEGER y2, INTEGER color, INTEGER mode)
{
	INTEGER x = 0, y = 0, dx = 0, dy = 0, d = 0, inc = 0;
	if (y2 - y1 < x1 - x2) {
		x = x1;
		x1 = x2;
		x2 = x;
		y = y1;
		y1 = y2;
		y2 = y;
	}
	dx = __ASHL(x2 - x1, 1);
	dy = __ASHL(y2 - y1, 1);
	x = x1;
	y = y1;
	inc = 1;
	if (dy > dx) {
		d = __ASHR(dy, 1);
		if (dx < 0) {
			inc = -1;
			dx = -dx;
		}
		while (y <= y2) {
			Display_DotC(f, color, x, y, mode);
			y += 1;
			d -= dx;
			if (d < 0) {
				d += dy;
				x += inc;
			}
		}
	} else {
		d = __ASHR(dx, 1);
		if (dy < 0) {
			inc = -1;
			dy = -dy;
		}
		while (x <= x2) {
			Display_DotC(f, color, x, y, mode);
			x += 1;
			d -= dy;
			if (d < 0) {
				d += dx;
				y += inc;
			}
		}
	}
}

static void DialogAnalogClocks_PCircle (Display_Frame f, INTEGER x0, INTEGER y0, INTEGER r, INTEGER color, INTEGER mode)
{
	Printer_Circle(x0, y0, r);
}

static void DialogAnalogClocks_PLine (Display_Frame f, INTEGER x1, INTEGER y1, INTEGER x2, INTEGER y2, INTEGER color, INTEGER mode)
{
	Printer_Line(x1, y1, x2, y2);
}

static void DialogAnalogClocks_Line2 (Display_Frame f, INTEGER ang, INTEGER x0, INTEGER y0, INTEGER r1, INTEGER r2, INTEGER color)
{
	INTEGER x1 = 0, y1 = 0, x2 = 0, y2 = 0;
	ang = (int)__MOD(15 - ang, 60);
	x1 = (int)__ENTIER(r1 * DialogAnalogClocks_cos[__X(ang, 60)] +   5.0000000e-001);
	y1 = (int)__ENTIER(r1 * DialogAnalogClocks_sin[__X(ang, 60)] +   5.0000000e-001);
	x2 = (int)__ENTIER(r2 * DialogAnalogClocks_cos[__X(ang, 60)] +   5.0000000e-001);
	y2 = (int)__ENTIER(r2 * DialogAnalogClocks_sin[__X(ang, 60)] +   5.0000000e-001);
	(*DialogAnalogClocks_Line)(f, x0 + x1, y0 + y1, x0 + x2, y0 + y2, color, 2);
}

static void DialogAnalogClocks_Line3 (Display_Frame f, INTEGER ang, INTEGER x0, INTEGER y0, INTEGER r1, INTEGER r2, INTEGER color)
{
	INTEGER x1 = 0, y1 = 0, x2 = 0, y2 = 0;
	ang = (int)__MOD(15 - ang, 60);
	x1 = (int)__ENTIER(r1 * DialogAnalogClocks_cos[__X(ang, 60)] +   5.0000000e-001);
	y1 = (int)__ENTIER(r1 * DialogAnalogClocks_sin[__X(ang, 60)] +   5.0000000e-001);
	x2 = (int)__ENTIER(r2 * DialogAnalogClocks_cos[__X(ang, 60)] +   5.0000000e-001);
	y2 = (int)__ENTIER(r2 * DialogAnalogClocks_sin[__X(ang, 60)] +   5.0000000e-001);
	(*DialogAnalogClocks_Line)(f, x0 + x1, y0 + y1, x0 + x2, y0 + y2, color, 1);
}

void DialogAnalogClocks_Item_Draw (DialogAnalogClocks_Item c, INTEGER x, INTEGER y, Display_Frame f)
{
	INTEGER r = 0, rh = 0, rm = 0, rs = 0, i = 0, sec = 0, min = 0, hour = 0, hourm = 0, mode = 0, ox = 0, oy = 0, w = 0, h = 0;
	DialogAnalogClocks_Line = DialogAnalogClocks_SLine;
	DialogAnalogClocks_Circle = DialogAnalogClocks_SCircle;
	__Dialogs_Object_GetDim((void*)c, &ox, &oy, &w, &h);
	Display_ReplConstC(f, __GUARDP(f, DialogFrames_FrameDesc, 1)->col, x, y, w, h, 1);
	r = __ASHR(DialogAnalogClocks_Min(w - 1, h - 1), 1);
	x = x + r;
	y = y + r;
	if (c->selected) {
		mode = 2;
	} else {
		mode = 0;
	}
	if (r >= 12) {
		rh = __DIV(7 * r, 11);
		rm = __DIV(9 * r, 11);
		rs = __DIV(10 * r, 11);
		i = 0;
		while (i < 60) {
			DialogAnalogClocks_Line3(f, i, x, y, rm, r, 15);
			i += 5;
		}
		DialogAnalogClocks_Format(DialogClocks_old.timeStamp, &sec, &min, &hour, &hourm);
		DialogAnalogClocks_Line2(f, sec, x, y, rm - r, rs, 15);
		DialogAnalogClocks_Line2(f, min, x, y, 0, rm, 15);
		DialogAnalogClocks_Line2(f, hourm, x, y, 0, rh, 15);
		(*DialogAnalogClocks_Circle)(f, x, y, 2, 15, mode);
	}
	(*DialogAnalogClocks_Circle)(f, x, y, r, 15, mode);
}

void DialogAnalogClocks_Item_Print (DialogAnalogClocks_Item c, INTEGER x, INTEGER y)
{
	INTEGER ox = 0, oy = 0, w = 0, h = 0, r = 0, sec = 0, min = 0, hour = 0, hourm = 0, i = 0, rh = 0, rm = 0, rs = 0;
	Display_Frame f = NIL;
	DialogAnalogClocks_Line = DialogAnalogClocks_PLine;
	DialogAnalogClocks_Circle = DialogAnalogClocks_PCircle;
	__Dialogs_Object_GetPDim((void*)c, &ox, &oy, &w, &h);
	r = __ASHR(DialogAnalogClocks_Min(w - 1, h - 1), 1);
	x = x + r;
	y = y + r;
	if (r >= 12) {
		rh = __DIV(7 * r, 11);
		rm = __DIV(9 * r, 11);
		rs = __DIV(10 * r, 11);
		i = 0;
		while (i < 60) {
			DialogAnalogClocks_Line2(f, i, x, y, rm, r, 15);
			i += 5;
		}
		DialogAnalogClocks_Format(DialogClocks_old.timeStamp, &sec, &min, &hour, &hourm);
		DialogAnalogClocks_Line2(f, sec, x, y, rm - r, rs, 15);
		DialogAnalogClocks_Line2(f, min, x, y, 0, rm, 15);
		DialogAnalogClocks_Line2(f, hourm, x, y, 0, rh, 15);
		(*DialogAnalogClocks_Circle)(f, x, y, (int)__DIV(__ASHL(Dialogs_dUnit, 1), Dialogs_pUnit), 15, -1);
	}
	(*DialogAnalogClocks_Circle)(f, x, y, r, 15, -1);
}

void DialogAnalogClocks_Item_Redraw (DialogAnalogClocks_Item c, Display_Frame f, INTEGER x, INTEGER y, DialogClocks_Time old, DialogClocks_Time new)
{
	INTEGER rh = 0, rm = 0, rs = 0, olds = 0, oldm = 0, oldh = 0, oldhm = 0, news = 0, newm = 0, newh = 0, newhm = 0, ox = 0, oy = 0, w = 0, h = 0, r = 0, mode = 0;
	__Dialogs_Object_GetDim((void*)c, &ox, &oy, &w, &h);
	mode = 0;
	r = __ASHR(DialogAnalogClocks_Min(w - 1, h - 1), 1);
	x = x + r;
	y = y + r;
	if (c->selected) {
		return;
	}
	DialogAnalogClocks_Line = DialogAnalogClocks_SLine;
	DialogAnalogClocks_Circle = DialogAnalogClocks_SCircle;
	if (r >= 12) {
		rh = __DIV(7 * r, 11);
		rm = __DIV(9 * r, 11);
		rs = __DIV(10 * r, 11);
		DialogAnalogClocks_Format(old.timeStamp, &olds, &oldm, &oldh, &oldhm);
		DialogAnalogClocks_Format(new.timeStamp, &news, &newm, &newh, &newhm);
		if (olds != news) {
			DialogAnalogClocks_Line2(f, olds, x, y, rm - r, rs, 15);
			DialogAnalogClocks_Line2(f, news, x, y, rm - r, rs, 15);
		}
		if (oldm != newm) {
			DialogAnalogClocks_Line2(f, oldm, x, y, 0, rm, 15);
			DialogAnalogClocks_Line2(f, newm, x, y, 0, rm, 15);
		}
		if (oldhm != newhm) {
			DialogAnalogClocks_Line2(f, oldhm, x, y, 0, rh, 15);
			DialogAnalogClocks_Line2(f, newhm, x, y, 0, rh, 15);
		}
		(*DialogAnalogClocks_Circle)(f, x, y, 2, 15, mode);
	}
}

void DialogAnalogClocks_Item_Copy (DialogAnalogClocks_Item c, Dialogs_Object *dup)
{
	DialogAnalogClocks_Item x = NIL;
	if (*dup == NIL) {
		__NEW(x, DialogAnalogClocks_ItemDesc);
		*dup = (Dialogs_Object)x;
	} else {
		x = __GUARDP(*dup, DialogAnalogClocks_ItemDesc, 2);
	}
	DialogClocks_Item_Copy((void*)c, &*dup);
}

void DialogAnalogClocks_Insert (void)
{
	INTEGER x = 0, y = 0, x1 = 0, y1 = 0, w = 0, h = 0;
	DialogAnalogClocks_Item c = NIL;
	Dialogs_Panel p = NIL;
	CHAR name[64];
	__NEW(c, DialogAnalogClocks_ItemDesc);
	DialogFrames_GetCaretPosition(&p, &x, &y);
	if (p != NIL) {
		__Dialogs_Object_Init((void*)c);
		In_Open();
		In_Name((void*)name, 64);
		if (!In_Done) {
			__COPY("", name, 64);
			In_Open();
		}
		__Dialogs_Object_SetName((void*)c, name, 64);
		In_Int(&x1);
		In_Int(&y1);
		In_Int(&w);
		In_Int(&h);
		if (!In_Done) {
			x1 = x;
			y1 = y;
			w = 65;
			h = 65;
		} else {
			if (w < 0) {
				w = 65;
			}
			if (h < 0) {
				h = 65;
			}
		}
		__Dialogs_Object_SetDim((void*)c, x, y, 65, 65, 0);
		__Dialogs_Panel_Insert(p, (void*)c, 0);
	} else {
		Dialogs_res = 5;
	}
	if (Dialogs_res != 0) {
		Dialogs_Error((CHAR*)"DialogClocks", (LONGINT)13);
	}
}

__TDESC(DialogAnalogClocks_ItemDesc, 27, 2) = {__TDFLDS("ItemDesc", 112), {0, 104, -12}};

export void *DialogAnalogClocks__init(void)
{
	__DEFMOD;
	__IMPORT(DialogClocks);
	__IMPORT(DialogFrames);
	__IMPORT(Dialogs);
	__IMPORT(Display);
	__IMPORT(In);
	__IMPORT(Math);
	__IMPORT(Oberon);
	__IMPORT(Printer);
	__REGMOD(DialogAnalogClocks, 0);
	__REGCMD("Insert", DialogAnalogClocks_Insert);
	__INITYP(DialogAnalogClocks_ItemDesc, DialogClocks_ItemDesc, 2);
	__INITBP(DialogAnalogClocks_ItemDesc, DialogAnalogClocks_Item_Copy, 1);
	__INITBP(DialogAnalogClocks_ItemDesc, DialogAnalogClocks_Item_Draw, 2);
	__INITBP(DialogAnalogClocks_ItemDesc, DialogAnalogClocks_Item_Print, 14);
	__INITBP(DialogAnalogClocks_ItemDesc, DialogAnalogClocks_Item_Redraw, 24);
/* BEGIN */
	DialogAnalogClocks_Init();
	__ENDMOD;
}
