/* Ofront 1.4 -xtspkaG */
#include "SYSTEM.h"
#include "Dialog.h"
#include "DialogFrames.h"
#include "Dialogs.h"
#include "Display.h"
#include "Files.h"
#include "Input.h"
#include "MenuViewers.h"
#include "Oberon.h"
#include "Printer.h"
#include "TextFrames.h"
#include "TextPrinter.h"
#include "Texts.h"
#include "Viewers.h"

typedef
	struct DialogElems_ElemDesc *DialogElems_Elem;

typedef
	struct DialogElems_ElemDesc { /* Texts_ElemDesc */
		char _prvt0[20];
		LONGINT W, H;
		Texts_Handler handle;
		char _prvt1[4];
		Dialogs_Panel dialog;
		INTEGER xg, yg;
		LONGINT sw, sh;
		BOOLEAN empty;
	} DialogElems_ElemDesc;

typedef
	struct DialogElems_FrameDesc *DialogElems_Frame;

typedef
	struct DialogElems_FrameDesc { /* DialogFrames_FrameDesc */
		Display_Frame dsc, next;
		INTEGER X, Y, W, H;
		Display_Handler handle;
		INTEGER col;
		Dialogs_Panel panel;
		INTEGER grid;
		LONGINT pat;
		DialogElems_Elem elem;
	} DialogElems_FrameDesc;


static INTEGER DialogElems_x0, DialogElems_x1, DialogElems_y0, DialogElems_y1;
static Texts_Writer DialogElems_w;
static LONGINT DialogElems_dUnit, DialogElems_pUnit;
static INTEGER DialogElems_gripW, DialogElems_gripH;

export LONGINT *DialogElems_ElemDesc__typ;
export LONGINT *DialogElems_FrameDesc__typ;

export void DialogElems_Alloc (void);
export void DialogElems_Changed (DialogElems_Elem e);
export void DialogElems_Copy (DialogElems_Elem se, DialogElems_Elem de);
static void DialogElems_FlipGrip (INTEGER x0, INTEGER y0, INTEGER w, INTEGER h);
export void DialogElems_Handle (Texts_Elem e, Texts_ElemMsg *msg, LONGINT *msg__typ);
export void DialogElems_Insert (void);
export void DialogElems_Open (DialogElems_Elem e, Dialogs_Panel p, INTEGER xg, INTEGER yg, BOOLEAN adjust);
export void DialogElems_OpenViewer (DialogElems_Elem e);
export void DialogElems_SetSize (DialogElems_Elem e, LONGINT w, LONGINT h);
static void DialogElems_Track (DialogElems_Elem e, SET keys, INTEGER x, INTEGER y, INTEGER x0, INTEGER y0);
export void DialogElems_Update (void);
static void DialogElems_box (Dialogs_Object obj, BOOLEAN *done);


void DialogElems_Changed (DialogElems_Elem e)
{
	Texts_Text t = NIL;
	LONGINT pos = 0;
	t = Texts_ElemBase((void*)e);
	pos = Texts_ElemPos((void*)e);
	(*t->notify)(t, 0, pos, pos + 1);
}

static void DialogElems_FlipGrip (INTEGER x0, INTEGER y0, INTEGER w, INTEGER h)
{
	Display_ReplConst(15, (x0 + w) - DialogElems_gripW, y0, DialogElems_gripW, DialogElems_gripH, 2);
}

void DialogElems_SetSize (DialogElems_Elem e, LONGINT w, LONGINT h)
{
	if (w < 108000) {
		w = 108000;
	}
	if (h < 108000) {
		h = 108000;
	}
	e->W = w;
	e->H = h;
	e->sw = w;
	e->sh = h;
}

static void DialogElems_box (Dialogs_Object obj, BOOLEAN *done)
{
	INTEGER x = 0, y = 0, w = 0, h = 0;
	__Dialogs_Object_GetDim(obj, &x, &y, &w, &h);
	y = __ABS(y);
	if (x < DialogElems_x0) {
		DialogElems_x0 = x;
	}
	if (DialogElems_x1 < x + w) {
		DialogElems_x1 = x + w;
	}
	if (y - h < DialogElems_y0) {
		DialogElems_y0 = y - h;
	}
	if (DialogElems_y1 < y) {
		DialogElems_y1 = y;
	}
}

void DialogElems_Open (DialogElems_Elem e, Dialogs_Panel p, INTEGER xg, INTEGER yg, BOOLEAN adjust)
{
	e->dialog = p;
	DialogElems_x0 = 32767;
	DialogElems_x1 = -32768;
	DialogElems_y0 = 32767;
	DialogElems_y1 = -32768;
	__Dialogs_Panel_Enumerate(p, DialogElems_box);
	if (DialogElems_x0 == 32767) {
		e->empty = 1;
		e->xg = 0;
		e->yg = 0;
		DialogElems_SetSize(e, 0, 0);
	} else {
		DialogElems_y1 = DialogElems_y1 + DialogElems_y0;
		DialogElems_x1 = DialogElems_x1 + DialogElems_x0;
		DialogElems_x0 = 0;
		DialogElems_y0 = 0;
		e->empty = 0;
		if (adjust) {
			e->xg = -DialogElems_x0;
			e->yg = -DialogElems_y1;
			DialogElems_SetSize(e, (LONGINT)(DialogElems_x1 - DialogElems_x0) * DialogElems_dUnit, (LONGINT)(DialogElems_y1 - DialogElems_y0) * DialogElems_dUnit);
		} else {
			e->xg = xg;
			e->yg = yg;
			DialogElems_SetSize(e, e->W, e->H);
		}
	}
}

void DialogElems_Copy (DialogElems_Elem se, DialogElems_Elem de)
{
	se->W = se->sw;
	se->H = se->sh;
	Texts_CopyElem((void*)se, (void*)de);
	DialogElems_Open(de, __Dialogs_Panel_Copy(se->dialog), se->xg, se->yg, 0);
}

void DialogElems_OpenViewer (DialogElems_Elem e)
{
	Viewers_Viewer v = NIL;
	DialogElems_Frame f = NIL;
	INTEGER x = 0, y = 0;
	__NEW(f, DialogElems_FrameDesc);
	__DialogFrames_Frame_Open((void*)f, Dialog_Handle, __Dialogs_Panel_Copy(e->dialog));
	f->col = 11;
	f->elem = e;
	Oberon_AllocateUserViewer(Oberon_Mouse.X, &x, &y);
	v = (Viewers_Viewer)MenuViewers_New((void*)TextFrames_NewMenu((CHAR*)"DialogElems.Panel", (LONGINT)18, (CHAR*)"^DialogElem.Menu.Text", (LONGINT)22), (void*)f, TextFrames_menuH, x, y);
}

static void DialogElems_Track (DialogElems_Elem e, SET keys, INTEGER x, INTEGER y, INTEGER x0, INTEGER y0)
{
	SET keysum = 0;
	INTEGER x1 = 0, y1 = 0, w = 0, h = 0;
	BOOLEAN hit = 0;
	if (__IN(1, keys)) {
		x1 = x - x0;
		y1 = y - y0;
		keysum = keys;
		w = (int)__DIV(e->W, DialogElems_dUnit);
		h = (int)__DIV(e->H, DialogElems_dUnit);
		hit = ((!e->empty && x1 >= w - DialogElems_gripW) && 0 <= y1) && y1 < DialogElems_gripH;
		if (hit) {
			DialogElems_FlipGrip(x0, y0, w, h);
		}
		do {
			Input_Mouse(&keys, &x, &y);
			Oberon_DrawCursor(&Oberon_Mouse, Oberon_Cursor__typ, &Oberon_Arrow, Oberon_Marker__typ, x, y);
			keysum = keysum | keys;
		} while (!(keys == 0x0));
		if (hit) {
			DialogElems_FlipGrip(x0, y0, w, h);
		}
		if (keysum == 0x02) {
			if (hit) {
				w += (x - x0) - x1;
				h -= (y - y0) - y1;
				DialogElems_SetSize(e, (LONGINT)w * DialogElems_dUnit, (LONGINT)h * DialogElems_dUnit);
				DialogElems_Changed(e);
			} else {
				DialogElems_OpenViewer(e);
			}
		}
	}
}

void DialogElems_Handle (Texts_Elem e, Texts_ElemMsg *msg, LONGINT *msg__typ)
{
	DialogElems_Elem e1 = NIL;
	INTEGER x = 0, y = 0, w = 0, h = 0, res = 0;
	DialogFrames_Frame f = NIL;
	Dialogs_Panel p = NIL;
	if (__ISP(e, DialogElems_ElemDesc, 2)) {
		if (__IS(msg__typ, TextFrames_DisplayMsg, 1)) {
			TextFrames_DisplayMsg *msg__ = (void*)msg;
			if ((*msg__).prepare) {
				(*(DialogElems_Elem*)&e)->W = (*(DialogElems_Elem*)&e)->sw;
				(*(DialogElems_Elem*)&e)->H = (*(DialogElems_Elem*)&e)->sh;
			} else {
				x = (*msg__).X0;
				y = (*msg__).Y0;
				w = (int)__DIV((*(DialogElems_Elem*)&e)->W, DialogElems_dUnit);
				h = (int)__DIV((*(DialogElems_Elem*)&e)->H, DialogElems_dUnit);
				if ((*(DialogElems_Elem*)&e)->empty) {
					Display_ReplPattern(15, Display_grey1, x, y, w, h, 0);
				} else {
					__NEW(f, DialogFrames_FrameDesc);
					__DialogFrames_Frame_Open(f, DialogFrames_Handle, (*(DialogElems_Elem*)&e)->dialog);
					f->X = x;
					f->Y = y;
					f->W = w;
					f->H = h;
					__DialogFrames_Frame_Restore(f);
					DialogElems_FlipGrip(x, y, w, h);
					(*msg__).elemFrame = (Display_Frame)f;
				}
			}
		} else if (__IS(msg__typ, TextPrinter_PrintMsg, 1)) {
			TextPrinter_PrintMsg *msg__ = (void*)msg;
			if ((*msg__).prepare) {
				(*(DialogElems_Elem*)&e)->W = __ASHL(__DIV((*(DialogElems_Elem*)&e)->sw, DialogElems_dUnit), 2) * DialogElems_pUnit;
				(*(DialogElems_Elem*)&e)->H = __ASHL(__DIV((*(DialogElems_Elem*)&e)->sh, DialogElems_dUnit), 2) * DialogElems_pUnit;
			} else {
				__Dialogs_Panel_Print((*(DialogElems_Elem*)&e)->dialog, (*msg__).X0, (*msg__).Y0 + (int)__DIV((*(DialogElems_Elem*)&e)->H, DialogElems_pUnit));
				(*(DialogElems_Elem*)&e)->W = (*(DialogElems_Elem*)&e)->sw;
				(*(DialogElems_Elem*)&e)->H = (*(DialogElems_Elem*)&e)->sh;
			}
		} else if (__IS(msg__typ, Texts_IdentifyMsg, 1)) {
			Texts_IdentifyMsg *msg__ = (void*)msg;
			__MOVE("DialogElems", (*msg__).mod, 12);
			__MOVE("Alloc", (*msg__).proc, 6);
		} else if (__IS(msg__typ, Texts_FileMsg, 1)) {
			Texts_FileMsg *msg__ = (void*)msg;
			if ((*msg__).id == 0) {
				__NEW(p, Dialogs_PanelDesc);
				__Dialogs_Panel_Load(p, &(*msg__).r, Files_Rider__typ);
				Files_ReadInt(&(*msg__).r, Files_Rider__typ, &(*(DialogElems_Elem*)&e)->xg);
				Files_ReadInt(&(*msg__).r, Files_Rider__typ, &(*(DialogElems_Elem*)&e)->yg);
				DialogElems_Open((*(DialogElems_Elem*)&e), p, (*(DialogElems_Elem*)&e)->xg, (*(DialogElems_Elem*)&e)->yg, 0);
			} else if ((*msg__).id == 1) {
				(*(DialogElems_Elem*)&e)->W = (*(DialogElems_Elem*)&e)->sw;
				(*(DialogElems_Elem*)&e)->H = (*(DialogElems_Elem*)&e)->sh;
				__Dialogs_Panel_Store((*(DialogElems_Elem*)&e)->dialog, &(*msg__).r, Files_Rider__typ);
				Files_WriteInt(&(*msg__).r, Files_Rider__typ, (*(DialogElems_Elem*)&e)->xg);
				Files_WriteInt(&(*msg__).r, Files_Rider__typ, (*(DialogElems_Elem*)&e)->yg);
			}
		} else if (__IS(msg__typ, Texts_CopyMsg, 1)) {
			Texts_CopyMsg *msg__ = (void*)msg;
			__NEW(e1, DialogElems_ElemDesc);
			DialogElems_Copy((*(DialogElems_Elem*)&e), e1);
			(*msg__).e = (Texts_Elem)e1;
		} else if (__IS(msg__typ, TextFrames_TrackMsg, 1)) {
			TextFrames_TrackMsg *msg__ = (void*)msg;
			DialogElems_Track((*(DialogElems_Elem*)&e), (*msg__).keys, (*msg__).X, (*msg__).Y, (*msg__).X0, (*msg__).Y0);
		} else if (__IS(msg__typ, TextFrames_FocusMsg, 1)) {
			TextFrames_FocusMsg *msg__ = (void*)msg;
			f = __GUARDP((*msg__).elemFrame, DialogFrames_FrameDesc, 1);
			__DialogFrames_Frame_Restore(f);
			if ((*msg__).focus && f->panel->cmd[0] != 0x00) {
				Dialogs_cmdPanel = f->panel;
				Oberon_Call(f->panel->cmd, 64, Oberon_Par, 0, &res);
			}
			if (!(*msg__).focus) {
				DialogElems_FlipGrip(f->X, f->Y, f->W, f->H);
			}
		}
	}
}

void DialogElems_Alloc (void)
{
	DialogElems_Elem e = NIL;
	__NEW(e, DialogElems_ElemDesc);
	e->handle = DialogElems_Handle;
	Texts_new = (Texts_Elem)e;
}

void DialogElems_Insert (void)
{
	Texts_Scanner s = {0};
	Texts_Text text = NIL;
	LONGINT beg = 0, end = 0, time = 0;
	Viewers_Viewer v = NIL;
	Dialogs_Panel p = NIL;
	DialogElems_Elem e = NIL;
	TextFrames_InsertElemMsg msg = {0};
	Files_File file = NIL;
	Files_Rider r = {0};
	Texts_OpenScanner(&s, Texts_Scanner__typ, Oberon_Par->text, Oberon_Par->pos);
	Texts_Scan(&s, Texts_Scanner__typ);
	if (s.class == 6 && s.c == '^') {
		Oberon_GetSelection(&text, &beg, &end, &time);
		if (time >= 0) {
			Texts_OpenScanner(&s, Texts_Scanner__typ, text, beg);
			Texts_Scan(&s, Texts_Scanner__typ);
		}
	}
	__NEW(p, Dialogs_PanelDesc);
	if (s.class == 6 && s.c == '*') {
		v = Oberon_MarkedViewer();
		if ((v->dsc != NIL && v->dsc->next != NIL) && __ISP(v->dsc->next, DialogFrames_FrameDesc, 1)) {
			p = __Dialogs_Panel_Copy(__GUARDP(v->dsc->next, DialogFrames_FrameDesc, 1)->panel);
		}
	} else if (s.class == 1) {
		file = Files_Old(s.s, 256);
		if (file != NIL) {
			Files_Set(&r, Files_Rider__typ, file, 0);
			__Dialogs_Panel_Load(p, &r, Files_Rider__typ);
		}
	}
	__NEW(e, DialogElems_ElemDesc);
	e->handle = DialogElems_Handle;
	DialogElems_Open(e, p, 0, 0, 1);
	msg.e = (Texts_Elem)e;
	(*Oberon_FocusViewer->handle)((void*)Oberon_FocusViewer, (void*)&msg, TextFrames_InsertElemMsg__typ);
}

void DialogElems_Update (void)
{
	Viewers_Viewer v = NIL;
	DialogElems_Frame f = NIL;
	Texts_Reader r = {0};
	Texts_Text t = NIL;
	CHAR ch = 0;
	v = Oberon_Par->vwr;
	f = __GUARDP(v->dsc->next, DialogElems_FrameDesc, 2);
	t = __GUARDP(v->dsc, TextFrames_FrameDesc, 1)->text;
	__Dialogs_Panel_RemoveSelections(f->panel);
	DialogElems_Open(f->elem, __Dialogs_Panel_Copy(f->panel), 0, 0, 1);
	DialogElems_Changed(f->elem);
	Texts_OpenReader(&r, Texts_Reader__typ, t, t->len - 1);
	Texts_Read(&r, Texts_Reader__typ, &ch);
	if (ch == '!') {
		Texts_Delete(t, t->len - 1, t->len);
	}
}

static void EnumPtrs(void (*P)(void*))
{
	__ENUMR(&DialogElems_w, Texts_Writer__typ, 36, 1, P);
}

__TDESC(DialogElems_ElemDesc, 1, 5) = {__TDFLDS("ElemDesc", 56), {0, 4, 12, 32, 36, -24}};
__TDESC(DialogElems_FrameDesc, 10, 4) = {__TDFLDS("FrameDesc", 40), {0, 4, 24, 36, -20}};

export void *DialogElems__init(void)
{
	__DEFMOD;
	__IMPORT(Dialog);
	__IMPORT(DialogFrames);
	__IMPORT(Dialogs);
	__IMPORT(Display);
	__IMPORT(Files);
	__IMPORT(Input);
	__IMPORT(MenuViewers);
	__IMPORT(Oberon);
	__IMPORT(Printer);
	__IMPORT(TextFrames);
	__IMPORT(TextPrinter);
	__IMPORT(Texts);
	__IMPORT(Viewers);
	__REGMOD(DialogElems, EnumPtrs);
	__REGCMD("Alloc", DialogElems_Alloc);
	__REGCMD("Insert", DialogElems_Insert);
	__REGCMD("Update", DialogElems_Update);
	__INITYP(DialogElems_ElemDesc, Texts_ElemDesc, 2);
	__INITYP(DialogElems_FrameDesc, DialogFrames_FrameDesc, 2);
/* BEGIN */
	DialogElems_dUnit = Dialogs_dUnit;
	DialogElems_pUnit = Dialogs_pUnit;
	DialogElems_gripW = (int)__DIV(54000, DialogElems_dUnit);
	DialogElems_gripH = DialogElems_gripW;
	__ENDMOD;
}
