/* Ofront 1.4 -xtspkaG */
#include "SYSTEM.h"
#include "Display.h"
#include "Fonts.h"
#include "Graphics.h"
#include "Input.h"
#include "MenuViewers.h"
#include "Oberon.h"
#include "Texts.h"
#include "Viewers.h"

typedef
	struct GraphicFrames_FrameDesc *GraphicFrames_Frame;

typedef
	struct GraphicFrames_CtrlMsg { /* Graphics_Msg */
		GraphicFrames_Frame f;
		INTEGER res;
	} GraphicFrames_CtrlMsg;

typedef
	struct GraphicFrames_DispMsg { /* Display_FrameMsg */
		INTEGER x1, y1, w;
		LONGINT pat;
		Graphics_Graph graph;
	} GraphicFrames_DispMsg;

typedef
	struct GraphicFrames_DrawMsg { /* Graphics_Msg */
		GraphicFrames_Frame f;
		INTEGER x, y, col, mode;
	} GraphicFrames_DrawMsg;

typedef
	struct GraphicFrames_FocusQuery { /* Display_FrameMsg */
		GraphicFrames_Frame f;
	} GraphicFrames_FocusQuery;

typedef
	struct GraphicFrames_LocDesc *GraphicFrames_Location;

typedef
	struct GraphicFrames_LocDesc {
		INTEGER x, y;
		GraphicFrames_Location next;
	} GraphicFrames_LocDesc;

typedef
	struct GraphicFrames_FrameDesc { /* Display_FrameDesc */
		Display_Frame dsc, next;
		INTEGER X, Y, W, H;
		Display_Handler handle;
		Graphics_Graph graph;
		INTEGER Xg, Yg, X1, Y1, x, y, col;
		BOOLEAN marked, ticked;
		GraphicFrames_LocDesc mark;
	} GraphicFrames_FrameDesc;

typedef
	struct GraphicFrames_PosQuery { /* Display_FrameMsg */
		GraphicFrames_Frame f;
		INTEGER x, y;
	} GraphicFrames_PosQuery;

typedef
	struct GraphicFrames_SelQuery { /* Display_FrameMsg */
		GraphicFrames_Frame f;
		LONGINT time;
	} GraphicFrames_SelQuery;

typedef
	struct GraphicFrames_UpdateMsg { /* Display_FrameMsg */
		INTEGER id;
		Graphics_Graph graph;
		Graphics_Object obj;
	} GraphicFrames_UpdateMsg;


static Graphics_Caption GraphicFrames_newcap;
static Texts_Writer GraphicFrames_W;

export LONGINT *GraphicFrames_LocDesc__typ;
export LONGINT *GraphicFrames_FrameDesc__typ;
export LONGINT *GraphicFrames_DrawMsg__typ;
export LONGINT *GraphicFrames_CtrlMsg__typ;
export LONGINT *GraphicFrames_UpdateMsg__typ;
export LONGINT *GraphicFrames_SelQuery__typ;
export LONGINT *GraphicFrames_FocusQuery__typ;
export LONGINT *GraphicFrames_PosQuery__typ;
export LONGINT *GraphicFrames_DispMsg__typ;

static Graphics_Caption GraphicFrames_CaptionCopy (GraphicFrames_Frame F, INTEGER x1, INTEGER y1, Texts_Text T, LONGINT beg, LONGINT end);
export void GraphicFrames_Change (GraphicFrames_Frame F, Graphics_Msg *msg, LONGINT *msg__typ);
export void GraphicFrames_Defocus (GraphicFrames_Frame F);
static void GraphicFrames_DeleteChar (GraphicFrames_Frame F);
export void GraphicFrames_Deselect (GraphicFrames_Frame F);
export void GraphicFrames_Draw (GraphicFrames_Frame F);
static void GraphicFrames_DrawCaption (Graphics_Object obj, Graphics_Msg *M, LONGINT *M__typ);
static void GraphicFrames_DrawLine (Graphics_Object obj, Graphics_Msg *M, LONGINT *M__typ);
static void GraphicFrames_DrawMacro (Graphics_Object obj, Graphics_Msg *M, LONGINT *M__typ);
static void GraphicFrames_DrawNorm (GraphicFrames_Frame F);
export void GraphicFrames_DrawObj (GraphicFrames_Frame F, Graphics_Object obj);
static void GraphicFrames_Edit (GraphicFrames_Frame F, INTEGER x0, INTEGER y0, SET k0);
export void GraphicFrames_Erase (GraphicFrames_Frame F);
export void GraphicFrames_EraseObj (GraphicFrames_Frame F, Graphics_Object obj);
static BOOLEAN GraphicFrames_ExtendedAscii (CHAR ch);
static void GraphicFrames_FlipMark (INTEGER x, INTEGER y);
export GraphicFrames_Frame GraphicFrames_Focus (void);
static void GraphicFrames_GetSelection (GraphicFrames_Frame F, Texts_Text *text, LONGINT *beg, LONGINT *end, LONGINT *time);
export void GraphicFrames_Handle (Display_Frame G, Display_FrameMsg *M, LONGINT *M__typ);
static void GraphicFrames_InsertChar (GraphicFrames_Frame F, CHAR ch);
export void GraphicFrames_Macro (CHAR *Lname, LONGINT Lname__len, CHAR *Mname, LONGINT Mname__len);
static void GraphicFrames_NewCaption (GraphicFrames_Frame F, INTEGER col, Fonts_Font font);
export void GraphicFrames_Open (GraphicFrames_Frame G, Graphics_Graph graph, INTEGER X, INTEGER Y, INTEGER col, BOOLEAN ticked);
export void GraphicFrames_Restore (GraphicFrames_Frame F);
export GraphicFrames_Frame GraphicFrames_Selected (void);
static void GraphicFrames_SendCaption (Graphics_Caption cap);
export GraphicFrames_Frame GraphicFrames_This (INTEGER x, INTEGER y);


static BOOLEAN GraphicFrames_ExtendedAscii (CHAR ch)
{
	return ch >= 0x80 && ch <= 0x85 || ch == 0xa7;
}

void GraphicFrames_Restore (GraphicFrames_Frame F)
{
	INTEGER x = 0, x0 = 0, y = 0;
	GraphicFrames_DrawMsg M = {0};
	F->X1 = F->X + F->W;
	F->Y1 = F->Y + F->H;
	F->x = F->X + F->Xg;
	F->y = F->Y1 + F->Yg;
	F->marked = 0;
	F->mark.next = NIL;
	Oberon_RemoveMarks(F->X, F->Y, F->W, F->H);
	Display_ReplConst(F->col, F->X, F->Y, F->W, F->H, 0);
	if (F->ticked) {
		y = (__MASK(F->Yg, -16) + F->Y1) - 16;
		x0 = __MASK(F->Xg, -16) + F->X;
		while (y >= F->Y) {
			x = x0;
			while (x < F->X1) {
				Display_Dot(15, x, y, 0);
				x += 16;
			}
			y -= 16;
		}
	}
	M.f = F;
	M.x = F->x;
	M.y = F->y;
	M.col = 0;
	M.mode = 0;
	Graphics_Draw(F->graph, (void*)&M, GraphicFrames_DrawMsg__typ);
}

GraphicFrames_Frame GraphicFrames_Focus (void)
{
	GraphicFrames_FocusQuery FQ = {0};
	FQ.f = NIL;
	Viewers_Broadcast((void*)&FQ, GraphicFrames_FocusQuery__typ);
	return FQ.f;
}

GraphicFrames_Frame GraphicFrames_Selected (void)
{
	GraphicFrames_SelQuery SQ = {0};
	SQ.f = NIL;
	SQ.time = 0;
	Viewers_Broadcast((void*)&SQ, GraphicFrames_SelQuery__typ);
	return SQ.f;
}

GraphicFrames_Frame GraphicFrames_This (INTEGER x, INTEGER y)
{
	GraphicFrames_PosQuery PQ = {0};
	PQ.f = NIL;
	PQ.x = x;
	PQ.y = y;
	Viewers_Broadcast((void*)&PQ, GraphicFrames_PosQuery__typ);
	return PQ.f;
}

void GraphicFrames_Draw (GraphicFrames_Frame F)
{
	GraphicFrames_UpdateMsg UM = {0};
	UM.id = 5;
	UM.graph = F->graph;
	Viewers_Broadcast((void*)&UM, GraphicFrames_UpdateMsg__typ);
}

static void GraphicFrames_DrawNorm (GraphicFrames_Frame F)
{
	GraphicFrames_UpdateMsg UM = {0};
	UM.id = 4;
	UM.graph = F->graph;
	Viewers_Broadcast((void*)&UM, GraphicFrames_UpdateMsg__typ);
}

void GraphicFrames_Erase (GraphicFrames_Frame F)
{
	GraphicFrames_UpdateMsg UM = {0};
	UM.id = 6;
	UM.graph = F->graph;
	Viewers_Broadcast((void*)&UM, GraphicFrames_UpdateMsg__typ);
}

void GraphicFrames_DrawObj (GraphicFrames_Frame F, Graphics_Object obj)
{
	GraphicFrames_UpdateMsg UM = {0};
	UM.id = 1;
	UM.graph = F->graph;
	UM.obj = obj;
	Viewers_Broadcast((void*)&UM, GraphicFrames_UpdateMsg__typ);
}

void GraphicFrames_EraseObj (GraphicFrames_Frame F, Graphics_Object obj)
{
	GraphicFrames_UpdateMsg UM = {0};
	UM.id = 3;
	UM.graph = F->graph;
	UM.obj = obj;
	Viewers_Broadcast((void*)&UM, GraphicFrames_UpdateMsg__typ);
}

void GraphicFrames_Change (GraphicFrames_Frame F, Graphics_Msg *msg, LONGINT *msg__typ)
{
	if (F != NIL) {
		GraphicFrames_Erase(F);
		Graphics_Handle(F->graph, &*msg, msg__typ);
		GraphicFrames_Draw(F);
	}
}

static void GraphicFrames_FlipMark (INTEGER x, INTEGER y)
{
	Display_ReplConst(15, x - 7, y, 15, 1, 2);
	Display_ReplConst(15, x, y - 7, 1, 15, 2);
}

void GraphicFrames_Defocus (GraphicFrames_Frame F)
{
	GraphicFrames_Location m = NIL;
	GraphicFrames_newcap = NIL;
	if (F->marked) {
		GraphicFrames_FlipMark(F->mark.x, F->mark.y);
		m = F->mark.next;
		while (m != NIL) {
			GraphicFrames_FlipMark(m->x, m->y);
			m = m->next;
		}
		F->marked = 0;
		F->mark.next = NIL;
	}
}

void GraphicFrames_Deselect (GraphicFrames_Frame F)
{
	GraphicFrames_UpdateMsg UM = {0};
	if (F != NIL) {
		UM.id = 4;
		UM.graph = F->graph;
		Viewers_Broadcast((void*)&UM, GraphicFrames_UpdateMsg__typ);
		Graphics_Deselect(F->graph);
	}
}

void GraphicFrames_Macro (CHAR *Lname, LONGINT Lname__len, CHAR *Mname, LONGINT Mname__len)
{
	INTEGER x = 0, y = 0;
	GraphicFrames_Frame F = NIL;
	Graphics_Macro mac = NIL;
	Graphics_MacHead mh = NIL;
	Graphics_Library L = NIL;
	F = GraphicFrames_Focus();
	if (F != NIL) {
		x = F->mark.x - F->x;
		y = F->mark.y - F->y;
		L = Graphics_ThisLib((void*)Lname, Lname__len, 0);
		if (L != NIL) {
			mh = Graphics_ThisMac(L, (void*)Mname, Mname__len);
			if (mh != NIL) {
				GraphicFrames_Deselect(F);
				GraphicFrames_Defocus(F);
				__NEW(mac, Graphics_MacroDesc);
				mac->x = x;
				mac->y = y;
				mac->w = mh->w;
				mac->h = mh->h;
				mac->mac = mh;
				mac->do_ = Graphics_MacMethod;
				mac->col = Oberon_CurCol;
				Graphics_Add(F->graph, (void*)mac);
				GraphicFrames_DrawObj(F, (void*)mac);
			}
		}
	}
}

static Graphics_Caption GraphicFrames_CaptionCopy (GraphicFrames_Frame F, INTEGER x1, INTEGER y1, Texts_Text T, LONGINT beg, LONGINT end)
{
	CHAR ch = 0;
	INTEGER dx = 0, w = 0, x2 = 0, y2 = 0, w1 = 0, h1 = 0;
	Graphics_Caption cap = NIL;
	LONGINT pat = 0;
	Texts_Reader R = {0};
	Texts_Write(&GraphicFrames_W, Texts_Writer__typ, 0x0d);
	__NEW(cap, Graphics_CaptionDesc);
	cap->len = (int)(end - beg);
	cap->pos = (int)Graphics_T->len + 1;
	cap->do_ = Graphics_CapMethod;
	Texts_OpenReader(&R, Texts_Reader__typ, T, beg);
	Texts_Read(&R, Texts_Reader__typ, &ch);
	GraphicFrames_W.fnt = R.fnt;
	GraphicFrames_W.col = R.col;
	w = 0;
	cap->x = x1 - F->x;
	cap->y = (y1 - F->y) + R.fnt->minY;
	while (beg < end) {
		Display_GetChar(R.fnt->raster, ch, &dx, &x2, &y2, &w1, &h1, &pat);
		w += dx;
		beg += 1;
		Texts_Write(&GraphicFrames_W, Texts_Writer__typ, ch);
		Texts_Read(&R, Texts_Reader__typ, &ch);
	}
	cap->w = w;
	cap->h = GraphicFrames_W.fnt->height;
	cap->col = GraphicFrames_W.col;
	Texts_Append(Graphics_T, GraphicFrames_W.buf);
	Graphics_Add(F->graph, (void*)cap);
	return cap;
}

static void GraphicFrames_SendCaption (Graphics_Caption cap)
{
	Oberon_CopyOverMsg M = {0};
	M.text = Graphics_T;
	M.beg = cap->pos;
	M.end = M.beg + (LONGINT)cap->len;
	Viewers_Broadcast((void*)&M, Oberon_CopyOverMsg__typ);
}

static struct Edit__12 {
	GraphicFrames_Frame *F;
	Graphics_Graph *G;
	struct Edit__12 *lnk;
} *Edit__12_s;

static void NewLine__13 (INTEGER x, INTEGER y, INTEGER w, INTEGER h);

static void NewLine__13 (INTEGER x, INTEGER y, INTEGER w, INTEGER h)
{
	Graphics_Line line = NIL;
	__NEW(line, Graphics_LineDesc);
	line->col = Oberon_CurCol;
	line->x = x - (*Edit__12_s->F)->x;
	line->y = y - (*Edit__12_s->F)->y;
	line->w = w;
	line->h = h;
	line->do_ = Graphics_LineMethod;
	Graphics_Add(*Edit__12_s->G, (void*)line);
}

static void GraphicFrames_Edit (GraphicFrames_Frame F, INTEGER x0, INTEGER y0, SET k0)
{
	Graphics_Object obj = NIL;
	INTEGER x1 = 0, y1 = 0, w = 0, h = 0, t = 0;
	LONGINT beg = 0, end = 0, time = 0;
	SET k1 = 0, k2 = 0;
	GraphicFrames_Location mark = NIL, newmark = NIL;
	Texts_Text T = NIL;
	GraphicFrames_Frame Fd = NIL;
	Graphics_Graph G = NIL;
	GraphicFrames_CtrlMsg CM = {0};
	struct Edit__12 _s;
	_s.F = &F;
	_s.G = &G;
	_s.lnk = Edit__12_s;
	Edit__12_s = &_s;
	k1 = k0;
	G = F->graph;
	if (k0 == 0x02) {
		obj = Graphics_ThisObj(G, x0 - F->x, y0 - F->y);
		if (obj != NIL && !obj->selected) {
			CM.f = F;
			CM.res = 0;
			(*obj->do_->handle)(obj, (void*)&CM, GraphicFrames_CtrlMsg__typ);
			if (CM.res != 0) {
				k0 = 0x0;
			}
		}
	}
	do {
		Input_Mouse(&k2, &x1, &y1);
		k1 = k1 | k2;
		x1 += __MASK(F->x - x1, -4);
		y1 += __MASK(F->y - y1, -4);
		Oberon_DrawCursor(&Oberon_Mouse, Oberon_Cursor__typ, &Oberon_Arrow, Oberon_Marker__typ, x1, y1);
	} while (!(k2 == 0x0));
	Oberon_FadeCursor(&Oberon_Mouse, Oberon_Cursor__typ);
	if (k0 == 0x04) {
		w = __ABS(x1 - x0);
		h = __ABS(y1 - y0);
		if (k1 == 0x04) {
			if (w < 7 && h < 7) {
				if (((x1 - 5 >= F->X && x1 + 5 < F->X1) && y1 - 5 >= F->Y) && y1 + 5 < F->Y1) {
					GraphicFrames_Defocus(F);
					Oberon_PassFocus(Viewers_This(F->X, F->Y));
					F->mark.x = x1;
					F->mark.y = y1;
					F->marked = 1;
					GraphicFrames_FlipMark(x1, y1);
				}
			} else {
				GraphicFrames_Deselect(F);
				if (w < h) {
					if (y1 < y0) {
						y0 = y1;
					}
					NewLine__13(x0, y0, Graphics_width, h);
				} else {
					if (x1 < x0) {
						x0 = x1;
					}
					NewLine__13(x0, y0, w, Graphics_width);
				}
				GraphicFrames_Draw(F);
			}
		} else if (k1 == 0x06) {
			GraphicFrames_Deselect(F);
			Oberon_GetSelection(&T, &beg, &end, &time);
			if (time >= 0) {
				GraphicFrames_DrawObj(F, (void*)GraphicFrames_CaptionCopy(F, x1, y1, T, beg, end));
			}
		} else if (k1 == 0x05) {
			if (F->marked) {
				__NEW(newmark, GraphicFrames_LocDesc);
				newmark->x = x1;
				newmark->y = y1;
				newmark->next = NIL;
				GraphicFrames_FlipMark(x1, y1);
				mark = F->mark.next;
				if (mark == NIL) {
					F->mark.next = newmark;
				} else {
					while (mark->next != NIL) {
						mark = mark->next;
					}
					mark->next = newmark;
				}
			}
		}
	} else if (k0 == 0x02) {
		if (k1 == 0x02) {
			if (x0 != x1 || y0 != y1) {
				Fd = GraphicFrames_This(x1, y1);
				GraphicFrames_Erase(F);
				if (Fd == F) {
					Graphics_Move(G, x1 - x0, y1 - y0);
				} else if (Fd != NIL && Fd->graph == G) {
					Graphics_Move(G, __ASHL(__ASHR(((x1 - Fd->x) - x0) + F->x, 2), 2), __ASHL(__ASHR(((y1 - Fd->y) - y0) + F->y, 2), 2));
				}
				GraphicFrames_Draw(F);
			}
		} else if (k1 == 0x06) {
			Fd = GraphicFrames_This(x1, y1);
			if (Fd != NIL) {
				GraphicFrames_DrawNorm(F);
				if (Fd == F) {
					Graphics_Copy(G, G, x1 - x0, y1 - y0);
				} else {
					GraphicFrames_Deselect(Fd);
					Graphics_Copy(G, Fd->graph, __ASHL(__ASHR(((x1 - Fd->x) - x0) + F->x, 2), 2), __ASHL(__ASHR(((y1 - Fd->y) - y0) + F->y, 2), 2));
				}
				GraphicFrames_Draw(Fd);
			}
		} else if (k1 == 0x03) {
			F->Xg += x1 - x0;
			F->Yg += y1 - y0;
			GraphicFrames_Restore(F);
		}
	} else if (k0 == 0x01) {
		GraphicFrames_newcap = NIL;
		if (__ABS(x0 - x1) < 7 && __ABS(y0 - y1) < 7) {
			if (!__IN(2, k1)) {
				GraphicFrames_Deselect(F);
			}
			obj = Graphics_ThisObj(G, x1 - F->x, y1 - F->y);
			if (obj != NIL) {
				Graphics_SelectObj(G, obj);
				GraphicFrames_DrawObj(F, obj);
				if (k1 == 0x03 && __ISP(obj, Graphics_CaptionDesc, 1)) {
					GraphicFrames_SendCaption(__GUARDP(obj, Graphics_CaptionDesc, 1));
				}
			}
		} else {
			GraphicFrames_Deselect(F);
			if (x1 < x0) {
				t = x0;
				x0 = x1;
				x1 = t;
			}
			if (y1 < y0) {
				t = y0;
				y0 = y1;
				y1 = t;
			}
			Graphics_SelectArea(G, x0 - F->x, y0 - F->y, x1 - F->x, y1 - F->y);
			GraphicFrames_Draw(F);
		}
	}
	Edit__12_s = _s.lnk;
}

static void GraphicFrames_NewCaption (GraphicFrames_Frame F, INTEGER col, Fonts_Font font)
{
	Texts_Write(&GraphicFrames_W, Texts_Writer__typ, 0x0d);
	__NEW(GraphicFrames_newcap, Graphics_CaptionDesc);
	GraphicFrames_newcap->x = F->mark.x - F->x;
	GraphicFrames_newcap->y = (F->mark.y - F->y) + font->minY;
	GraphicFrames_newcap->w = 0;
	GraphicFrames_newcap->h = font->height;
	GraphicFrames_newcap->col = col;
	GraphicFrames_newcap->pos = (int)(Graphics_T->len + 1);
	GraphicFrames_newcap->len = 0;
	GraphicFrames_newcap->do_ = Graphics_CapMethod;
	Graphics_Add(F->graph, (void*)GraphicFrames_newcap);
	GraphicFrames_W.fnt = font;
}

static void GraphicFrames_InsertChar (GraphicFrames_Frame F, CHAR ch)
{
	INTEGER w1 = 0, h1 = 0;
	GraphicFrames_DispMsg DM = {0};
	DM.graph = F->graph;
	Display_GetChar(GraphicFrames_W.fnt->raster, ch, &DM.w, &DM.x1, &DM.y1, &w1, &h1, &DM.pat);
	DM.y1 -= GraphicFrames_W.fnt->minY;
	if (((GraphicFrames_newcap->x + GraphicFrames_newcap->w) + DM.w) + F->x < F->X1) {
		Viewers_Broadcast((void*)&DM, GraphicFrames_DispMsg__typ);
		GraphicFrames_newcap->w += DM.w;
		GraphicFrames_newcap->len += 1;
		Texts_Write(&GraphicFrames_W, Texts_Writer__typ, ch);
	}
	Texts_Append(Graphics_T, GraphicFrames_W.buf);
}

static void GraphicFrames_DeleteChar (GraphicFrames_Frame F)
{
	INTEGER w1 = 0, h1 = 0;
	CHAR ch = 0;
	LONGINT pos = 0;
	GraphicFrames_DispMsg DM = {0};
	Texts_Reader R = {0};
	DM.graph = F->graph;
	if (GraphicFrames_newcap->len > 0) {
		pos = Graphics_T->len;
		Texts_OpenReader(&R, Texts_Reader__typ, Graphics_T, pos - 1);
		Texts_Read(&R, Texts_Reader__typ, &ch);
		if (ch >= ' ') {
			Display_GetChar(R.fnt->raster, ch, &DM.w, &DM.x1, &DM.y1, &w1, &h1, &DM.pat);
			GraphicFrames_newcap->w -= DM.w;
			GraphicFrames_newcap->len -= 1;
			DM.y1 -= R.fnt->minY;
			Viewers_Broadcast((void*)&DM, GraphicFrames_DispMsg__typ);
			Texts_Delete(Graphics_T, pos - 1, pos);
		}
	}
}

static void GraphicFrames_GetSelection (GraphicFrames_Frame F, Texts_Text *text, LONGINT *beg, LONGINT *end, LONGINT *time)
{
	Graphics_Object obj = NIL;
	obj = F->graph->sel;
	if ((obj != NIL && __ISP(obj, Graphics_CaptionDesc, 1)) && F->graph->time >= *time) {
		if (__ISP(obj, Graphics_CaptionDesc, 1)) {
			*beg = (*(Graphics_Caption*)&obj)->pos;
			*end = (*(Graphics_Caption*)&obj)->pos + (*(Graphics_Caption*)&obj)->len;
		} else __WITHCHK;
		*text = Graphics_T;
		*time = F->graph->time;
	}
}

static struct Handle__21 {
	struct Handle__21 *lnk;
} *Handle__21_s;

static void move__22 (GraphicFrames_Frame G, INTEGER dx, INTEGER dy);

static void move__22 (GraphicFrames_Frame G, INTEGER dx, INTEGER dy)
{
	GraphicFrames_UpdateMsg M = {0};
	GraphicFrames_Defocus(G);
	Oberon_FadeCursor(&Oberon_Mouse, Oberon_Cursor__typ);
	M.id = 6;
	M.graph = G->graph;
	Viewers_Broadcast((void*)&M, GraphicFrames_UpdateMsg__typ);
	Graphics_Move(G->graph, dx, dy);
	M.id = 5;
	Viewers_Broadcast((void*)&M, GraphicFrames_UpdateMsg__typ);
}

void GraphicFrames_Handle (Display_Frame G, Display_FrameMsg *M, LONGINT *M__typ)
{
	INTEGER x = 0, y = 0;
	GraphicFrames_DispMsg DM = {0};
	GraphicFrames_DrawMsg dM = {0};
	GraphicFrames_Frame G1 = NIL;
	struct Handle__21 _s;
	_s.lnk = Handle__21_s;
	Handle__21_s = &_s;
	if (__ISP(G, GraphicFrames_FrameDesc, 1)) {
		if (__IS(M__typ, Oberon_InputMsg, 1)) {
			if (__IS(M__typ, Oberon_InputMsg, 1)) {
				Oberon_InputMsg *M__ = (void*)M;
				if ((*M__).id == 1) {
					x = (*M__).X - __MASK((*M__).X - (*(GraphicFrames_Frame*)&G)->x, -4);
					y = (*M__).Y - __MASK((*M__).Y - (*(GraphicFrames_Frame*)&G)->y, -4);
					if ((*M__).keys != 0x0) {
						GraphicFrames_Edit((*(GraphicFrames_Frame*)&G), x, y, (*M__).keys);
					} else {
						Oberon_DrawCursor(&Oberon_Mouse, Oberon_Cursor__typ, &Oberon_Arrow, Oberon_Marker__typ, x, y);
					}
				} else if ((*M__).id == 0) {
					if ((*M__).ch == 0x7f) {
						if (GraphicFrames_newcap != NIL) {
							GraphicFrames_DeleteChar((*(GraphicFrames_Frame*)&G));
						} else {
							Oberon_FadeCursor(&Oberon_Mouse, Oberon_Cursor__typ);
							GraphicFrames_Defocus((*(GraphicFrames_Frame*)&G));
							GraphicFrames_Erase((*(GraphicFrames_Frame*)&G));
							Graphics_Delete((*(GraphicFrames_Frame*)&G)->graph);
						}
					} else if ((*M__).ch == 0xc1) {
						move__22((*(GraphicFrames_Frame*)&G), 0, 1);
					} else if ((*M__).ch == 0xc2) {
						move__22((*(GraphicFrames_Frame*)&G), 0, -1);
					} else if ((*M__).ch == 0xc3) {
						move__22((*(GraphicFrames_Frame*)&G), 1, 0);
					} else if ((*M__).ch == 0xc4) {
						move__22((*(GraphicFrames_Frame*)&G), -1, 0);
					} else if ((*M__).ch >= ' ' && (*M__).ch <= 0x7f || GraphicFrames_ExtendedAscii((*M__).ch)) {
						if (GraphicFrames_newcap != NIL) {
							GraphicFrames_InsertChar((*(GraphicFrames_Frame*)&G), (*M__).ch);
						} else if ((*(GraphicFrames_Frame*)&G)->marked) {
							GraphicFrames_Defocus((*(GraphicFrames_Frame*)&G));
							GraphicFrames_Deselect((*(GraphicFrames_Frame*)&G));
							GraphicFrames_NewCaption((*(GraphicFrames_Frame*)&G), (*M__).col, (*M__).fnt);
							GraphicFrames_InsertChar((*(GraphicFrames_Frame*)&G), (*M__).ch);
						}
					}
				}
			} else __WITHCHK;
		} else if (__IS(M__typ, GraphicFrames_UpdateMsg, 1)) {
			if (__IS(M__typ, GraphicFrames_UpdateMsg, 1)) {
				GraphicFrames_UpdateMsg *M__ = (void*)M;
				if ((*M__).graph == (*(GraphicFrames_Frame*)&G)->graph) {
					dM.f = (*(GraphicFrames_Frame*)&G);
					dM.x = (*(GraphicFrames_Frame*)&G)->x;
					dM.y = (*(GraphicFrames_Frame*)&G)->y;
					dM.col = 0;
					switch ((*M__).id) {
						case 0: 
							GraphicFrames_Restore((*(GraphicFrames_Frame*)&G));
							break;
						case 1: 
							dM.mode = 0;
							(*(*M__).obj->do_->draw)((*M__).obj, (void*)&dM, GraphicFrames_DrawMsg__typ);
							break;
						case 2: 
							dM.mode = 1;
							(*(*M__).obj->do_->draw)((*M__).obj, (void*)&dM, GraphicFrames_DrawMsg__typ);
							break;
						case 3: 
							dM.mode = 3;
							(*(*M__).obj->do_->draw)((*M__).obj, (void*)&dM, GraphicFrames_DrawMsg__typ);
							break;
						case 5: 
							dM.mode = 0;
							Graphics_DrawSel((*(GraphicFrames_Frame*)&G)->graph, (void*)&dM, GraphicFrames_DrawMsg__typ);
							break;
						case 4: 
							dM.mode = 2;
							Graphics_DrawSel((*(GraphicFrames_Frame*)&G)->graph, (void*)&dM, GraphicFrames_DrawMsg__typ);
							break;
						case 6: 
							dM.mode = 3;
							Graphics_DrawSel((*(GraphicFrames_Frame*)&G)->graph, (void*)&dM, GraphicFrames_DrawMsg__typ);
							break;
						default: __CASECHK;
					}
				}
			} else __WITHCHK;
		} else if (__IS(M__typ, GraphicFrames_SelQuery, 1)) {
			if (__IS(M__typ, GraphicFrames_SelQuery, 1)) {
				GraphicFrames_SelQuery *M__ = (void*)M;
				if ((*(GraphicFrames_Frame*)&G)->graph->sel != NIL && (*M__).time < (*(GraphicFrames_Frame*)&G)->graph->time) {
					(*M__).f = (*(GraphicFrames_Frame*)&G);
					(*M__).time = (*(GraphicFrames_Frame*)&G)->graph->time;
				}
			} else __WITHCHK;
		} else if (__IS(M__typ, GraphicFrames_FocusQuery, 1)) {
			if ((*(GraphicFrames_Frame*)&G)->marked) {
				__GUARDR(M, GraphicFrames_FocusQuery, 1).f = (*(GraphicFrames_Frame*)&G);
			}
		} else if (__IS(M__typ, GraphicFrames_PosQuery, 1)) {
			if (__IS(M__typ, GraphicFrames_PosQuery, 1)) {
				GraphicFrames_PosQuery *M__ = (void*)M;
				if ((((*(GraphicFrames_Frame*)&G)->X <= (*M__).x && (*M__).x < (*(GraphicFrames_Frame*)&G)->X1) && (*(GraphicFrames_Frame*)&G)->Y <= (*M__).y) && (*M__).y < (*(GraphicFrames_Frame*)&G)->Y1) {
					(*M__).f = (*(GraphicFrames_Frame*)&G);
				}
			} else __WITHCHK;
		} else if (__IS(M__typ, GraphicFrames_DispMsg, 1)) {
			DM = __GUARDR(M, GraphicFrames_DispMsg, 1);
			x = ((*(GraphicFrames_Frame*)&G)->x + GraphicFrames_newcap->x) + GraphicFrames_newcap->w;
			y = (*(GraphicFrames_Frame*)&G)->y + GraphicFrames_newcap->y;
			if ((((DM.graph == (*(GraphicFrames_Frame*)&G)->graph && x >= (*(GraphicFrames_Frame*)&G)->X) && x + DM.w < (*(GraphicFrames_Frame*)&G)->X1) && y >= (*(GraphicFrames_Frame*)&G)->Y) && y < (*(GraphicFrames_Frame*)&G)->Y1) {
				Display_CopyPattern(Oberon_CurCol, DM.pat, x + DM.x1, y + DM.y1, 2);
				Display_ReplConst(15, x, y, DM.w, GraphicFrames_newcap->h, 2);
			}
		} else if (__IS(M__typ, Oberon_ControlMsg, 1)) {
			if (__IS(M__typ, Oberon_ControlMsg, 1)) {
				Oberon_ControlMsg *M__ = (void*)M;
				if ((*M__).id == 1) {
					Oberon_RemoveMarks((*(GraphicFrames_Frame*)&G)->X, (*(GraphicFrames_Frame*)&G)->Y, (*(GraphicFrames_Frame*)&G)->W, (*(GraphicFrames_Frame*)&G)->H);
					GraphicFrames_Defocus((*(GraphicFrames_Frame*)&G));
					GraphicFrames_DrawNorm((*(GraphicFrames_Frame*)&G));
					Graphics_Deselect((*(GraphicFrames_Frame*)&G)->graph);
				} else if ((*M__).id == 0) {
					GraphicFrames_Defocus((*(GraphicFrames_Frame*)&G));
				}
			} else __WITHCHK;
		} else if (__IS(M__typ, Oberon_SelectionMsg, 1)) {
			if (__IS(M__typ, Oberon_SelectionMsg, 1)) {
				Oberon_SelectionMsg *M__ = (void*)M;
				GraphicFrames_GetSelection((*(GraphicFrames_Frame*)&G), &(*M__).text, &(*M__).beg, &(*M__).end, &(*M__).time);
			} else __WITHCHK;
		} else if (__IS(M__typ, Oberon_CopyMsg, 1)) {
			Oberon_RemoveMarks((*(GraphicFrames_Frame*)&G)->X, (*(GraphicFrames_Frame*)&G)->Y, (*(GraphicFrames_Frame*)&G)->W, (*(GraphicFrames_Frame*)&G)->H);
			GraphicFrames_Defocus((*(GraphicFrames_Frame*)&G));
			__NEW(G1, GraphicFrames_FrameDesc);
			__GUARDEQP(G1, GraphicFrames_FrameDesc) = *(*(GraphicFrames_Frame*)&G);
			__GUARDR(M, Oberon_CopyMsg, 1).F = (Display_Frame)G1;
		} else if (__IS(M__typ, MenuViewers_ModifyMsg, 1)) {
			if (__IS(M__typ, MenuViewers_ModifyMsg, 1)) {
				MenuViewers_ModifyMsg *M__ = (void*)M;
				(*(GraphicFrames_Frame*)&G)->Y = (*M__).Y;
				(*(GraphicFrames_Frame*)&G)->H = (*M__).H;
				GraphicFrames_Restore((*(GraphicFrames_Frame*)&G));
			} else __WITHCHK;
		} else if (__IS(M__typ, Oberon_CopyOverMsg, 1)) {
			if (__IS(M__typ, Oberon_CopyOverMsg, 1)) {
				Oberon_CopyOverMsg *M__ = (void*)M;
				if ((*(GraphicFrames_Frame*)&G)->marked) {
					GraphicFrames_DrawObj((*(GraphicFrames_Frame*)&G), (void*)GraphicFrames_CaptionCopy((*(GraphicFrames_Frame*)&G), (*(GraphicFrames_Frame*)&G)->mark.x, (*(GraphicFrames_Frame*)&G)->mark.y, (*M__).text, (*M__).beg, (*M__).end));
				}
			} else __WITHCHK;
		}
	} else __WITHCHK;
	Handle__21_s = _s.lnk;
}

static void GraphicFrames_DrawLine (Graphics_Object obj, Graphics_Msg *M, LONGINT *M__typ)
{
	INTEGER x = 0, y = 0, w = 0, h = 0, col = 0;
	GraphicFrames_Frame f = NIL;
	if (__IS(M__typ, GraphicFrames_DrawMsg, 1)) {
		GraphicFrames_DrawMsg *M__ = (void*)M;
		x = obj->x + (*M__).x;
		y = obj->y + (*M__).y;
		w = obj->w;
		h = obj->h;
		f = (*M__).f;
		if (((x + w > f->X && x < f->X1) && y + h > f->Y) && y < f->Y1) {
			if ((*M__).col == 0) {
				col = obj->col;
			} else {
				col = (*M__).col;
			}
			if ((*M__).mode == 0 && obj->selected || (*M__).mode == 1) {
				Display_ReplPatternC((void*)f, col, Display_grey2, x, y, w, h, x, y, 0);
			} else if ((*M__).mode == 3) {
				Display_ReplConstC((void*)f, f->col, x, y, w, h, 0);
			} else {
				Display_ReplConstC((void*)f, col, x, y, w, h, 0);
			}
		}
	} else __WITHCHK;
}

static void GraphicFrames_DrawCaption (Graphics_Object obj, Graphics_Msg *M, LONGINT *M__typ)
{
	INTEGER x = 0, y = 0, dx = 0, x0 = 0, x1 = 0, y0 = 0, y1 = 0, w = 0, h = 0, w1 = 0, h1 = 0, col = 0;
	GraphicFrames_Frame f = NIL;
	CHAR ch = 0;
	LONGINT pat = 0;
	Fonts_Font fnt = NIL;
	Texts_Reader R = {0};
	if (__IS(M__typ, GraphicFrames_DrawMsg, 1)) {
		GraphicFrames_DrawMsg *M__ = (void*)M;
		x = obj->x + (*M__).x;
		y = obj->y + (*M__).y;
		w = obj->w;
		h = obj->h;
		f = (*M__).f;
		if (((f->X <= x && x <= f->X1) && f->Y <= y) && y + h <= f->Y1) {
			if (x + w > f->X1) {
				w = f->X1 - x;
			}
			Texts_OpenReader(&R, Texts_Reader__typ, Graphics_T, __GUARDP(obj, Graphics_CaptionDesc, 1)->pos);
			Texts_Read(&R, Texts_Reader__typ, &ch);
			if ((*M__).mode == 0) {
				if (ch >= ' ') {
					if ((*M__).col == 0) {
						col = obj->col;
					} else {
						col = (*M__).col;
					}
					fnt = R.fnt;
					x0 = x;
					y0 = y - fnt->minY;
					for (;;) {
						Display_GetChar(fnt->raster, ch, &dx, &x1, &y1, &w1, &h1, &pat);
						if ((x0 + x1) + w1 <= f->X1) {
							Display_CopyPattern(col, pat, x0 + x1, y0 + y1, 1);
						} else {
							break;
						}
						x0 += dx;
						Texts_Read(&R, Texts_Reader__typ, &ch);
						if (ch < ' ') {
							break;
						}
					}
					if (obj->selected) {
						Display_ReplConst(15, x, y, w, h, 2);
					}
				}
			} else if ((*M__).mode < 3) {
				Display_ReplConst(15, x, y, w, h, 2);
			} else {
				Display_ReplConst(f->col, x, y, w, h, 0);
			}
		}
	} else __WITHCHK;
}

static void GraphicFrames_DrawMacro (Graphics_Object obj, Graphics_Msg *M, LONGINT *M__typ)
{
	INTEGER x = 0, y = 0, w = 0, h = 0;
	GraphicFrames_Frame f = NIL;
	GraphicFrames_DrawMsg M1 = {0};
	if (__IS(M__typ, GraphicFrames_DrawMsg, 1)) {
		GraphicFrames_DrawMsg *M__ = (void*)M;
		x = obj->x + (*M__).x;
		y = obj->y + (*M__).y;
		w = obj->w;
		h = obj->h;
		f = (*M__).f;
		if (((x + w > f->X && x < f->X1) && y + h > f->Y) && y < f->Y1) {
			M1.x = x;
			M1.y = y;
			if ((*M__).mode == 0) {
				M1.f = f;
				M1.col = obj->col;
				M1.mode = 0;
				Graphics_DrawMac(__GUARDP(obj, Graphics_MacroDesc, 1)->mac, (void*)&M1, GraphicFrames_DrawMsg__typ);
				if (obj->selected) {
					Display_ReplConstC((void*)f, 15, x, y, w, h, 2);
				}
			} else if ((*M__).mode < 3) {
				Display_ReplConstC((void*)f, 15, x, y, w, h, 2);
			} else {
				Display_ReplConstC((void*)f, f->col, x, y, w, h, 0);
			}
		}
	} else __WITHCHK;
}

void GraphicFrames_Open (GraphicFrames_Frame G, Graphics_Graph graph, INTEGER X, INTEGER Y, INTEGER col, BOOLEAN ticked)
{
	G->graph = graph;
	G->Xg = X;
	G->Yg = Y;
	G->col = col;
	G->marked = 0;
	G->mark.next = NIL;
	G->ticked = ticked;
	G->handle = GraphicFrames_Handle;
}

static void EnumPtrs(void (*P)(void*))
{
	P(GraphicFrames_newcap);
	__ENUMR(&GraphicFrames_W, Texts_Writer__typ, 36, 1, P);
}

__TDESC(GraphicFrames_LocDesc, 1, 1) = {__TDFLDS("LocDesc", 8), {4, -8}};
__TDESC(GraphicFrames_FrameDesc, 1, 4) = {__TDFLDS("FrameDesc", 48), {0, 4, 20, 44, -20}};
__TDESC(GraphicFrames_DrawMsg, 1, 1) = {__TDFLDS("DrawMsg", 12), {0, -8}};
__TDESC(GraphicFrames_CtrlMsg, 1, 1) = {__TDFLDS("CtrlMsg", 8), {0, -8}};
__TDESC(GraphicFrames_UpdateMsg, 1, 2) = {__TDFLDS("UpdateMsg", 12), {4, 8, -12}};
__TDESC(GraphicFrames_SelQuery, 1, 1) = {__TDFLDS("SelQuery", 8), {0, -8}};
__TDESC(GraphicFrames_FocusQuery, 1, 1) = {__TDFLDS("FocusQuery", 4), {0, -8}};
__TDESC(GraphicFrames_PosQuery, 1, 1) = {__TDFLDS("PosQuery", 8), {0, -8}};
__TDESC(GraphicFrames_DispMsg, 1, 1) = {__TDFLDS("DispMsg", 16), {12, -8}};

export void *GraphicFrames__init(void)
{
	__DEFMOD;
	__IMPORT(Display);
	__IMPORT(Fonts);
	__IMPORT(Graphics);
	__IMPORT(Input);
	__IMPORT(MenuViewers);
	__IMPORT(Oberon);
	__IMPORT(Texts);
	__IMPORT(Viewers);
	__REGMOD(GraphicFrames, EnumPtrs);
	__INITYP(GraphicFrames_LocDesc, GraphicFrames_LocDesc, 0);
	__INITYP(GraphicFrames_FrameDesc, Display_FrameDesc, 1);
	__INITYP(GraphicFrames_DrawMsg, Graphics_Msg, 1);
	__INITYP(GraphicFrames_CtrlMsg, Graphics_Msg, 1);
	__INITYP(GraphicFrames_UpdateMsg, Display_FrameMsg, 1);
	__INITYP(GraphicFrames_SelQuery, Display_FrameMsg, 1);
	__INITYP(GraphicFrames_FocusQuery, Display_FrameMsg, 1);
	__INITYP(GraphicFrames_PosQuery, Display_FrameMsg, 1);
	__INITYP(GraphicFrames_DispMsg, Display_FrameMsg, 1);
/* BEGIN */
	Texts_OpenWriter(&GraphicFrames_W, Texts_Writer__typ);
	Graphics_LineMethod->draw = GraphicFrames_DrawLine;
	Graphics_CapMethod->draw = GraphicFrames_DrawCaption;
	Graphics_MacMethod->draw = GraphicFrames_DrawMacro;
	__ENDMOD;
}
