/* Ofront 1.4 -xtspkaG */
#include "SYSTEM.h"
#include "Display.h"
#include "Input.h"
#include "MathL.h"
#include "Printer.h"

struct GraphicOps__1 {
	INTEGER x1, x2, x3, x4;
};

typedef
	void (*GraphicOps_BlockProc)(Display_Frame, INTEGER, LONGINT, INTEGER, INTEGER, INTEGER, INTEGER, INTEGER, INTEGER, INTEGER);

typedef
	void (*GraphicOps_DotProc)(Display_Frame, INTEGER, INTEGER, INTEGER, INTEGER);


export void (*GraphicOps_Scale)(Display_Frame, LONGINT, LONGINT, INTEGER*, INTEGER*);
export Display_Frame GraphicOps_Screen, GraphicOps_Printer;

static LONGINT *LineParms__35__typ;
export LONGINT *GraphicOps__1__typ;

static void GraphicOps_DBlock (Display_Frame F, INTEGER col, LONGINT pat, INTEGER X, INTEGER Y, INTEGER W, INTEGER H, INTEGER Xp, INTEGER Yp, INTEGER mode);
export void GraphicOps_Ellipse (Display_Frame F, LONGINT x, LONGINT y, LONGINT a, LONGINT b, LONGINT d, LONGINT pat, INTEGER col);
export void GraphicOps_Line (Display_Frame F, LONGINT x1, LONGINT y1, LONGINT x2, LONGINT y2, LONGINT d, LONGINT pat, INTEGER col);
static void GraphicOps_PBlock (Display_Frame F, INTEGER col, LONGINT pat, INTEGER X, INTEGER Y, INTEGER W, INTEGER H, INTEGER Xp, INTEGER Yp, INTEGER mode);
static void GraphicOps_PDot (Display_Frame F, INTEGER col, INTEGER X, INTEGER Y, INTEGER mode);
export void GraphicOps_Quadrangle (Display_Frame F, LONGINT x1, LONGINT y1, LONGINT x2, LONGINT y2, LONGINT x3, LONGINT y3, LONGINT x4, LONGINT y4, LONGINT pat, INTEGER col);
export void GraphicOps_Rect (Display_Frame F, LONGINT x, LONGINT y, LONGINT w, LONGINT h, LONGINT d, LONGINT pat, INTEGER col);
static void GraphicOps_SetUp (Display_Frame F, GraphicOps_DotProc *Dot, GraphicOps_BlockProc *Block, INTEGER col, INTEGER *mode, LONGINT pat);
static void GraphicOps_StdScale (Display_Frame F, LONGINT x, LONGINT y, INTEGER *X, INTEGER *Y);


static void GraphicOps_StdScale (Display_Frame F, LONGINT x, LONGINT y, INTEGER *X, INTEGER *Y)
{
	*X = (int)x;
	*Y = (int)y;
}

static void GraphicOps_PDot (Display_Frame F, INTEGER col, INTEGER X, INTEGER Y, INTEGER mode)
{
	if (((F->X <= X && X < F->X + F->W) && F->Y <= Y) && Y < F->Y + F->H) {
		Printer_ReplConst(X, Y, 1, 1);
	}
}

static void GraphicOps_DBlock (Display_Frame F, INTEGER col, LONGINT pat, INTEGER X, INTEGER Y, INTEGER W, INTEGER H, INTEGER Xp, INTEGER Yp, INTEGER mode)
{
	Display_ReplConstC(F, col, X, Y, W, H, mode);
}

static void GraphicOps_PBlock (Display_Frame F, INTEGER col, LONGINT pat, INTEGER X, INTEGER Y, INTEGER W, INTEGER H, INTEGER Xp, INTEGER Yp, INTEGER mode)
{
	if (X < F->X) {
		W -= F->X - X;
		X = F->X;
	}
	if (X + W > F->X + F->W) {
		W = (F->X + F->W) - X;
	}
	if (Y < F->Y) {
		H -= F->Y - Y;
		Y = F->Y;
	}
	if (Y + H > F->Y + F->H) {
		H = (F->Y + F->H) - Y;
	}
	if (W > 0 && H > 0) {
		Printer_ReplConst(X, Y, W, H);
	}
}

static void GraphicOps_SetUp (Display_Frame F, GraphicOps_DotProc *Dot, GraphicOps_BlockProc *Block, INTEGER col, INTEGER *mode, LONGINT pat)
{
	if (F == GraphicOps_Printer) {
		*Dot = GraphicOps_PDot;
		*Block = GraphicOps_PBlock;
	} else {
		*Dot = Display_DotC;
		if (pat == 0) {
			*Block = GraphicOps_DBlock;
		} else {
			*Block = Display_ReplPatternC;
		}
	}
	if (col == -1) {
		*mode = 2;
	} else {
		*mode = 0;
	}
}

typedef
	struct LineParms__35 {
		INTEGER x, y, d, dx, dy, inx, iny, drawX, drawY;
	} LineParms__35;

static struct Quadrangle__32 {
	struct Quadrangle__32 *lnk;
} *Quadrangle__32_s;

static void InitLineParms__33 (INTEGER x1, INTEGER y1, INTEGER x2, INTEGER y2, LineParms__35 *p, LONGINT *p__typ);
static void LineStep__36 (LineParms__35 *p, LONGINT *p__typ);
static void MinMax__38 (INTEGER x1, INTEGER x2, INTEGER x3, INTEGER x4, INTEGER *min, INTEGER *max);

static void MinMax__38 (INTEGER x1, INTEGER x2, INTEGER x3, INTEGER x4, INTEGER *min, INTEGER *max)
{
	*min = x1;
	*max = x1;
	if (x2 < *min) {
		*min = x2;
	} else if (x2 > *max) {
		*max = x2;
	}
	if (x3 < *min) {
		*min = x3;
	} else if (x3 > *max) {
		*max = x3;
	}
	if (x4 < *min) {
		*min = x4;
	} else if (x4 > *max) {
		*max = x4;
	}
}

static void InitLineParms__33 (INTEGER x1, INTEGER y1, INTEGER x2, INTEGER y2, LineParms__35 *p, LONGINT *p__typ)
{
	(*p).x = x1;
	(*p).dx = x2 - x1;
	if ((*p).dx > 0) {
		(*p).inx = 1;
	} else if ((*p).dx < 0) {
		(*p).inx = -1;
		(*p).dx = -(*p).dx;
	} else {
		(*p).inx = 0;
	}
	(*p).y = y1;
	(*p).dy = y2 - y1;
	if ((*p).dy > 0) {
		(*p).iny = 1;
	} else if ((*p).dy < 0) {
		(*p).iny = -1;
		(*p).dy = -(*p).dy;
	} else {
		(*p).iny = 0;
	}
	(*p).d = (*p).dy - (*p).dx;
	(*p).dx = __ASHL((*p).dx, 1);
	(*p).dy = __ASHL((*p).dy, 1);
}

static void LineStep__36 (LineParms__35 *p, LONGINT *p__typ)
{
	while ((*p).d < 0) {
		(*p).x += (*p).inx;
		(*p).d += (*p).dy;
	}
	(*p).drawX = (*p).x;
	(*p).drawY = __ASHR((*p).iny, 1) + (*p).y;
	(*p).d -= (*p).dx;
	(*p).y += (*p).iny;
}

void GraphicOps_Quadrangle (Display_Frame F, LONGINT x1, LONGINT y1, LONGINT x2, LONGINT y2, LONGINT x3, LONGINT y3, LONGINT x4, LONGINT y4, LONGINT pat, INTEGER col)
{
	LineParms__35 left = {0}, right = {0};
	INTEGER X0 = 0, Y0 = 0, X1 = 0, Y1 = 0, X2 = 0, Y2 = 0, X3 = 0, Y3 = 0, X4 = 0, Y4 = 0, X = 0, Y = 0, RHS2 = 0, RHS3 = 0, Xmin = 0, Xmax = 0, Ymin = 0, Ymax = 0;
	GraphicOps_DotProc Dot = NIL;
	GraphicOps_BlockProc Block = NIL;
	INTEGER mode = 0;
	struct Quadrangle__32 _s;
	_s.lnk = Quadrangle__32_s;
	Quadrangle__32_s = &_s;
	(*GraphicOps_Scale)(F, x1, y1, &X1, &Y1);
	(*GraphicOps_Scale)(F, x2, y2, &X2, &Y2);
	(*GraphicOps_Scale)(F, x3, y3, &X3, &Y3);
	(*GraphicOps_Scale)(F, x4, y4, &X4, &Y4);
	MinMax__38(X1, X2, X3, X4, &Xmin, &Xmax);
	MinMax__38(Y1, Y2, Y3, Y4, &Ymin, &Ymax);
	if (((F->X < Xmax && Xmin < F->X + F->W) && F->Y < Ymax) && Ymin < F->Y + F->H) {
		GraphicOps_SetUp(F, &Dot, &Block, col, &mode, pat);
		(*GraphicOps_Scale)(F, 0, 0, &X0, &Y0);
		if (Y1 > Y2 || Y1 == Y2 && X1 > X2) {
			X = X1;
			X1 = X2;
			X2 = X;
			Y = Y1;
			Y1 = Y2;
			Y2 = Y;
		}
		if (Y2 > Y3 || Y2 == Y3 && X2 > X3) {
			X = X2;
			X2 = X3;
			X3 = X;
			Y = Y2;
			Y2 = Y3;
			Y3 = Y;
		}
		if (Y3 > Y4 || Y3 == Y4 && X3 > X4) {
			X = X3;
			X3 = X4;
			X4 = X;
			Y = Y3;
			Y3 = Y4;
			Y4 = Y;
		}
		if (Y1 > Y2 || Y1 == Y2 && X1 > X2) {
			X = X1;
			X1 = X2;
			X2 = X;
			Y = Y1;
			Y1 = Y2;
			Y2 = Y;
		}
		if (Y2 > Y3 || Y2 == Y3 && X2 > X3) {
			X = X2;
			X2 = X3;
			X3 = X;
			Y = Y2;
			Y2 = Y3;
			Y3 = Y;
		}
		if (Y1 > Y2 || Y1 == Y2 && X1 > X2) {
			X = X1;
			X1 = X2;
			X2 = X;
			Y = Y1;
			Y1 = Y2;
			Y2 = Y;
		}
		if ((LONGINT)(X2 - X1) * (LONGINT)(Y4 - Y1) > (LONGINT)(Y2 - Y1) * (LONGINT)(X4 - X1)) {
			RHS2 = 2;
		} else {
			RHS2 = 0;
		}
		if ((LONGINT)(X3 - X1) * (LONGINT)(Y4 - Y1) > (LONGINT)(Y3 - Y1) * (LONGINT)(X4 - X1)) {
			RHS3 = 1;
		} else {
			RHS3 = 0;
		}
		switch (RHS2 + RHS3) {
			case 0: 
				InitLineParms__33(X1, Y1, X2, Y2, &left, LineParms__35__typ);
				InitLineParms__33(X1, Y1, X4, Y4, &right, LineParms__35__typ);
				break;
			case 1: 
				InitLineParms__33(X1, Y1, X2, Y2, &left, LineParms__35__typ);
				InitLineParms__33(X1, Y1, X3, Y3, &right, LineParms__35__typ);
				break;
			case 2: 
				InitLineParms__33(X1, Y1, X3, Y3, &left, LineParms__35__typ);
				InitLineParms__33(X1, Y1, X2, Y2, &right, LineParms__35__typ);
				break;
			case 3: 
				InitLineParms__33(X1, Y1, X4, Y4, &left, LineParms__35__typ);
				InitLineParms__33(X1, Y1, X2, Y2, &right, LineParms__35__typ);
				break;
			default: __CASECHK;
		}
		while (left.y != Y2) {
			LineStep__36(&left, LineParms__35__typ);
			LineStep__36(&right, LineParms__35__typ);
			(*Block)(F, col, pat, left.drawX, left.drawY, right.drawX - left.drawX, 1, X0, Y0, mode);
		}
		switch (RHS2 + RHS3) {
			case 0: 
				InitLineParms__33(X2, Y2, X3, Y3, &left, LineParms__35__typ);
				break;
			case 1: 
				InitLineParms__33(X2, Y2, X4, Y4, &left, LineParms__35__typ);
				break;
			case 2: 
				InitLineParms__33(X2, Y2, X4, Y4, &right, LineParms__35__typ);
				break;
			case 3: 
				InitLineParms__33(X2, Y2, X3, Y3, &right, LineParms__35__typ);
				break;
			default: __CASECHK;
		}
		while (left.y != Y3) {
			LineStep__36(&left, LineParms__35__typ);
			LineStep__36(&right, LineParms__35__typ);
			(*Block)(F, col, pat, left.drawX, left.drawY, right.drawX - left.drawX, 1, X0, Y0, mode);
		}
		switch (RHS2 + RHS3) {
			case 0: case 2: 
				InitLineParms__33(X3, Y3, X4, Y4, &left, LineParms__35__typ);
				break;
			case 1: case 3: 
				InitLineParms__33(X3, Y3, X4, Y4, &right, LineParms__35__typ);
				break;
			default: __CASECHK;
		}
		while (left.y != Y4) {
			LineStep__36(&left, LineParms__35__typ);
			LineStep__36(&right, LineParms__35__typ);
			(*Block)(F, col, pat, left.drawX, left.drawY, right.drawX - left.drawX, 1, X0, Y0, mode);
		}
	}
	Quadrangle__32_s = _s.lnk;
}

static struct Line__27 {
	LONGINT *pat;
	struct Line__27 *lnk;
} *Line__27_s;

static void HairLine__28 (Display_Frame F, INTEGER X0, INTEGER Y0, INTEGER X1, INTEGER Y1, INTEGER col);

static void HairLine__28 (Display_Frame F, INTEGER X0, INTEGER Y0, INTEGER X1, INTEGER Y1, INTEGER col)
{
	INTEGER x = 0, y = 0, dx = 0, dy = 0, d = 0, inc = 0, L = 0, B = 0, R = 0, T = 0, Xmin = 0, Xmax = 0, Ymin = 0, Ymax = 0;
	GraphicOps_DotProc Dot = NIL;
	GraphicOps_BlockProc Block = NIL;
	INTEGER mode = 0;
	L = F->X;
	B = F->Y;
	R = F->X + F->W;
	T = F->Y + F->H;
	if (X0 < X1) {
		Xmin = X0;
		Xmax = X1;
	} else {
		Xmin = X1;
		Xmax = X0;
	}
	if (Y0 < Y1) {
		Ymin = Y0;
		Ymax = Y1;
	} else {
		Ymin = Y1;
		Ymax = Y0;
	}
	if (((L <= Xmax && Xmin < R) && B <= Ymax) && Ymin < T) {
		GraphicOps_SetUp(F, &Dot, &Block, col, &mode, *Line__27_s->pat);
		if (Xmin == Xmax) {
			(*Block)(F, col, *Line__27_s->pat, Xmin, Ymin, 1, (Ymax - Ymin) + 1, col, 0, 0);
		} else if (Ymin == Ymax) {
			(*Block)(F, col, *Line__27_s->pat, Xmin, Ymin, (Xmax - Xmin) + 1, 1, col, 0, 0);
		} else {
			if (Y1 - Y0 < X0 - X1) {
				x = X0;
				X0 = X1;
				X1 = x;
				y = Y0;
				Y0 = Y1;
				Y1 = y;
			}
			dx = __ASHL(X1 - X0, 1);
			dy = __ASHL(Y1 - Y0, 1);
			x = X0;
			y = Y0;
			inc = 1;
			if (((L <= Xmin && Xmax < R) && B <= Ymin) && Ymax < T) {
				if (dy > dx) {
					d = __ASHR(dy, 1);
					if (dx < 0) {
						inc = -1;
						dx = -dx;
					}
					while (y <= Y1) {
						(*Dot)(F, col, x, y, mode);
						y += 1;
						d -= dx;
						if (d < 0) {
							d += dy;
							x += inc;
						}
					}
				} else {
					d = __ASHR(dx, 1);
					if (dy < 0) {
						inc = -1;
						dy = -dy;
					}
					while (x <= X1) {
						(*Dot)(F, col, x, y, mode);
						x += 1;
						d -= dy;
						if (d < 0) {
							d += dx;
							y += inc;
						}
					}
				}
			} else {
				if (dy > dx) {
					d = __ASHR(dy, 1);
					if (dx < 0) {
						inc = -1;
						dx = -dx;
					}
					while (y <= Y1) {
						if (((L <= x && x < R) && B <= y) && y < T) {
							(*Dot)(F, col, x, y, mode);
						}
						y += 1;
						d -= dx;
						if (d < 0) {
							d += dy;
							x += inc;
						}
					}
				} else {
					d = __ASHR(dx, 1);
					if (dy < 0) {
						inc = -1;
						dy = -dy;
					}
					while (x <= X1) {
						if (((L <= x && x < R) && B <= y) && y < T) {
							(*Dot)(F, col, x, y, mode);
						}
						x += 1;
						d -= dy;
						if (d < 0) {
							d += dx;
							y += inc;
						}
					}
				}
			}
		}
	}
}

void GraphicOps_Line (Display_Frame F, LONGINT x1, LONGINT y1, LONGINT x2, LONGINT y2, LONGINT d, LONGINT pat, INTEGER col)
{
	INTEGER X1 = 0, Y1 = 0, X2 = 0, Y2 = 0;
	LONGREAL dx = 0, dy = 0, c = 0;
	LONGINT u1 = 0, v1 = 0, u2 = 0, v2 = 0, u3 = 0, v3 = 0, u4 = 0, v4 = 0;
	struct Line__27 _s;
	_s.pat = &pat;
	_s.lnk = Line__27_s;
	Line__27_s = &_s;
	if (d <= 0) {
		(*GraphicOps_Scale)(F, x1, y1, &X1, &Y1);
		(*GraphicOps_Scale)(F, x2, y2, &X2, &Y2);
		HairLine__28(F, X1, Y1, X2, Y2, col);
	} else {
		dx = x2 - x1;
		dy = y2 - y1;
		c = (LONGREAL)2 * MathL_sqrt(dx * dx + dy * dy);
		if (c > (LONGREAL)0) {
			c = d / c;
			dx = dx * c;
			dy = dy * c;
			u1 = __ENTIER(x1 - dy);
			v1 = __ENTIER(y1 + dx);
			u2 = __ENTIER(x1 + dy);
			v2 = __ENTIER(y1 - dx);
			u3 = __ENTIER(x2 - dy);
			v3 = __ENTIER(y2 + dx);
			u4 = __ENTIER(x2 + dy);
			v4 = __ENTIER(y2 - dx);
			GraphicOps_Quadrangle(F, u1, v1, u2, v2, u3, v3, u4, v4, pat, col);
		}
	}
	Line__27_s = _s.lnk;
}

void GraphicOps_Rect (Display_Frame F, LONGINT x, LONGINT y, LONGINT w, LONGINT h, LONGINT d, LONGINT pat, INTEGER col)
{
	INTEGER X0 = 0, Y0 = 0, L = 0, B = 0, R = 0, T = 0, Li = 0, Bi = 0, Ri = 0, Ti = 0;
	GraphicOps_DotProc Dot = NIL;
	GraphicOps_BlockProc Block = NIL;
	INTEGER mode = 0;
	(*GraphicOps_Scale)(F, x, y, &L, &B);
	(*GraphicOps_Scale)(F, x + w, y + h, &R, &T);
	if (((F->X < R && L < F->X + F->W) && F->Y < T) && B < F->Y + F->H) {
		GraphicOps_SetUp(F, &Dot, &Block, col, &mode, pat);
		if (d <= 0) {
			Li = L + 1;
			Bi = B + 1;
			Ri = R - 1;
			Ti = T - 1;
			col = __MASK(col, -256);
			X0 = 0;
			Y0 = 0;
		} else {
			(*GraphicOps_Scale)(F, 0, 0, &X0, &Y0);
			(*GraphicOps_Scale)(F, x + d, y + d, &Li, &Bi);
			(*GraphicOps_Scale)(F, (x + w) - d, (y + h) - d, &Ri, &Ti);
		}
		if (Li < Ri && Bi < Ti) {
			(*Block)(F, col, pat, L, B, R - L, Bi - B, X0, Y0, mode);
			(*Block)(F, col, pat, L, Ti, R - L, T - Ti, X0, Y0, mode);
			(*Block)(F, col, pat, L, Bi, Li - L, Ti - Bi, X0, Y0, mode);
			(*Block)(F, col, pat, Ri, Bi, R - Ri, Ti - Bi, X0, Y0, mode);
		} else {
			(*Block)(F, col, pat, L, B, R - L, T - B, X0, Y0, mode);
		}
	}
}

static struct Ellipse__2 {
	LONGINT *pat;
	struct GraphicOps__1 (*line)[4096];
	GraphicOps_DotProc *Dot;
	GraphicOps_BlockProc *Block;
	INTEGER *mode;
	struct Ellipse__2 *lnk;
} *Ellipse__2_s;

static void HairCircle__3 (Display_Frame F, INTEGER X, INTEGER Y, INTEGER R, INTEGER col);
static void HairEllipse__9 (Display_Frame F, INTEGER X, INTEGER Y, INTEGER A, INTEGER B, INTEGER col);
static void Line2__15 (Display_Frame F, INTEGER x, INTEGER y1, INTEGER y2, INTEGER w, INTEGER h, INTEGER col, INTEGER X0, INTEGER Y0);
static void Line2i__17 (INTEGER x, INTEGER y1, INTEGER y2, INTEGER w, INTEGER h);
static void Line2o__19 (INTEGER x, INTEGER y1, INTEGER y2, INTEGER w, INTEGER h);
static void ScanLines__21 (Display_Frame F, INTEGER y, INTEGER h, INTEGER y0, INTEGER col, INTEGER X0, INTEGER Y0);
static void ThickCircle__23 (Display_Frame F, INTEGER X, INTEGER Y, INTEGER R, INTEGER Ri, INTEGER col, INTEGER X0, INTEGER Y0);
static void ThickEllipse__25 (Display_Frame F, INTEGER X, INTEGER Y, INTEGER A, INTEGER B, INTEGER Ai, INTEGER Bi, INTEGER col, INTEGER X0, INTEGER Y0);

static struct HairCircle__3__4 {
	Display_Frame *F;
	INTEGER *R, *col;
	INTEGER *L, *B, *T;
	struct HairCircle__3__4 *lnk;
} *HairCircle__3__4_s;

static void Dot4__5 (INTEGER x1, INTEGER x2, INTEGER y1, INTEGER y2);
static void Dot4c__7 (INTEGER x1, INTEGER x2, INTEGER y1, INTEGER y2);

static void Dot4__5 (INTEGER x1, INTEGER x2, INTEGER y1, INTEGER y2)
{
	(**Ellipse__2_s->Dot)(*HairCircle__3__4_s->F, *HairCircle__3__4_s->col, x1, y1, *Ellipse__2_s->mode);
	(**Ellipse__2_s->Dot)(*HairCircle__3__4_s->F, *HairCircle__3__4_s->col, x1, y2, *Ellipse__2_s->mode);
	(**Ellipse__2_s->Dot)(*HairCircle__3__4_s->F, *HairCircle__3__4_s->col, x2, y1, *Ellipse__2_s->mode);
	(**Ellipse__2_s->Dot)(*HairCircle__3__4_s->F, *HairCircle__3__4_s->col, x2, y2, *Ellipse__2_s->mode);
}

static void Dot4c__7 (INTEGER x1, INTEGER x2, INTEGER y1, INTEGER y2)
{
	if (*HairCircle__3__4_s->L <= x1 && x1 < *HairCircle__3__4_s->R) {
		if (*HairCircle__3__4_s->B <= y1 && y1 < *HairCircle__3__4_s->T) {
			(**Ellipse__2_s->Dot)(*HairCircle__3__4_s->F, *HairCircle__3__4_s->col, x1, y1, *Ellipse__2_s->mode);
		}
		if (*HairCircle__3__4_s->B <= y2 && y2 < *HairCircle__3__4_s->T) {
			(**Ellipse__2_s->Dot)(*HairCircle__3__4_s->F, *HairCircle__3__4_s->col, x1, y2, *Ellipse__2_s->mode);
		}
	}
	if (*HairCircle__3__4_s->L <= x2 && x2 < *HairCircle__3__4_s->R) {
		if (*HairCircle__3__4_s->B <= y1 && y1 < *HairCircle__3__4_s->T) {
			(**Ellipse__2_s->Dot)(*HairCircle__3__4_s->F, *HairCircle__3__4_s->col, x2, y1, *Ellipse__2_s->mode);
		}
		if (*HairCircle__3__4_s->B <= y2 && y2 < *HairCircle__3__4_s->T) {
			(**Ellipse__2_s->Dot)(*HairCircle__3__4_s->F, *HairCircle__3__4_s->col, x2, y2, *Ellipse__2_s->mode);
		}
	}
}

static void HairCircle__3 (Display_Frame F, INTEGER X, INTEGER Y, INTEGER R, INTEGER col)
{
	INTEGER x = 0, y = 0, dx = 0, dy = 0, d = 0, L = 0, B = 0, Rt = 0, T = 0;
	struct HairCircle__3__4 _s;
	_s.F = &F;
	_s.R = &R;
	_s.col = &col;
	_s.L = &L;
	_s.B = &B;
	_s.T = &T;
	_s.lnk = HairCircle__3__4_s;
	HairCircle__3__4_s = &_s;
	L = F->X;
	B = F->Y;
	Rt = F->X + F->W;
	T = F->Y + F->H;
	if (((L < X + R && X - R < Rt) && B < Y + R) && Y - R < T) {
		x = R - 1;
		y = 0;
		dx = __ASHL(x - 1, 3);
		dy = __ASHL(y, 3) + 4;
		d = 3 - __ASHL(R, 2);
		if (((L <= X - R && X + R <= Rt) && B <= Y - R) && Y + R <= T) {
			while (x > y) {
				Dot4__5((X - x) - 1, X + x, (Y - y) - 1, Y + y);
				Dot4__5((X - y) - 1, X + y, (Y - x) - 1, Y + x);
				d += dy;
				dy += 8;
				y += 1;
				if (d >= 0) {
					d -= dx;
					dx -= 8;
					x -= 1;
				}
			}
			if (x == y) {
				Dot4__5((X - x) - 1, X + x, (Y - y) - 1, Y + y);
			}
		} else {
			while (x > y) {
				Dot4c__7((X - x) - 1, X + x, (Y - y) - 1, Y + y);
				Dot4c__7((X - y) - 1, X + y, (Y - x) - 1, Y + x);
				d += dy;
				dy += 8;
				y += 1;
				if (d >= 0) {
					d -= dx;
					dx -= 8;
					x -= 1;
				}
			}
			if (x == y) {
				Dot4c__7((X - x) - 1, X + x, (Y - y) - 1, Y + y);
			}
		}
	}
	HairCircle__3__4_s = _s.lnk;
}

static struct HairEllipse__9__10 {
	Display_Frame *F;
	INTEGER *col;
	INTEGER *L, *Bt, *R, *T;
	struct HairEllipse__9__10 *lnk;
} *HairEllipse__9__10_s;

static void Dot4__11 (INTEGER x1, INTEGER x2, INTEGER y1, INTEGER y2);
static void Dot4c__13 (INTEGER x1, INTEGER x2, INTEGER y1, INTEGER y2);

static void Dot4__11 (INTEGER x1, INTEGER x2, INTEGER y1, INTEGER y2)
{
	(**Ellipse__2_s->Dot)(*HairEllipse__9__10_s->F, *HairEllipse__9__10_s->col, x1, y1, *Ellipse__2_s->mode);
	(**Ellipse__2_s->Dot)(*HairEllipse__9__10_s->F, *HairEllipse__9__10_s->col, x1, y2, *Ellipse__2_s->mode);
	(**Ellipse__2_s->Dot)(*HairEllipse__9__10_s->F, *HairEllipse__9__10_s->col, x2, y1, *Ellipse__2_s->mode);
	(**Ellipse__2_s->Dot)(*HairEllipse__9__10_s->F, *HairEllipse__9__10_s->col, x2, y2, *Ellipse__2_s->mode);
}

static void Dot4c__13 (INTEGER x1, INTEGER x2, INTEGER y1, INTEGER y2)
{
	if (*HairEllipse__9__10_s->L <= x1 && x1 < *HairEllipse__9__10_s->R) {
		if (*HairEllipse__9__10_s->Bt <= y1 && y1 < *HairEllipse__9__10_s->T) {
			(**Ellipse__2_s->Dot)(*HairEllipse__9__10_s->F, *HairEllipse__9__10_s->col, x1, y1, *Ellipse__2_s->mode);
		}
		if (*HairEllipse__9__10_s->Bt <= y2 && y2 < *HairEllipse__9__10_s->T) {
			(**Ellipse__2_s->Dot)(*HairEllipse__9__10_s->F, *HairEllipse__9__10_s->col, x1, y2, *Ellipse__2_s->mode);
		}
	}
	if (*HairEllipse__9__10_s->L <= x2 && x2 < *HairEllipse__9__10_s->R) {
		if (*HairEllipse__9__10_s->Bt <= y1 && y1 < *HairEllipse__9__10_s->T) {
			(**Ellipse__2_s->Dot)(*HairEllipse__9__10_s->F, *HairEllipse__9__10_s->col, x2, y1, *Ellipse__2_s->mode);
		}
		if (*HairEllipse__9__10_s->Bt <= y2 && y2 < *HairEllipse__9__10_s->T) {
			(**Ellipse__2_s->Dot)(*HairEllipse__9__10_s->F, *HairEllipse__9__10_s->col, x2, y2, *Ellipse__2_s->mode);
		}
	}
}

static void HairEllipse__9 (Display_Frame F, INTEGER X, INTEGER Y, INTEGER A, INTEGER B, INTEGER col)
{
	INTEGER x = 0, y = 0, L = 0, Bt = 0, R = 0, T = 0;
	LONGINT d = 0, dx = 0, dy = 0, x2 = 0, y2 = 0, a = 0, a2 = 0, a8 = 0, b = 0, b2 = 0, b8 = 0;
	struct HairEllipse__9__10 _s;
	_s.F = &F;
	_s.col = &col;
	_s.L = &L;
	_s.Bt = &Bt;
	_s.R = &R;
	_s.T = &T;
	_s.lnk = HairEllipse__9__10_s;
	HairEllipse__9__10_s = &_s;
	if (A > 0 && B > 0) {
		L = F->X;
		Bt = F->Y;
		R = F->X + F->W;
		T = F->Y + F->H;
		if (((L < X + A && X - A < R) && Bt < Y + B) && Y - B < T) {
			a = A - 1;
			a2 = a * a;
			a8 = __ASHL(a2, 3);
			b = B - 1;
			b2 = b * b;
			b8 = __ASHL(b2, 3);
			x = A - 1;
			y = 0;
			x2 = a * b2;
			y2 = 0;
			dx = b8 * (a - 1);
			dy = __ASHL(a2, 2);
			d = b2 * (1 - __ASHL(a, 2));
			if (((L <= X - A && X + A <= R) && Bt <= Y - B) && Y + B <= T) {
				while (y2 < x2) {
					Dot4__11((X - x) - 1, X + x, (Y - y) - 1, Y + y);
					d += dy;
					dy += a8;
					y += 1;
					y2 += a2;
					if (d >= 0) {
						d -= dx;
						dx -= b8;
						x -= 1;
						x2 -= b2;
					}
				}
				d += (__ASHL(x2 + y2, 2) - b2) + a2;
				while (x >= 0) {
					Dot4__11((X - x) - 1, X + x, (Y - y) - 1, Y + y);
					d -= dx;
					dx -= b8;
					x -= 1;
					if (d < 0) {
						d += dy;
						dy += a8;
						y += 1;
					}
				}
			} else {
				while (y2 < x2) {
					Dot4c__13((X - x) - 1, X + x, (Y - y) - 1, Y + y);
					d += dy;
					dy += a8;
					y += 1;
					y2 += a2;
					if (d >= 0) {
						d -= dx;
						dx -= b8;
						x -= 1;
						x2 -= b2;
					}
				}
				d += (__ASHL(x2 + y2, 2) - b2) + a2;
				while (x >= 0) {
					Dot4c__13((X - x) - 1, X + x, (Y - y) - 1, Y + y);
					d -= dx;
					dx -= b8;
					x -= 1;
					if (d < 0) {
						d += dy;
						dy += a8;
						y += 1;
					}
				}
			}
		}
	}
	HairEllipse__9__10_s = _s.lnk;
}

static void Line2__15 (Display_Frame F, INTEGER x, INTEGER y1, INTEGER y2, INTEGER w, INTEGER h, INTEGER col, INTEGER X0, INTEGER Y0)
{
	(**Ellipse__2_s->Block)(F, col, *Ellipse__2_s->pat, x, y1, w, h, X0, Y0, *Ellipse__2_s->mode);
	(**Ellipse__2_s->Block)(F, col, *Ellipse__2_s->pat, x, y2, w, h, X0, Y0, *Ellipse__2_s->mode);
}

static void Line2o__19 (INTEGER x, INTEGER y1, INTEGER y2, INTEGER w, INTEGER h)
{
	INTEGER yt = 0;
	yt = y1 + h;
	if (y1 < 0) {
		y1 = 0;
	}
	if (yt > 4096) {
		yt = 4096;
	}
	while (y1 < yt) {
		(*Ellipse__2_s->line)[__X(y1, 4096)].x1 = x;
		(*Ellipse__2_s->line)[__X(y1, 4096)].x2 = x;
		(*Ellipse__2_s->line)[__X(y1, 4096)].x3 = x;
		(*Ellipse__2_s->line)[__X(y1, 4096)].x4 = x + w;
		y1 += 1;
	}
	yt = y2 + h;
	if (y2 < 0) {
		y2 = 0;
	}
	if (yt > 4096) {
		yt = 4096;
	}
	while (y2 < yt) {
		(*Ellipse__2_s->line)[__X(y2, 4096)].x1 = x;
		(*Ellipse__2_s->line)[__X(y2, 4096)].x2 = x;
		(*Ellipse__2_s->line)[__X(y2, 4096)].x3 = x;
		(*Ellipse__2_s->line)[__X(y2, 4096)].x4 = x + w;
		y2 += 1;
	}
}

static void Line2i__17 (INTEGER x, INTEGER y1, INTEGER y2, INTEGER w, INTEGER h)
{
	INTEGER yt = 0;
	yt = y1 + h;
	if (y1 < 0) {
		y1 = 0;
	}
	if (yt > 4096) {
		yt = 4096;
	}
	while (y1 < yt) {
		(*Ellipse__2_s->line)[__X(y1, 4096)].x2 = x;
		(*Ellipse__2_s->line)[__X(y1, 4096)].x3 = x + w;
		y1 += 1;
	}
	yt = y2 + h;
	if (y2 < 0) {
		y2 = 0;
	}
	if (yt > 4096) {
		yt = 4096;
	}
	while (y2 < yt) {
		(*Ellipse__2_s->line)[__X(y2, 4096)].x2 = x;
		(*Ellipse__2_s->line)[__X(y2, 4096)].x3 = x + w;
		y2 += 1;
	}
}

static void ScanLines__21 (Display_Frame F, INTEGER y, INTEGER h, INTEGER y0, INTEGER col, INTEGER X0, INTEGER Y0)
{
	INTEGER yt = 0, x1 = 0, x2 = 0, x3 = 0, x4 = 0;
	yt = y + h;
	if (y < 0) {
		y = 0;
	}
	if (yt > 4096) {
		yt = 4096;
	}
	while (y < yt) {
		x1 = (*Ellipse__2_s->line)[__X(y, 4096)].x1;
		x2 = (*Ellipse__2_s->line)[__X(y, 4096)].x2;
		x3 = (*Ellipse__2_s->line)[__X(y, 4096)].x3;
		x4 = (*Ellipse__2_s->line)[__X(y, 4096)].x4;
		if (x2 < x3) {
			(**Ellipse__2_s->Block)(F, col, *Ellipse__2_s->pat, x1, y0 + y, x2 - x1, 1, X0, Y0, *Ellipse__2_s->mode);
			(**Ellipse__2_s->Block)(F, col, *Ellipse__2_s->pat, x3, y0 + y, x4 - x3, 1, X0, Y0, *Ellipse__2_s->mode);
		} else {
			(**Ellipse__2_s->Block)(F, col, *Ellipse__2_s->pat, x1, y0 + y, x4 - x1, 1, X0, Y0, *Ellipse__2_s->mode);
		}
		y += 1;
	}
}

static void ThickCircle__23 (Display_Frame F, INTEGER X, INTEGER Y, INTEGER R, INTEGER Ri, INTEGER col, INTEGER X0, INTEGER Y0)
{
	INTEGER x = 0, y = 0, d = 0, dx = 0, dy = 0, yb = 0;
	R -= 1;
	Ri -= 1;
	x = R;
	y = 0;
	dx = __ASHL(x - 1, 3);
	dy = __ASHL(y, 3) + 4;
	d = 1 - __ASHL(R, 2);
	yb = 0;
	if (Ri < 0) {
		while (x > y) {
			d += dy;
			dy += 8;
			y += 1;
			if (d >= 0) {
				Line2__15(F, (X - x) - 1, Y - y, Y + yb, __ASHL(x + 1, 1), y - yb, col, X0, Y0);
				Line2__15(F, X - y, (Y - x) - 1, Y + x, __ASHL(y, 1), 1, col, X0, Y0);
				d -= dx;
				dx -= 8;
				x -= 1;
				yb = y;
			}
		}
		if (x == y) {
			y += 1;
			Line2__15(F, (X - x) - 1, Y - y, Y + yb, __ASHL(x + 1, 1), y - yb, col, X0, Y0);
		}
	} else {
		Y -= F->Y;
		while (x > y) {
			d += dy;
			dy += 8;
			y += 1;
			if (d >= 0) {
				Line2o__19((X - x) - 1, Y - y, Y + yb, __ASHL(x + 1, 1), y - yb);
				Line2o__19(X - y, (Y - x) - 1, Y + x, __ASHL(y, 1), 1);
				d -= dx;
				dx -= 8;
				x -= 1;
				yb = y;
			}
		}
		if (x == y) {
			y += 1;
			Line2o__19((X - x) - 1, Y - y, Y + yb, __ASHL(x + 1, 1), y - yb);
		}
		x = Ri;
		y = 0;
		dx = __ASHL(x - 1, 3);
		dy = __ASHL(y, 3) + 4;
		d = 1 - __ASHL(Ri, 2);
		yb = 0;
		while (x > y) {
			d += dy;
			dy += 8;
			y += 1;
			if (d >= 0) {
				Line2i__17((X - x) - 1, Y - y, Y + yb, __ASHL(x + 1, 1), y - yb);
				Line2i__17(X - y, (Y - x) - 1, Y + x, __ASHL(y, 1), 1);
				d -= dx;
				dx -= 8;
				x -= 1;
				yb = y;
			}
		}
		if (x == y) {
			y += 1;
			Line2i__17((X - x) - 1, Y - y, Y + yb, __ASHL(x + 1, 1), y - yb);
		}
		ScanLines__21(F, (Y - R) - 1, __ASHL(R, 1) + 2, F->Y, col, X0, Y0);
	}
}

static void ThickEllipse__25 (Display_Frame F, INTEGER X, INTEGER Y, INTEGER A, INTEGER B, INTEGER Ai, INTEGER Bi, INTEGER col, INTEGER X0, INTEGER Y0)
{
	INTEGER x = 0, y = 0, xb = 0, yb = 0;
	LONGINT d = 0, dx = 0, dy = 0, x2 = 0, y2 = 0, a = 0, a2 = 0, a8 = 0, b = 0, b2 = 0, b8 = 0;
	if (A > 0 && B > 0) {
		A -= 1;
		Ai -= 1;
		B -= 1;
		Bi -= 1;
		a = A;
		a2 = a * a;
		a8 = __ASHL(a2, 3);
		b = B;
		b2 = b * b;
		b8 = __ASHL(b2, 3);
		x = A;
		y = 0;
		x2 = a * b2;
		y2 = 0;
		dx = b8 * (a - 1);
		dy = __ASHL(a2, 2);
		d = b2 * (1 - __ASHL(a, 2));
		yb = 0;
		if (Ai < 0 || Bi < 0) {
			while (y2 < x2) {
				d += dy;
				dy += a8;
				y += 1;
				y2 += a2;
				if (d >= 0) {
					Line2__15(F, (X - x) - 1, Y - y, Y + yb, __ASHL(x + 1, 1), y - yb, col, X0, Y0);
					d -= dx;
					dx -= b8;
					x -= 1;
					x2 -= b2;
					yb = y;
				}
			}
			if (y > yb) {
				Line2__15(F, (X - x) - 1, (Y - y) - 1, Y + yb, __ASHL(x + 1, 1), (y - yb) + 1, col, X0, Y0);
			}
			d += (__ASHL(x2 + y2, 2) - b2) + a2;
			xb = x;
			while (x >= 0) {
				d -= dx;
				dx -= b8;
				x -= 1;
				if (d < 0) {
					Line2__15(F, (X - xb) - 1, (Y - y) - 1, Y + y, __ASHL(xb + 1, 1), 1, col, X0, Y0);
					d += dy;
					dy += a8;
					y += 1;
					xb = x;
				}
			}
			if (x < xb) {
				Line2__15(F, (X - xb) - 1, (Y - y) - 1, Y + y, __ASHL(xb + 1, 1), 1, col, X0, Y0);
			}
		} else {
			Y -= F->Y;
			while (y2 < x2) {
				d += dy;
				dy += a8;
				y += 1;
				y2 += a2;
				if (d >= 0) {
					Line2o__19((X - x) - 1, Y - y, Y + yb, __ASHL(x + 1, 1), y - yb);
					d -= dx;
					dx -= b8;
					x -= 1;
					x2 -= b2;
					yb = y;
				}
			}
			if (y > yb) {
				Line2o__19((X - x) - 1, (Y - y) - 1, Y + yb, __ASHL(x + 1, 1), (y - yb) + 1);
			}
			d += (__ASHL(x2 + y2, 2) - b2) + a2;
			xb = x;
			while (x >= 0) {
				d -= dx;
				dx -= b8;
				x -= 1;
				if (d < 0) {
					Line2o__19((X - xb) - 1, (Y - y) - 1, Y + y, __ASHL(xb + 1, 1), 1);
					d += dy;
					dy += a8;
					y += 1;
					xb = x;
				}
			}
			if (x < xb) {
				Line2o__19((X - xb) - 1, (Y - y) - 1, Y + y, __ASHL(xb + 1, 1), 1);
			}
			a = Ai;
			a2 = a * a;
			a8 = __ASHL(a2, 3);
			b = Bi;
			b2 = b * b;
			b8 = __ASHL(b2, 3);
			x = Ai;
			y = 0;
			x2 = a * b2;
			y2 = 0;
			dx = b8 * (a - 1);
			dy = __ASHL(a2, 2);
			d = b2 * (1 - __ASHL(a, 2));
			yb = 0;
			while (y2 < x2) {
				d += dy;
				dy += a8;
				y += 1;
				y2 += a2;
				if (d >= 0) {
					Line2i__17((X - x) - 1, Y - y, Y + yb, __ASHL(x + 1, 1), y - yb);
					d -= dx;
					dx -= b8;
					x -= 1;
					x2 -= b2;
					yb = y;
				}
			}
			if (y > yb) {
				Line2i__17((X - x) - 1, (Y - y) - 1, Y + yb, __ASHL(x + 1, 1), (y - yb) + 1);
			}
			d += (__ASHL(x2 + y2, 2) - b2) + a2;
			xb = x;
			while (x >= 0) {
				d -= dx;
				dx -= b8;
				x -= 1;
				if (d < 0) {
					Line2i__17((X - xb) - 1, (Y - y) - 1, Y + y, __ASHL(xb + 1, 1), 1);
					d += dy;
					dy += a8;
					y += 1;
					xb = x;
				}
			}
			if (x < xb) {
				Line2i__17((X - xb) - 1, (Y - y) - 1, Y + y, __ASHL(xb + 1, 1), 1);
			}
			ScanLines__21(F, (Y - B) - 1, __ASHL(B, 1) + 2, F->Y, col, X0, Y0);
		}
	}
}

void GraphicOps_Ellipse (Display_Frame F, LONGINT x, LONGINT y, LONGINT a, LONGINT b, LONGINT d, LONGINT pat, INTEGER col)
{
	INTEGER X0 = 0, Y0 = 0, X1 = 0, Y1 = 0, X2 = 0, Y2 = 0, X1i = 0, Y1i = 0, X2i = 0, Y2i = 0, A = 0, B = 0, Ai = 0, Bi = 0;
	struct GraphicOps__1 line[4096];
	GraphicOps_DotProc Dot = NIL;
	GraphicOps_BlockProc Block = NIL;
	INTEGER mode = 0;
	struct Ellipse__2 _s;
	_s.pat = &pat;
	_s.line = (void*)line;
	_s.Dot = &Dot;
	_s.Block = &Block;
	_s.mode = &mode;
	_s.lnk = Ellipse__2_s;
	Ellipse__2_s = &_s;
	(*GraphicOps_Scale)(F, x - a, y - b, &X1, &Y1);
	(*GraphicOps_Scale)(F, x + a, y + b, &X2, &Y2);
	if (((F->X < X2 && X1 < F->X + F->W) && F->Y < Y2) && Y1 < F->Y + F->H) {
		GraphicOps_SetUp(F, &Dot, &Block, col, &mode, pat);
		A = __ASHR(X2 - X1, 1);
		B = __ASHR(Y2 - Y1, 1);
		if (d <= 0) {
			if (A == B) {
				HairCircle__3(F, X1 + A, Y1 + B, A, col);
			} else {
				HairEllipse__9(F, X1 + A, Y1 + B, A, B, col);
			}
		} else {
			(*GraphicOps_Scale)(F, 0, 0, &X0, &Y0);
			(*GraphicOps_Scale)(F, (x - a) + d, (y - b) + d, &X1i, &Y1i);
			(*GraphicOps_Scale)(F, (x + a) - d, (y + b) - d, &X2i, &Y2i);
			Ai = __ASHR(X2i - X1i, 1);
			Bi = __ASHR(Y2i - Y1i, 1);
			if (A == B && Ai == Bi) {
				ThickCircle__23(F, X1 + A, Y1 + B, A, Ai, col, X0, Y0);
			} else {
				ThickEllipse__25(F, X1 + A, Y1 + B, A, B, Ai, Bi, col, X0, Y0);
			}
		}
	}
	Ellipse__2_s = _s.lnk;
}

static void EnumPtrs(void (*P)(void*))
{
	P(GraphicOps_Screen);
	P(GraphicOps_Printer);
}

__TDESC(LineParms__35, 1, 0) = {__TDFLDS("LineParms__35", 18), {-4}};
__TDESC(GraphicOps__1, 1, 0) = {__TDFLDS("", 8), {-4}};

export void *GraphicOps__init(void)
{
	__DEFMOD;
	__IMPORT(Display);
	__IMPORT(Input);
	__IMPORT(MathL);
	__IMPORT(Printer);
	__REGMOD(GraphicOps, EnumPtrs);
	__INITYP(LineParms__35, LineParms__35, 0);
	__INITYP(GraphicOps__1, GraphicOps__1, 0);
/* BEGIN */
	GraphicOps_Scale = GraphicOps_StdScale;
	__NEW(GraphicOps_Screen, Display_FrameDesc);
	GraphicOps_Screen->X = 0;
	GraphicOps_Screen->Y = 0;
	GraphicOps_Screen->W = Display_Width;
	GraphicOps_Screen->H = Display_Height;
	__NEW(GraphicOps_Printer, Display_FrameDesc);
	GraphicOps_Printer->X = 0;
	GraphicOps_Printer->Y = 0;
	GraphicOps_Printer->W = 2200;
	GraphicOps_Printer->H = 3300;
	__ENDMOD;
}
