/* Ofront 1.4 -xtspkaG */
#include "SYSTEM.h"
#include "Display.h"
#include "Files.h"
#include "Fonts.h"
#include "Input.h"
#include "MenuViewers.h"
#include "Oberon.h"
#include "TextFrames.h"
#include "Texts.h"
#include "Viewers.h"

typedef
	struct Hex_CursorCoordDesc *Hex_CursorCoord;

typedef
	struct Hex_CursorCoordDesc {
		INTEGER X, W;
	} Hex_CursorCoordDesc;

typedef
	struct Hex_CursorMsg { /* Display_FrameMsg */
		LONGINT pos;
		Files_File file;
	} Hex_CursorMsg;

typedef
	struct Hex_FrameDesc *Hex_Frame;

typedef
	struct Hex_ModelDesc *Hex_Model;

typedef
	struct Hex_FrameDesc { /* Display_FrameDesc */
		Display_Frame dsc, next;
		INTEGER X, Y, W, H;
		Display_Handler handle;
		BOOLEAN virgin, hasCursor;
		Hex_CursorCoord cursor1, cursor2;
		INTEGER cursorY;
		LONGINT cursorBytePos;
		Hex_Model model;
		LONGINT org, len;
	} Hex_FrameDesc;

typedef
	struct Hex_ModelDesc {
		CHAR name[64];
		Files_File file;
	} Hex_ModelDesc;

typedef
	struct Hex_UpdateMsg { /* Display_FrameMsg */
		INTEGER id;
		Files_File file;
		LONGINT pos;
		CHAR ch;
	} Hex_UpdateMsg;


static Fonts_Font Hex_font;
static INTEGER Hex_fontwidth, Hex_fontheight, Hex_hmin, Hex_hmax, Hex_amin, Hex_amax, Hex_cursorH, Hex_greybar1, Hex_greybar2, Hex_greybar3;
static Hex_CursorCoord Hex_hexcurs, Hex_asccurs;
static CHAR Hex_nextline[16];
static Files_Rider Hex_R;
static Texts_Writer Hex_W;
static INTEGER Hex_res;

export LONGINT *Hex_CursorCoordDesc__typ;
export LONGINT *Hex_ModelDesc__typ;
export LONGINT *Hex_FrameDesc__typ;
export LONGINT *Hex_UpdateMsg__typ;
export LONGINT *Hex_CursorMsg__typ;

static void Hex_AscUpdateByte (Hex_Frame F, CHAR ch);
static CHAR Hex_Cap (CHAR ch);
static void Hex_ChangeFont (CHAR *name, LONGINT name__len, INTEGER *res);
static void Hex_Copy (Hex_Frame src, Hex_Frame dst);
static void Hex_CopyFile (Hex_Frame F);
static CHAR Hex_DecToHex (LONGINT d);
static void Hex_DeleteBang (Hex_Frame F);
static void Hex_Draw (Hex_Frame F, INTEGER Y, INTEGER maxY, LONGINT pos);
static void Hex_DrawClip (Hex_Frame F);
static void Hex_DrawCursor (Hex_Frame F);
static void Hex_DrawFrame (Hex_Frame F);
static void Hex_DrawGreyBars (Hex_Frame F);
static void Hex_Edit (Hex_Frame F, CHAR ch);
static void Hex_FontLogText (CHAR *name, LONGINT name__len, INTEGER res);
export void Hex_GetBool (void);
static void Hex_GetFrame (Hex_Frame *F, CHAR *name, LONGINT name__len);
export void Hex_GetInt (void);
export void Hex_GetLInt (void);
export void Hex_GetLReal (void);
static void Hex_GetLine (Hex_Frame F, INTEGER Y, INTEGER *line);
static void Hex_GetName (CHAR *name, LONGINT name__len);
export void Hex_GetNum (void);
static void Hex_GetOffset (Hex_Frame F, INTEGER X, INTEGER *off);
export void Hex_GetReal (void);
export void Hex_GetSInt (void);
export void Hex_GetSet (void);
static void Hex_GetX (Hex_Frame F, LONGINT pos, INTEGER *hX, INTEGER *aX);
static void Hex_GetY (Hex_Frame F, LONGINT pos, INTEGER *Y);
static void Hex_Handle (Display_Frame F, Display_FrameMsg *M, LONGINT *M__typ);
static INTEGER Hex_HexToDec (CHAR ch);
static void Hex_HexUpdateByte (Hex_Frame F, INTEGER ord);
static void Hex_InitDisplayVars (void);
static void Hex_InvertCursor (Hex_Frame F);
static void Hex_Modify (Hex_Frame F, INTEGER Y, INTEGER H);
export void Hex_Open (void);
static void Hex_OpenFrame (Hex_Frame F, Files_File file, CHAR *name, LONGINT name__len, Display_Handler handle);
export void Hex_OpenText (void);
static void Hex_OpenTextViewer (Hex_Frame F, CHAR *name, LONGINT name__len);
static CHAR Hex_ReadableChar (CHAR ch);
static void Hex_RemoveCursor (Hex_Frame F);
static void Hex_Scroll (Hex_Frame F, INTEGER X, INTEGER Y, SET keysum);
static void Hex_ScrollFrame (Hex_Frame F, LONGINT pos, INTEGER line);
export void Hex_Search (void);
static void Hex_SearchPat (Hex_Frame F, CHAR *pat, LONGINT pat__len, INTEGER len);
static void Hex_SendUpdateMsg (Hex_Frame F);
static void Hex_SetCursor (Hex_Frame F, INTEGER X, INTEGER Y);
export void Hex_SetFont (void);
static void Hex_SetRider (BOOLEAN *done);
static void Hex_ShowAddress (Hex_Frame F, LONGINT pos, INTEGER *X, INTEGER Y);
static void Hex_ShowAscPart (Hex_Frame F, INTEGER max, INTEGER *X, INTEGER Y);
static void Hex_ShowChar (Hex_Frame F, CHAR ch, INTEGER *X, INTEGER Y);
static void Hex_ShowHexPart (Hex_Frame F, INTEGER max, INTEGER *X, INTEGER Y);
static void Hex_ShowLine (Hex_Frame F, INTEGER Y, INTEGER nr, LONGINT adr);
static void Hex_ShowSpaces (Hex_Frame F, INTEGER num, INTEGER *X, INTEGER Y);
export void Hex_Store (void);
static void Hex_StoreFile (Hex_Frame F, CHAR *name, LONGINT name__len);
static void Hex_TrackMouse (Hex_Frame F, INTEGER X, INTEGER Y, SET *keys);
static void Hex_Update (Hex_Frame F, LONGINT pos, CHAR ch);
static void Hex_WriteAddress (LONGINT pos);
static void Hex_WriteAscPart (INTEGER max);
static void Hex_WriteBang (Hex_Frame F);
static void Hex_WriteHexPart (INTEGER max);
static void Hex_WriteLine (INTEGER nr, LONGINT adr);
static void Hex_WriteSpaces (INTEGER num);


static CHAR Hex_Cap (CHAR ch)
{
	switch (ch) {
		case 'a': case 'b': case 'c': case 'd': case 'e': 
		case 'f': case 'g': case 'h': case 'i': case 'j': 
		case 'k': case 'l': case 'm': case 'n': case 'o': 
		case 'p': case 'q': case 'r': case 's': case 't': 
		case 'u': case 'v': case 'w': case 'x': case 'y': 
		case 'z': 
			return __CAP(ch);
			break;
		default: 
			return ch;
			break;
	}
	__RETCHK;
}

static CHAR Hex_DecToHex (LONGINT d)
{
	if (d < 10) {
		d = d + 48;
	} else {
		d = (d + 65) - 10;
	}
	return (CHAR)d;
}

static INTEGER Hex_HexToDec (CHAR ch)
{
	switch (ch) {
		case 'A': case 'B': case 'C': case 'D': case 'E': 
		case 'F': 
			return ((int)ch - 65) + 10;
			break;
		case 'a': case 'b': case 'c': case 'd': case 'e': 
		case 'f': 
			return ((int)ch - 97) + 10;
			break;
		case '0': case '1': case '2': case '3': case '4': 
		case '5': case '6': case '7': case '8': case '9': 
			return (int)ch - 48;
			break;
		default: 
			return -1;
			break;
	}
	__RETCHK;
}

static CHAR Hex_ReadableChar (CHAR ch)
{
	switch ((int)ch) {
		case 32: case 33: case 34: case 35: case 36: 
		case 37: case 38: case 39: case 40: case 41: 
		case 42: case 43: case 44: case 45: case 46: 
		case 47: case 48: case 49: case 50: case 51: 
		case 52: case 53: case 54: case 55: case 56: 
		case 57: case 58: case 59: case 60: case 61: 
		case 62: case 63: case 64: case 65: case 66: 
		case 67: case 68: case 69: case 70: case 71: 
		case 72: case 73: case 74: case 75: case 76: 
		case 77: case 78: case 79: case 80: case 81: 
		case 82: case 83: case 84: case 85: case 86: 
		case 87: case 88: case 89: case 90: case 91: 
		case 92: case 93: case 94: case 95: case 96: 
		case 97: case 98: case 99: case 100: case 101: 
		case 102: case 103: case 104: case 105: case 106: 
		case 107: case 108: case 109: case 110: case 111: 
		case 112: case 113: case 114: case 115: case 116: 
		case 117: case 118: case 119: case 120: case 121: 
		case 122: case 123: case 124: case 125: case 126: 
		case 128: case 129: case 130: case 131: case 132: 
		case 133: case 134: case 135: case 136: case 137: 
		case 138: case 139: case 140: case 141: case 142: 
		case 143: case 144: case 145: case 146: case 147: 
		case 148: case 149: case 155: 
			return ch;
			break;
		default: 
			return '.';
			break;
	}
	__RETCHK;
}

static void Hex_InitDisplayVars (void)
{
	INTEGER dx = 0, x = 0, y = 0, w = 0, h = 0;
	LONGINT p = 0;
	Display_GetChar(Hex_font->raster, '0', &dx, &x, &y, &w, &h, &p);
	Hex_fontwidth = dx;
	Hex_fontheight = Hex_font->height + 1;
	Hex_hmin = 20 + 9 * Hex_fontwidth;
	Hex_hmax = Hex_hmin + 47 * Hex_fontwidth;
	Hex_amin = Hex_hmax + 3 * Hex_fontwidth;
	Hex_amax = Hex_amin + __ASHL(Hex_fontwidth, 4);
	Hex_greybar1 = Hex_hmin + __ASHR((Hex_hmax - Hex_hmin) - Hex_fontwidth, 2);
	Hex_greybar2 = Hex_hmin + __ASHR(Hex_hmax - Hex_hmin, 1);
	Hex_greybar3 = Hex_hmax - __ASHR((Hex_hmax - Hex_hmin) - Hex_fontwidth, 2);
	__NEW(Hex_hexcurs, Hex_CursorCoordDesc);
	Hex_hexcurs->W = __ASHL(Hex_fontwidth, 1) + 3;
	__NEW(Hex_asccurs, Hex_CursorCoordDesc);
	Hex_asccurs->W = Hex_fontwidth;
	Hex_cursorH = Hex_fontheight;
}

static void Hex_GetLine (Hex_Frame F, INTEGER Y, INTEGER *line)
{
	if (Y >= F->Y) {
		*line = __DIV(((F->Y + F->H) - Y) - 3, Hex_fontheight);
		if ((*line + 1) * Hex_fontheight >= F->H - 3) {
			*line -= 1;
		}
		if (*line < 0) {
			*line = 0;
		}
	} else {
		*line = __DIV(F->H - 3, Hex_fontheight) - 1;
	}
}

static void Hex_GetOffset (Hex_Frame F, INTEGER X, INTEGER *off)
{
	if (Hex_hmin <= X - F->X && X - F->X <= Hex_hmax) {
		*off = __DIV(((X - F->X) - Hex_hmin) + __ASHR(Hex_fontwidth, 1), 3 * Hex_fontwidth);
	} else if (Hex_amin <= X - F->X && X - F->X <= Hex_amax) {
		*off = __DIV((X - F->X) - Hex_amin, Hex_fontwidth);
	} else {
		*off = -1;
	}
}

static void Hex_GetX (Hex_Frame F, LONGINT pos, INTEGER *hX, INTEGER *aX)
{
	if (pos < F->len) {
		pos -= F->org;
		pos = __MASK(pos, -16);
		*hX = (F->X + Hex_hmin) + ((int)pos * 3) * Hex_fontwidth;
		*aX = (F->X + Hex_amin) + (int)pos * Hex_fontwidth;
	} else {
		*hX = -1;
		*aX = -1;
	}
}

static void Hex_GetY (Hex_Frame F, LONGINT pos, INTEGER *Y)
{
	if (pos < F->len) {
		pos -= F->org;
		pos = __ASHR(pos, 4);
		*Y = (F->Y + F->H) - ((int)pos + 1) * Hex_fontheight;
	} else {
		*Y = -1;
	}
}

static void Hex_WriteBang (Hex_Frame F)
{
	Texts_Reader R = {0};
	Viewers_Viewer V = NIL;
	Texts_Text T = NIL;
	CHAR ch = 0;
	V = Viewers_This(F->X, F->Y);
	if (__ISP(V, MenuViewers_ViewerDesc, 2) && __ISP(V->dsc, TextFrames_FrameDesc, 1)) {
		T = __GUARDP(V->dsc, TextFrames_FrameDesc, 1)->text;
		if (T->len > 0) {
			Texts_OpenReader(&R, Texts_Reader__typ, T, T->len - 1);
			Texts_Read(&R, Texts_Reader__typ, &ch);
		} else {
			ch = 0x00;
		}
		if (ch != '!') {
			Texts_Write(&Hex_W, Texts_Writer__typ, '!');
			Texts_Append(T, Hex_W.buf);
		}
	}
}

static void Hex_DeleteBang (Hex_Frame F)
{
	Texts_Reader R = {0};
	Viewers_Viewer V = NIL;
	Texts_Text T = NIL;
	CHAR ch = 0;
	V = Viewers_This(F->X, F->Y);
	if (__ISP(V, MenuViewers_ViewerDesc, 2) && __ISP(V->dsc, TextFrames_FrameDesc, 1)) {
		T = __GUARDP(V->dsc, TextFrames_FrameDesc, 1)->text;
		if (T->len > 0) {
			Texts_OpenReader(&R, Texts_Reader__typ, T, T->len - 1);
			Texts_Read(&R, Texts_Reader__typ, &ch);
		} else {
			ch = 0x00;
		}
		if (ch == '!') {
			Texts_Delete(T, T->len - 1, T->len);
		}
	}
}

static void Hex_InvertCursor (Hex_Frame F)
{
	if (((F->X < F->cursor1->X && F->cursor1->X + F->cursor1->W < F->X + F->W) && F->Y < F->cursorY) && F->cursorY + Hex_cursorH <= F->Y + F->H) {
		F->hasCursor = !F->hasCursor;
		Oberon_RemoveMarks(F->X, F->Y, F->W, F->H);
		Display_ReplConst(15, F->cursor1->X, F->cursorY, F->cursor1->W, Hex_cursorH, 2);
		Display_ReplConstC((void*)F, 15, F->cursor2->X, F->cursorY, F->cursor2->W, 2, 2);
	}
}

static void Hex_RemoveCursor (Hex_Frame F)
{
	if (F->hasCursor) {
		Hex_InvertCursor(F);
		F->cursorBytePos = -1;
	}
}

static void Hex_DrawCursor (Hex_Frame F)
{
	Oberon_RemoveMarks(F->X, F->Y, F->W, F->H);
	Display_ReplConstC((void*)F, 15, F->cursor1->X, F->cursorY, F->cursor1->W, Hex_cursorH, 2);
	Display_ReplConstC((void*)F, 15, F->cursor2->X, F->cursorY, F->cursor2->W, 2, 2);
}

static void Hex_SetCursor (Hex_Frame F, INTEGER X, INTEGER Y)
{
	INTEGER offset = 0, line = 0;
	LONGINT pos = 0;
	Hex_GetOffset(F, X, &offset);
	Hex_GetLine(F, Y, &line);
	pos = (__ASHL((LONGINT)line, 4) + (LONGINT)offset) + F->org;
	if (pos < F->len) {
		if (F->cursor1 == Hex_hexcurs) {
			Hex_GetX(F, pos, &F->cursor1->X, &F->cursor2->X);
			F->cursor1->X -= 1;
		} else {
			Hex_GetX(F, pos, &F->cursor2->X, &F->cursor1->X);
			F->cursor2->X -= 1;
		}
		Hex_GetY(F, pos, &F->cursorY);
		F->cursorY -= 3;
		F->cursorBytePos = pos;
		Hex_InvertCursor(F);
	}
}

static void Hex_ShowChar (Hex_Frame F, CHAR ch, INTEGER *X, INTEGER Y)
{
	INTEGER dx = 0, x = 0, y = 0, w = 0, h = 0;
	LONGINT p = 0;
	if (((F->X < *X && *X + Hex_fontwidth < F->X + F->W) && F->Y + 3 < Y) && Y + Hex_fontheight <= F->Y + F->H) {
		Oberon_RemoveMarks(F->X, F->Y, F->W, F->H);
		Display_GetChar(Hex_font->raster, ch, &dx, &x, &y, &w, &h, &p);
		Display_CopyPattern(15, p, *X + x, Y + y, 0);
		*X += dx;
	}
}

static void Hex_ShowSpaces (Hex_Frame F, INTEGER num, INTEGER *X, INTEGER Y)
{
	INTEGER i = 0;
	i = 0;
	while (i < num) {
		Hex_ShowChar(F, ' ', &*X, Y);
		i += 1;
	}
}

static void Hex_ShowAddress (Hex_Frame F, LONGINT pos, INTEGER *X, INTEGER Y)
{
	LONGINT div = 0;
	div = 1048576;
	do {
		Hex_ShowChar(F, Hex_DecToHex(__DIV(pos, div)), &*X, Y);
		pos = __MOD(pos, div);
		div = __ASHR(div, 4);
	} while (!(div == 0));
}

static void Hex_ShowHexPart (Hex_Frame F, INTEGER max, INTEGER *X, INTEGER Y)
{
	INTEGER i = 0;
	i = 0;
	while (i < max) {
		Hex_ShowChar(F, Hex_DecToHex(__ASHR((LONGINT)Hex_nextline[__X(i, 16)], 4)), &*X, Y);
		Hex_ShowChar(F, Hex_DecToHex(__MASK((int)Hex_nextline[__X(i, 16)], -16)), &*X, Y);
		Hex_ShowSpaces(F, 1, &*X, Y);
		i += 1;
	}
	Hex_ShowSpaces(F, (16 - i) * 3, &*X, Y);
}

static void Hex_ShowAscPart (Hex_Frame F, INTEGER max, INTEGER *X, INTEGER Y)
{
	INTEGER i = 0;
	i = 0;
	while (i < max) {
		Hex_ShowChar(F, Hex_ReadableChar(Hex_nextline[__X(i, 16)]), &*X, Y);
		i += 1;
	}
}

static void Hex_ShowLine (Hex_Frame F, INTEGER Y, INTEGER nr, LONGINT adr)
{
	INTEGER X = 0;
	X = F->X + 20;
	Hex_ShowAddress(F, adr, &X, Y);
	Hex_ShowSpaces(F, 3, &X, Y);
	Hex_ShowHexPart(F, nr, &X, Y);
	Hex_ShowSpaces(F, 2, &X, Y);
	Hex_ShowAscPart(F, nr, &X, Y);
}

static void Hex_DrawGreyBars (Hex_Frame F)
{
	INTEGER Y = 0, H = 0, line = 0;
	LONGINT help = 0;
	Hex_GetLine(F, F->Y + 1, &line);
	help = F->len - F->org;
	if ((LONGINT)__ASHL(line + 1, 4) > help) {
		Y = ((F->Y + F->H) - (int)(__ASHR(help - 1, 4) + 1) * Hex_fontheight) - 3;
		H = (int)(__ASHR(help - 1, 4) + 1) * Hex_fontheight;
	} else {
		Y = ((F->Y + F->H) - (line + 1) * Hex_fontheight) - 3;
		H = (line + 1) * Hex_fontheight;
	}
	if (__DIV((F->H - 1) - 3, Hex_fontheight) > 0) {
		Display_ReplPattern(15, Display_grey1, F->X + Hex_greybar1, Y, 1, H, 0);
		Display_ReplPattern(15, Display_grey1, F->X + Hex_greybar2, Y, 1, H, 0);
		Display_ReplPattern(15, Display_grey1, F->X + Hex_greybar3, Y, 1, H, 0);
	}
}

static void Hex_DrawClip (Hex_Frame F)
{
	INTEGER Y = 0;
	Oberon_RemoveMarks(F->X, F->Y, F->W, F->H);
	Display_ReplConst(0, F->X + 1, F->Y, 11, F->H, 0);
	Y = ((F->Y + F->H) - 2) - (int)__DIV((LONGINT)(F->H - 2) * F->org, F->len);
	Display_ReplConst(15, F->X + 1, Y, 8, 2, 0);
}

static void Hex_Draw (Hex_Frame F, INTEGER Y, INTEGER maxY, LONGINT pos)
{
	INTEGER X = 0, rest = 0;
	Y -= Hex_fontheight;
	if (F->len > 0) {
		Files_Set(&Hex_R, Files_Rider__typ, F->model->file, pos);
		Files_ReadBytes(&Hex_R, Files_Rider__typ, (void*)Hex_nextline, 16, 16);
		while (!Hex_R.eof && Y > maxY) {
			Hex_ShowLine(F, Y, 16, Files_Pos(&Hex_R, Files_Rider__typ) - 16);
			Y -= Hex_fontheight;
			Files_ReadBytes(&Hex_R, Files_Rider__typ, (void*)Hex_nextline, 16, 16);
		}
		rest = 16 - (int)Hex_R.res;
		if (Y > maxY && rest > 0) {
			Hex_ShowLine(F, Y, rest, Files_Pos(&Hex_R, Files_Rider__typ) - (LONGINT)rest);
		}
		Hex_DrawClip(F);
	}
}

static void Hex_DrawFrame (Hex_Frame F)
{
	INTEGER line = 0;
	Hex_RemoveCursor(F);
	Oberon_RemoveMarks(F->X, F->Y, F->W, F->H);
	Display_ReplConst(0, F->X, F->Y, F->W, F->H, 0);
	Display_ReplConst(15, F->X + 13, F->Y, 1, F->H, 0);
	Hex_Draw(F, F->Y + F->H, F->Y + 3, F->org);
	Hex_DrawGreyBars(F);
}

static void Hex_AscUpdateByte (Hex_Frame F, CHAR ch)
{
	Files_Set(&Hex_R, Files_Rider__typ, F->model->file, F->cursorBytePos);
	Files_Write(&Hex_R, Files_Rider__typ, ch);
}

static void Hex_HexUpdateByte (Hex_Frame F, INTEGER ord)
{
	CHAR help = 0;
	Files_Set(&Hex_R, Files_Rider__typ, F->model->file, F->cursorBytePos);
	Files_Read(&Hex_R, Files_Rider__typ, (void*)&help);
	help = (CHAR)(__ASHL((int)help, 4) + ord);
	Files_Set(&Hex_R, Files_Rider__typ, F->model->file, F->cursorBytePos);
	Files_Write(&Hex_R, Files_Rider__typ, help);
}

static void Hex_Update (Hex_Frame F, LONGINT pos, CHAR ch)
{
	INTEGER hX = 0, aX = 0, Y = 0;
	Hex_GetX(F, pos, &hX, &aX);
	Hex_GetY(F, pos, &Y);
	Oberon_RemoveMarks(F->X, F->Y, F->W, F->H);
	Display_ReplConstC((void*)F, 0, hX - 1, Y - 3, Hex_hexcurs->W, Hex_cursorH, 0);
	Hex_ShowChar(F, Hex_DecToHex(__ASHR((LONGINT)ch, 4)), &hX, Y);
	Hex_ShowChar(F, Hex_DecToHex(__MASK((int)ch, -16)), &hX, Y);
	Display_ReplConstC((void*)F, 0, aX, Y - 3, Hex_asccurs->W, Hex_cursorH, 0);
	Hex_ShowChar(F, Hex_ReadableChar(ch), &aX, Y);
}

static void Hex_SendUpdateMsg (Hex_Frame F)
{
	Hex_UpdateMsg M = {0};
	CHAR ch = 0;
	Files_Set(&Hex_R, Files_Rider__typ, F->model->file, F->cursorBytePos);
	Files_Read(&Hex_R, Files_Rider__typ, (void*)&ch);
	M.id = 0;
	M.file = F->model->file;
	M.ch = ch;
	M.pos = F->cursorBytePos;
	Viewers_Broadcast((void*)&M, Hex_UpdateMsg__typ);
}

static void Hex_ScrollFrame (Hex_Frame F, LONGINT pos, INTEGER line)
{
	INTEGER H = 0, d = 0, maxline = 0;
	Oberon_RemoveMarks(F->X, F->Y, F->W, F->H);
	Hex_GetLine(F, F->Y + 1, &maxline);
	d = F->H - (maxline + 1) * Hex_fontheight;
	if (F->org < pos && pos <= F->org + (LONGINT)__ASHL(maxline, 4)) {
		Hex_RemoveCursor(F);
		H = (F->H - line * Hex_fontheight) - d;
		F->org = pos;
		Display_CopyBlock((F->X + 13) + 1, (F->Y + d) - 3, (F->W - 13) - 1, H, (F->X + 13) + 1, ((F->Y + F->H) - H) - 3, 0);
		Display_ReplConst(0, (F->X + 13) + 1, F->Y, (F->W - 13) - 1, (F->H - H) - 3, 0);
		Hex_Draw(F, (F->Y + F->H) - H, F->Y + 3, F->org + (LONGINT)__ASHL((maxline - line) + 1, 4));
		Hex_DrawGreyBars(F);
	} else if (pos < F->org && F->org <= pos + (LONGINT)__ASHL(maxline, 4)) {
		Hex_RemoveCursor(F);
		if (__ASHR(F->len, 4) <= (LONGINT)maxline) {
			d = F->H - (int)(__ASHR(F->len, 4) + 1) * Hex_fontheight;
		}
		H = (line + 1) * Hex_fontheight;
		F->org = pos;
		Display_CopyBlock((F->X + 13) + 1, ((F->Y + F->H) - H) - 3, (F->W - 13) - 1, H, (F->X + 13) + 1, (F->Y + d) - 3, 0);
		Display_ReplConst(0, (F->X + 13) + 1, ((F->Y + H) + d) - 3, (F->W - 13) - 1, ((F->H - H) - d) + 3, 0);
		Hex_Draw(F, F->Y + F->H, ((F->Y + H) + d) - 1, F->org);
		Hex_DrawGreyBars(F);
	} else {
		F->org = pos;
		Hex_DrawFrame(F);
	}
}

static struct Scroll__43 {
	Hex_Frame *F;
	INTEGER *line, *Ybar;
	struct Scroll__43 *lnk;
} *Scroll__43_s;

static void Track__44 (INTEGER *X, INTEGER *Y, SET *keysum);
static void Underscore__46 (INTEGER col, INTEGER mode);

static void Underscore__46 (INTEGER col, INTEGER mode)
{
	Oberon_RemoveMarks((*Scroll__43_s->F)->X, (*Scroll__43_s->F)->Y, (*Scroll__43_s->F)->W, (*Scroll__43_s->F)->H);
	Display_ReplConstC((void*)*Scroll__43_s->F, col, (*Scroll__43_s->F)->X + 20, *Scroll__43_s->Ybar - 3, 6 * Hex_fontwidth, 2, mode);
}

static void Track__44 (INTEGER *X, INTEGER *Y, SET *keysum)
{
	SET keys = 0, prim = 0;
	INTEGER Y1 = 0, oldline = 0;
	keys = *keysum;
	prim = *keysum;
	oldline = -1;
	*Scroll__43_s->Ybar = -1;
	while (keys != 0x0) {
		Oberon_DrawCursor(&Oberon_Mouse, Oberon_Cursor__typ, &Oberon_Arrow, Oberon_Marker__typ, *X, *Y);
		Hex_GetLine(*Scroll__43_s->F, *Y, &*Scroll__43_s->line);
		if ((LONGINT)__ASHL(*Scroll__43_s->line, 4) + (*Scroll__43_s->F)->org >= (*Scroll__43_s->F)->len) {
			*Scroll__43_s->line = (int)__ASHR(((*Scroll__43_s->F)->len - (*Scroll__43_s->F)->org) - 1, 4);
		}
		if (*Scroll__43_s->line != oldline) {
			if (!__IN(1, prim)) {
				Underscore__46(0, 0);
			}
			Hex_GetY(*Scroll__43_s->F, (LONGINT)__ASHL(*Scroll__43_s->line, 4) + (*Scroll__43_s->F)->org, &*Scroll__43_s->Ybar);
			if (!__IN(1, prim)) {
				Underscore__46(15, 0);
			}
			oldline = *Scroll__43_s->line;
		}
		Input_Mouse(&keys, &*X, &*Y);
		*keysum = *keysum | keys;
	}
}

static void Hex_Scroll (Hex_Frame F, INTEGER X, INTEGER Y, SET keysum)
{
	LONGINT pos = 0;
	INTEGER line = 0, line1 = 0, Ybar = 0;
	struct Scroll__43 _s;
	_s.F = &F;
	_s.line = &line;
	_s.Ybar = &Ybar;
	_s.lnk = Scroll__43_s;
	Scroll__43_s = &_s;
	pos = F->org;
	if (__IN(0, keysum)) {
		Track__44(&X, &Y, &keysum);
		if (keysum == 0x07) {
			Underscore__46(0, 0);
			Scroll__43_s = _s.lnk;
			return;
		} else {
			Hex_GetLine(F, F->Y + 1, &line1);
			pos = F->org - (LONGINT)__ASHL(line1 - line, 4);
			if (pos < 0) {
				if (__ASHR(F->len, 4) > (LONGINT)line1) {
					line = __ASHR(__ASHL(line1 + 1, 4) - (int)F->org, 4) - 1;
				}
				pos = 0;
			}
			Underscore__46(0, 0);
		}
	} else if (__IN(1, keysum)) {
		Track__44(&X, &Y, &keysum);
		if (keysum == 0x07) {
			Scroll__43_s = _s.lnk;
			return;
		} else if (__IN(0, keysum)) {
			pos = 0;
			if (pos != F->org) {
				F->org = pos;
				Hex_DrawFrame(F);
				Scroll__43_s = _s.lnk;
				return;
			}
		} else if (__IN(2, keysum)) {
			pos = __ASHL(__ASHR(F->len, 4) - 2, 4);
			if (pos < 0) {
				pos = 0;
			}
			if (pos != F->org) {
				F->org = pos;
				Hex_DrawFrame(F);
				Scroll__43_s = _s.lnk;
				return;
			}
		} else {
			pos = __DIV((LONGINT)((F->Y + F->H) - Y) * F->len, (LONGINT)F->H);
			pos = __ASHL(__ASHR(pos, 4), 4);
			line = __ASHR((int)(pos - F->org), 4);
			if (line < 0) {
				Hex_GetLine(F, F->Y + 1, &line1);
				if (__ASHR(F->len, 4) > (LONGINT)line1) {
					line = line1 + line;
				} else {
					line = __ASHR((int)F->len, 4) + line;
				}
			}
		}
	} else if (__IN(2, keysum)) {
		Track__44(&X, &Y, &keysum);
		if (keysum == 0x07) {
			Underscore__46(0, 0);
			Scroll__43_s = _s.lnk;
			return;
		} else {
			pos = (LONGINT)__ASHL(line, 4) + F->org;
			if (pos > F->len) {
				pos = __ASHL(__ASHR(F->len, 4), 4);
			}
			Underscore__46(0, 0);
		}
	}
	if (F->org != pos) {
		Hex_ScrollFrame(F, pos, line);
	}
	Scroll__43_s = _s.lnk;
}

static void Hex_TrackMouse (Hex_Frame F, INTEGER X, INTEGER Y, SET *keys)
{
	INTEGER off = 0, line = 0;
	BOOLEAN track = 0;
	Hex_CursorCoord prim = NIL, sec = NIL;
	if (!F->hasCursor && *keys == 0x04) {
		Oberon_PassFocus(Viewers_This(X, Y));
		track = 1;
	} else if (*keys == 0x04) {
		track = 1;
	} else {
		track = 0;
	}
	while (*keys != 0x0) {
		Input_Mouse(&*keys, &X, &Y);
		if (F->X + Hex_hmin < X && X < F->X + Hex_hmax) {
			prim = Hex_hexcurs;
			sec = Hex_asccurs;
		} else if (F->X + Hex_amin < X && X < F->X + Hex_amax) {
			prim = Hex_asccurs;
			sec = Hex_hexcurs;
		} else {
			Hex_RemoveCursor(F);
			prim = NIL;
			sec = NIL;
		}
		Hex_GetLine(F, Y, &line);
		Hex_GetOffset(F, X, &off);
		if (track) {
			if (prim != NIL && (F->cursor1 != prim || (F->org + (LONGINT)__ASHL(line, 4)) + (LONGINT)off != F->cursorBytePos)) {
				Hex_RemoveCursor(F);
				F->cursor1 = prim;
				F->cursor2 = sec;
				Hex_SetCursor(F, X, Y);
			}
		}
		Oberon_DrawCursor(&Oberon_Mouse, Oberon_Cursor__typ, &Oberon_Arrow, Oberon_Marker__typ, X, Y);
	}
}

static void Hex_CopyFile (Hex_Frame F)
{
	Files_File new = NIL;
	Files_Rider writer = {0};
	CHAR buf[512];
	Files_Set(&Hex_R, Files_Rider__typ, F->model->file, 0);
	new = Files_New(F->model->name, 64);
	Files_Set(&writer, Files_Rider__typ, new, 0);
	Files_ReadBytes(&Hex_R, Files_Rider__typ, (void*)buf, 512, 512);
	while (!Hex_R.eof) {
		Files_WriteBytes(&writer, Files_Rider__typ, (void*)buf, 512, 512);
		Files_ReadBytes(&Hex_R, Files_Rider__typ, (void*)buf, 512, 512);
	}
	Files_WriteBytes(&writer, Files_Rider__typ, (void*)buf, 512, 512 - Hex_R.res);
	F->model->file = new;
}

static void Hex_Edit (Hex_Frame F, CHAR ch)
{
	INTEGER hX = 0, aX = 0, Y = 0;
	if (F->hasCursor) {
		if (ch == 0xc3 && F->cursorBytePos != F->len - 1) {
			Hex_InvertCursor(F);
			F->cursorBytePos += 1;
			Hex_GetX(F, F->cursorBytePos, &hX, &aX);
			Hex_GetY(F, F->cursorBytePos, &Y);
			if (F->cursor1 == Hex_hexcurs) {
				Hex_SetCursor(F, hX, Y);
			} else {
				Hex_SetCursor(F, aX, Y);
			}
		} else if (ch == 0xc4 && F->cursorBytePos != 0) {
			Hex_InvertCursor(F);
			F->cursorBytePos -= 1;
			Hex_GetX(F, F->cursorBytePos, &hX, &aX);
			Hex_GetY(F, F->cursorBytePos, &Y);
			if (F->cursor1 == Hex_hexcurs) {
				Hex_SetCursor(F, hX, Y);
			} else {
				Hex_SetCursor(F, aX, Y);
			}
		} else if (F->cursor1 == Hex_hexcurs) {
			if (Hex_HexToDec(ch) >= 0) {
				if (F->virgin) {
					F->virgin = 0;
					Hex_CopyFile(F);
				}
				Hex_HexUpdateByte(F, Hex_HexToDec(ch));
				Hex_SendUpdateMsg(F);
				Hex_DrawCursor(F);
			}
		} else if (F->cursor1 == Hex_asccurs) {
			if (ch == '.' || Hex_ReadableChar(ch) != '.') {
				if (F->virgin) {
					F->virgin = 0;
					Hex_CopyFile(F);
				}
				Hex_AscUpdateByte(F, ch);
				Hex_SendUpdateMsg(F);
				Hex_DrawCursor(F);
				if (F->cursorBytePos != F->len - 1) {
					Hex_InvertCursor(F);
					F->cursorBytePos += 1;
					Hex_GetX(F, F->cursorBytePos, &hX, &aX);
					Hex_GetY(F, F->cursorBytePos, &Y);
					Hex_SetCursor(F, aX, Y);
				}
			}
		}
	}
}

static void Hex_Copy (Hex_Frame src, Hex_Frame dst)
{
	dst->virgin = src->virgin;
	dst->hasCursor = 0;
	dst->cursor1 = NIL;
	dst->cursor2 = NIL;
	dst->cursorBytePos = -1;
	__NEW(dst->model, Hex_ModelDesc);
	dst->model = src->model;
	dst->org = src->org;
	dst->len = src->len;
	dst->handle = src->handle;
}

static void Hex_Modify (Hex_Frame F, INTEGER Y, INTEGER H)
{
	INTEGER line = 0, dH = 0;
	dH = H - F->H;
	if (dH > 0) {
		Oberon_RemoveMarks(F->X, F->Y, F->W, F->H);
		Hex_GetLine(F, F->Y, &line);
		if (F->Y + F->H != Y + H) {
			Display_CopyBlock(F->X, F->Y, F->W, F->H, F->X, Y + dH, 0);
		}
		F->Y = Y;
		F->H = H;
		Display_ReplConst(0, F->X, F->Y, F->W, dH, 0);
		Display_ReplConst(15, F->X + 13, F->Y, 1, dH, 0);
		Hex_Draw(F, (Y + H) - line * Hex_fontheight, F->Y + 3, F->org + (LONGINT)__ASHL(line, 4));
		Hex_DrawGreyBars(F);
	} else if (dH < 0) {
		Oberon_RemoveMarks(F->X, F->Y, F->W, F->H);
		line = __DIV((H - 1) - 3, Hex_fontheight);
		if ((line + 1) * Hex_fontheight >= H - 3) {
			line -= 1;
		}
		dH = (line + 1) * Hex_fontheight;
		if (F->Y + F->H != Y + H) {
			Display_CopyBlock(F->X, ((F->Y + F->H) - dH) - 3, F->W, dH + 3, F->X, ((Y + H) - dH) - 3, 0);
		}
		F->Y = Y;
		F->H = H;
		if (dH < 0) {
			dH = 0;
		}
		Display_ReplConst(0, (F->X + 13) + 1, F->Y, (F->W - 13) - 1, (H - dH) - 3, 0);
		Hex_DrawClip(F);
		Hex_DrawGreyBars(F);
	}
}

static void Hex_Handle (Display_Frame F, Display_FrameMsg *M, LONGINT *M__typ)
{
	Hex_Frame dest = NIL;
	if (__ISP(F, Hex_FrameDesc, 1)) {
		if (__IS(M__typ, Oberon_InputMsg, 1)) {
			if (__IS(M__typ, Oberon_InputMsg, 1)) {
				Oberon_InputMsg *M__ = (void*)M;
				if ((*M__).id == 1) {
					if ((*M__).X < (*(Hex_Frame*)&F)->X + 13) {
						Oberon_DrawCursor(&Oberon_Mouse, Oberon_Cursor__typ, &Oberon_Arrow, Oberon_Marker__typ, (*M__).X, (*M__).Y);
						Hex_Scroll((*(Hex_Frame*)&F), (*M__).X, (*M__).Y, (*M__).keys);
					} else {
						Oberon_DrawCursor(&Oberon_Mouse, Oberon_Cursor__typ, &Oberon_Arrow, Oberon_Marker__typ, (*M__).X, (*M__).Y);
						Hex_TrackMouse((*(Hex_Frame*)&F), (*M__).X, (*M__).Y, &(*M__).keys);
					}
				} else if ((*M__).id == 0) {
					Hex_Edit((*(Hex_Frame*)&F), (*M__).ch);
				}
			} else __WITHCHK;
		} else if (__IS(M__typ, MenuViewers_ModifyMsg, 1)) {
			if (__IS(M__typ, MenuViewers_ModifyMsg, 1)) {
				MenuViewers_ModifyMsg *M__ = (void*)M;
				Hex_RemoveCursor((*(Hex_Frame*)&F));
				Hex_Modify((*(Hex_Frame*)&F), (*M__).Y, (*M__).H);
			} else __WITHCHK;
		} else if (__IS(M__typ, Oberon_CopyMsg, 1)) {
			if (__IS(M__typ, Oberon_CopyMsg, 1)) {
				Oberon_CopyMsg *M__ = (void*)M;
				if ((*M__).F == NIL) {
					__NEW(dest, Hex_FrameDesc);
					(*M__).F = (Display_Frame)dest;
				}
				Hex_RemoveCursor((*(Hex_Frame*)&F));
				Hex_Copy((*(Hex_Frame*)&F), __GUARDP((*M__).F, Hex_FrameDesc, 1));
			} else __WITHCHK;
		} else if (__IS(M__typ, Hex_UpdateMsg, 1)) {
			if (__IS(M__typ, Hex_UpdateMsg, 1)) {
				Hex_UpdateMsg *M__ = (void*)M;
				if ((*M__).id == 1) {
					Hex_DrawFrame((*(Hex_Frame*)&F));
				} else if ((*M__).id == 0) {
					if ((*M__).file == (*(Hex_Frame*)&F)->model->file) {
						Hex_WriteBang((*(Hex_Frame*)&F));
						Hex_Update((*(Hex_Frame*)&F), (*M__).pos, (*M__).ch);
					}
				}
			} else __WITHCHK;
		} else if (__IS(M__typ, Oberon_ControlMsg, 1)) {
			Hex_RemoveCursor((*(Hex_Frame*)&F));
		} else if (__IS(M__typ, Hex_CursorMsg, 1)) {
			if (__IS(M__typ, Hex_CursorMsg, 1)) {
				Hex_CursorMsg *M__ = (void*)M;
				if ((*(Hex_Frame*)&F)->hasCursor) {
					(*M__).file = (*(Hex_Frame*)&F)->model->file;
					(*M__).pos = (*(Hex_Frame*)&F)->cursorBytePos;
				}
			} else __WITHCHK;
		}
	}
}

static void Hex_WriteSpaces (INTEGER num)
{
	INTEGER i = 0;
	i = 0;
	while (i < num) {
		Texts_Write(&Hex_W, Texts_Writer__typ, ' ');
		i += 1;
	}
}

static void Hex_WriteAddress (LONGINT pos)
{
	LONGINT div = 0;
	div = 1048576;
	do {
		Texts_Write(&Hex_W, Texts_Writer__typ, Hex_DecToHex(__DIV(pos, div)));
		pos = __MOD(pos, div);
		div = __ASHR(div, 4);
	} while (!(div == 0));
}

static void Hex_WriteHexPart (INTEGER max)
{
	INTEGER i = 0;
	i = 0;
	while (i < max) {
		Texts_Write(&Hex_W, Texts_Writer__typ, Hex_DecToHex(__ASHR((LONGINT)Hex_nextline[__X(i, 16)], 4)));
		Texts_Write(&Hex_W, Texts_Writer__typ, Hex_DecToHex(__MASK((int)Hex_nextline[__X(i, 16)], -16)));
		Hex_WriteSpaces(1);
		i += 1;
	}
	Hex_WriteSpaces((16 - i) * 3);
}

static void Hex_WriteAscPart (INTEGER max)
{
	INTEGER i = 0;
	i = 0;
	while (i < max) {
		Texts_Write(&Hex_W, Texts_Writer__typ, Hex_ReadableChar(Hex_nextline[__X(i, 16)]));
		i += 1;
	}
}

static void Hex_WriteLine (INTEGER nr, LONGINT adr)
{
	Hex_WriteAddress(adr);
	Hex_WriteSpaces(3);
	Hex_WriteHexPart(nr);
	Hex_WriteSpaces(2);
	Hex_WriteAscPart(nr);
	Texts_WriteLn(&Hex_W, Texts_Writer__typ);
}

static void Hex_OpenFrame (Hex_Frame F, Files_File file, CHAR *name, LONGINT name__len, Display_Handler handle)
{
	__DUP(name, name__len, CHAR);
	F->virgin = 1;
	F->hasCursor = 0;
	F->cursor1 = NIL;
	F->cursor2 = NIL;
	F->cursorBytePos = -1;
	__NEW(F->model, Hex_ModelDesc);
	F->model->file = file;
	__COPY(name, F->model->name, 64);
	F->org = 0;
	F->len = Files_Length(file);
	F->handle = handle;
	__DEL(name);
}

static void Hex_StoreFile (Hex_Frame F, CHAR *name, LONGINT name__len)
{
	__DUP(name, name__len, CHAR);
	F->virgin = 1;
	Hex_DeleteBang(F);
	__COPY(name, F->model->name, 64);
	Hex_CopyFile(F);
	Files_Register(F->model->file);
	__DEL(name);
}

static void Hex_OpenTextViewer (Hex_Frame F, CHAR *name, LONGINT name__len)
{
	Texts_Text T = NIL;
	INTEGER rest = 0;
	Fonts_Font oldfont = NIL;
	INTEGER x = 0, y = 0;
	MenuViewers_Viewer V = NIL;
	TextFrames_Frame Main = NIL, Menu = NIL;
	Texts_Buffer buf = NIL;
	__DUP(name, name__len, CHAR);
	T = TextFrames_Text((CHAR*)"", (LONGINT)1);
	oldfont = Hex_W.fnt;
	Texts_SetFont(&Hex_W, Texts_Writer__typ, Hex_font);
	Files_Set(&Hex_R, Files_Rider__typ, F->model->file, 0);
	Files_ReadBytes(&Hex_R, Files_Rider__typ, (void*)Hex_nextline, 16, 16);
	while (!Hex_R.eof) {
		Hex_WriteLine(16, Files_Pos(&Hex_R, Files_Rider__typ) - 16);
		Files_ReadBytes(&Hex_R, Files_Rider__typ, (void*)Hex_nextline, 16, 16);
	}
	rest = 16 - (int)Hex_R.res;
	if (rest > 0) {
		Hex_WriteLine(rest, Files_Pos(&Hex_R, Files_Rider__typ) - (LONGINT)rest);
	}
	Texts_Append(T, Hex_W.buf);
	Oberon_AllocateUserViewer(Oberon_Mouse.X, &x, &y);
	Main = TextFrames_NewText(T, 0);
	if (Files_Old((CHAR*)"Edit.Menu.Text", (LONGINT)15) == NIL) {
		Menu = TextFrames_NewMenu(name, name__len, (CHAR*)"System.Close  System.Copy  System.Grow  Edit.Search  Edit.Replace  Edit.Parcs  Edit.Store ", (LONGINT)91);
	} else {
		Menu = TextFrames_NewMenu(name, name__len, (CHAR*)"", (LONGINT)1);
		__NEW(T, Texts_TextDesc);
		Texts_Open(T, (CHAR*)"Edit.Menu.Text", (LONGINT)15);
		__NEW(buf, Texts_BufDesc);
		Texts_OpenBuf(buf);
		Texts_Save(T, 0, T->len, buf);
		Texts_Append(Menu->text, buf);
	}
	V = MenuViewers_New((void*)Menu, (void*)Main, TextFrames_menuH, x, y);
	Texts_SetFont(&Hex_W, Texts_Writer__typ, oldfont);
	__DEL(name);
}

static void Hex_ChangeFont (CHAR *name, LONGINT name__len, INTEGER *res)
{
	Fonts_Font newfont = NIL;
	Hex_UpdateMsg M = {0};
	INTEGER dx1 = 0, dx2 = 0, x = 0, y = 0, w = 0, h = 0;
	LONGINT p = 0;
	__DUP(name, name__len, CHAR);
	newfont = Fonts_This(name, name__len);
	if (__STRCMP(name, Fonts_Default->name) != 0) {
		if (Fonts_Default == newfont || newfont == NIL) {
			*res = 1;
			__DEL(name);
			return;
		}
	}
	Display_GetChar(newfont->raster, 'W', &dx1, &x, &y, &w, &h, &p);
	Display_GetChar(newfont->raster, 'i', &dx2, &x, &y, &w, &h, &p);
	if (dx1 != dx2) {
		*res = 2;
	} else {
		*res = 0;
		Hex_font = newfont;
		Hex_InitDisplayVars();
		M.id = 1;
		Viewers_Broadcast((void*)&M, Hex_UpdateMsg__typ);
	}
	__DEL(name);
}

static void Hex_SearchPat (Hex_Frame F, CHAR *pat, LONGINT pat__len, INTEGER len)
{
	LONGINT org = 0, pos = 0, cursorpos = 0;
	CHAR ch = 0;
	INTEGER patpos = 0, hX = 0, aX = 0, Y = 0;
	__DUP(pat, pat__len, CHAR);
	if (F->hasCursor) {
		pos = F->cursorBytePos;
	} else {
		pos = 0;
	}
	do {
		Files_Set(&Hex_R, Files_Rider__typ, F->model->file, pos);
		Files_Read(&Hex_R, Files_Rider__typ, (void*)&ch);
		while (!Hex_R.eof && ch != pat[0]) {
			Files_Read(&Hex_R, Files_Rider__typ, (void*)&ch);
		}
		if (ch == pat[0]) {
			pos = Files_Pos(&Hex_R, Files_Rider__typ);
			Files_Read(&Hex_R, Files_Rider__typ, (void*)&ch);
			patpos = 1;
			while (patpos < len && ch == pat[__X(patpos, pat__len)]) {
				Files_Read(&Hex_R, Files_Rider__typ, (void*)&ch);
				patpos += 1;
			}
			if (patpos == len) {
				if (!F->hasCursor) {
					Oberon_PassFocus(Viewers_This(F->X, F->Y));
				}
				cursorpos = (pos + (LONGINT)len) - 1;
				org = __ASHL(__ASHR(cursorpos, 4) - 1, 4);
				if (org < 0) {
					org = 0;
				}
				F->org = org;
				Hex_DrawFrame(F);
				F->cursor1 = Hex_asccurs;
				F->cursor2 = Hex_hexcurs;
				Hex_GetX(F, cursorpos, &hX, &aX);
				Hex_GetY(F, cursorpos, &Y);
				Hex_SetCursor(F, aX, Y);
				__DEL(pat);
				return;
			}
		}
	} while (!Hex_R.eof);
	Hex_RemoveCursor(F);
	__DEL(pat);
}

static void Hex_GetFrame (Hex_Frame *F, CHAR *name, LONGINT name__len)
{
	Oberon_ParList par = NIL;
	Viewers_Viewer V = NIL;
	Texts_Scanner S = {0};
	par = Oberon_Par;
	if (par->frame == par->vwr->dsc) {
		V = par->vwr;
	} else {
		V = Oberon_MarkedViewer();
	}
	Texts_OpenScanner(&S, Texts_Scanner__typ, __GUARDP(V->dsc, TextFrames_FrameDesc, 1)->text, 0);
	Texts_Scan(&S, Texts_Scanner__typ);
	if ((__IN(S.class, 0x06) && V->dsc != NIL) && __ISP(V->dsc->next, Hex_FrameDesc, 1)) {
		*F = __GUARDP(V->dsc->next, Hex_FrameDesc, 1);
		__COPY(S.s, name, name__len);
	} else {
		*F = NIL;
	}
}

static void Hex_GetName (CHAR *name, LONGINT name__len)
{
	Texts_Text T = NIL;
	Texts_Scanner S = {0};
	LONGINT beg = 0, end = 0, time = 0;
	Texts_OpenScanner(&S, Texts_Scanner__typ, Oberon_Par->text, Oberon_Par->pos);
	Texts_Scan(&S, Texts_Scanner__typ);
	if (S.class == 6 && S.c == '^') {
		Oberon_GetSelection(&T, &beg, &end, &time);
		if (time >= 0) {
			Texts_OpenScanner(&S, Texts_Scanner__typ, T, beg);
			Texts_Scan(&S, Texts_Scanner__typ);
		}
	}
	if (__IN(S.class, 0x06)) {
		__COPY(S.s, name, name__len);
	} else {
		name[0] = 0x00;
	}
}

static void Hex_FontLogText (CHAR *name, LONGINT name__len, INTEGER res)
{
	__DUP(name, name__len, CHAR);
	Texts_WriteString(&Hex_W, Texts_Writer__typ, name, name__len);
	if (res == 1) {
		Texts_WriteString(&Hex_W, Texts_Writer__typ, (CHAR*)" not found", (LONGINT)11);
	} else if (res == 2) {
		Texts_WriteString(&Hex_W, Texts_Writer__typ, (CHAR*)" is not a fixed-width font", (LONGINT)27);
	}
	Texts_WriteLn(&Hex_W, Texts_Writer__typ);
	Texts_Append(Oberon_Log, Hex_W.buf);
	__DEL(name);
}

static void Hex_SetRider (BOOLEAN *done)
{
	Hex_CursorMsg M = {0};
	M.pos = -1;
	Viewers_Broadcast((void*)&M, Hex_CursorMsg__typ);
	if (M.pos >= 0) {
		Files_Set(&Hex_R, Files_Rider__typ, M.file, M.pos);
		*done = 1;
	} else {
		*done = 0;
	}
}

void Hex_Open (void)
{
	Hex_Frame F = NIL;
	TextFrames_Frame M = NIL;
	Viewers_Viewer V = NIL;
	Texts_Text T = NIL;
	Texts_Buffer buf = NIL;
	Files_File File = NIL;
	INTEGER X = 0, Y = 0;
	CHAR name[64];
	INTEGER res = 0;
	Hex_GetName((void*)name, 64);
	if (name[0] != 0x00) {
		File = Files_Old(name, 64);
		if (File != NIL) {
			__NEW(F, Hex_FrameDesc);
			Hex_OpenFrame(F, File, name, 64, Hex_Handle);
			if (Files_Old((CHAR*)"Hex.Menu.Text", (LONGINT)14) == NIL) {
				M = TextFrames_NewMenu(name, 64, (CHAR*)"System.Close  System.Copy  System.Grow  Hex.Search  Hex.OpenText  Hex.Store ", (LONGINT)77);
			} else {
				M = TextFrames_NewMenu(name, 64, (CHAR*)"", (LONGINT)1);
				__NEW(T, Texts_TextDesc);
				Texts_Open(T, (CHAR*)"Hex.Menu.Text", (LONGINT)14);
				__NEW(buf, Texts_BufDesc);
				Texts_OpenBuf(buf);
				Texts_Save(T, 0, T->len, buf);
				Texts_Append(M->text, buf);
			}
			Oberon_AllocateUserViewer(Oberon_Mouse.X, &X, &Y);
			V = (Viewers_Viewer)MenuViewers_New((void*)M, (void*)F, TextFrames_menuH, X, Y);
		} else {
			Texts_WriteString(&Hex_W, Texts_Writer__typ, name, 64);
			Texts_WriteString(&Hex_W, Texts_Writer__typ, (CHAR*)" not found", (LONGINT)11);
			Texts_WriteLn(&Hex_W, Texts_Writer__typ);
			Texts_Append(Oberon_Log, Hex_W.buf);
		}
	}
}

static struct Store__61 {
	struct Store__61 *lnk;
} *Store__61_s;

static void Backup__62 (CHAR *name, LONGINT name__len);

static void Backup__62 (CHAR *name, LONGINT name__len)
{
	INTEGER res = 0, i = 0;
	CHAR bak[64];
	i = 0;
	while (name[__X(i, name__len)] != 0x00) {
		bak[__X(i, 64)] = name[__X(i, name__len)];
		i += 1;
	}
	bak[__X(i, 64)] = '.';
	bak[__X(i + 1, 64)] = 'B';
	bak[__X(i + 2, 64)] = 'a';
	bak[__X(i + 3, 64)] = 'k';
	bak[__X(i + 4, 64)] = 0x00;
	Files_Rename(name, name__len, bak, 64, &res);
}

void Hex_Store (void)
{
	Hex_Frame F = NIL;
	CHAR name[64];
	struct Store__61 _s;
	_s.lnk = Store__61_s;
	Store__61_s = &_s;
	Hex_GetFrame(&F, (void*)name, 64);
	if (F != NIL) {
		Texts_WriteString(&Hex_W, Texts_Writer__typ, (CHAR*)"Hex.Store ", (LONGINT)11);
		Texts_Append(Oberon_Log, Hex_W.buf);
		Backup__62((void*)name, 64);
		Hex_StoreFile(F, name, 64);
		Texts_WriteString(&Hex_W, Texts_Writer__typ, name, 64);
		Texts_WriteLn(&Hex_W, Texts_Writer__typ);
		Texts_Append(Oberon_Log, Hex_W.buf);
	}
	Store__61_s = _s.lnk;
}

static struct OpenText__37 {
	struct OpenText__37 *lnk;
} *OpenText__37_s;

static void NewName__38 (CHAR *name, LONGINT name__len);

static void NewName__38 (CHAR *name, LONGINT name__len)
{
	INTEGER i = 0;
	i = 0;
	while (name[__X(i, name__len)] != 0x00) {
		i += 1;
	}
	name[__X(i, name__len)] = '.';
	name[__X(i + 1, name__len)] = 'T';
	name[__X(i + 2, name__len)] = 'e';
	name[__X(i + 3, name__len)] = 'x';
	name[__X(i + 4, name__len)] = 't';
	name[__X(i + 5, name__len)] = 0x00;
}

void Hex_OpenText (void)
{
	Hex_Frame F = NIL;
	CHAR name[64];
	struct OpenText__37 _s;
	_s.lnk = OpenText__37_s;
	OpenText__37_s = &_s;
	Hex_GetFrame(&F, (void*)name, 64);
	if (F != NIL) {
		NewName__38((void*)name, 64);
		Hex_OpenTextViewer(F, name, 64);
	}
	OpenText__37_s = _s.lnk;
}

void Hex_SetFont (void)
{
	INTEGER res = 0;
	CHAR name[32];
	Hex_GetName((void*)name, 32);
	if (name[0] != 0x00) {
		Hex_ChangeFont(name, 32, &res);
		if (res != 0) {
			Hex_FontLogText(name, 32, res);
		}
	}
}

void Hex_GetSInt (void)
{
	CHAR x = 0;
	BOOLEAN done = 0;
	Hex_SetRider(&done);
	if (done) {
		Files_Read(&Hex_R, Files_Rider__typ, (void*)&x);
		Texts_WriteString(&Hex_W, Texts_Writer__typ, (CHAR*)"SHORTINT", (LONGINT)9);
		Texts_Write(&Hex_W, Texts_Writer__typ, 0x09);
		Texts_WriteInt(&Hex_W, Texts_Writer__typ, (int)x, 0);
		Texts_WriteLn(&Hex_W, Texts_Writer__typ);
		Texts_Append(Oberon_Log, Hex_W.buf);
	}
}

void Hex_GetInt (void)
{
	INTEGER x = 0;
	BOOLEAN done = 0;
	Hex_SetRider(&done);
	if (done) {
		Files_ReadInt(&Hex_R, Files_Rider__typ, &x);
		Texts_WriteString(&Hex_W, Texts_Writer__typ, (CHAR*)"INTEGER", (LONGINT)8);
		Texts_Write(&Hex_W, Texts_Writer__typ, 0x09);
		Texts_WriteInt(&Hex_W, Texts_Writer__typ, x, 0);
		Texts_WriteLn(&Hex_W, Texts_Writer__typ);
		Texts_Append(Oberon_Log, Hex_W.buf);
	}
}

void Hex_GetLInt (void)
{
	LONGINT x = 0;
	BOOLEAN done = 0;
	Hex_SetRider(&done);
	if (done) {
		Files_ReadLInt(&Hex_R, Files_Rider__typ, &x);
		Texts_WriteString(&Hex_W, Texts_Writer__typ, (CHAR*)"LONGINT", (LONGINT)8);
		Texts_Write(&Hex_W, Texts_Writer__typ, 0x09);
		Texts_WriteInt(&Hex_W, Texts_Writer__typ, x, 0);
		Texts_WriteLn(&Hex_W, Texts_Writer__typ);
		Texts_Append(Oberon_Log, Hex_W.buf);
	}
}

void Hex_GetReal (void)
{
	REAL x = 0;
	BOOLEAN done = 0;
	Hex_SetRider(&done);
	if (done) {
		Files_ReadReal(&Hex_R, Files_Rider__typ, &x);
		Texts_WriteString(&Hex_W, Texts_Writer__typ, (CHAR*)"REAL", (LONGINT)5);
		Texts_Write(&Hex_W, Texts_Writer__typ, 0x09);
		Texts_WriteReal(&Hex_W, Texts_Writer__typ, x, 20);
		Texts_WriteLn(&Hex_W, Texts_Writer__typ);
		Texts_Append(Oberon_Log, Hex_W.buf);
	}
}

void Hex_GetLReal (void)
{
	LONGREAL x = 0;
	BOOLEAN done = 0;
	Hex_SetRider(&done);
	if (done) {
		Files_ReadLReal(&Hex_R, Files_Rider__typ, &x);
		Texts_WriteString(&Hex_W, Texts_Writer__typ, (CHAR*)"LONGREAL", (LONGINT)9);
		Texts_Write(&Hex_W, Texts_Writer__typ, 0x09);
		Texts_WriteLongReal(&Hex_W, Texts_Writer__typ, x, 20);
		Texts_WriteLn(&Hex_W, Texts_Writer__typ);
		Texts_Append(Oberon_Log, Hex_W.buf);
	}
}

void Hex_GetNum (void)
{
	LONGINT x = 0, n = 0;
	BOOLEAN done = 0;
	Hex_SetRider(&done);
	if (done) {
		n = Files_Pos(&Hex_R, Files_Rider__typ);
		Files_ReadNum(&Hex_R, Files_Rider__typ, &x);
		n = Files_Pos(&Hex_R, Files_Rider__typ) - n;
		Texts_WriteString(&Hex_W, Texts_Writer__typ, (CHAR*)"Number (", (LONGINT)9);
		Texts_WriteInt(&Hex_W, Texts_Writer__typ, n, 0);
		if (n > 1) {
			Texts_WriteString(&Hex_W, Texts_Writer__typ, (CHAR*)" Bytes)", (LONGINT)8);
		} else {
			Texts_WriteString(&Hex_W, Texts_Writer__typ, (CHAR*)" Byte)", (LONGINT)7);
		}
		Texts_Write(&Hex_W, Texts_Writer__typ, 0x09);
		Texts_WriteInt(&Hex_W, Texts_Writer__typ, x, 0);
		Texts_WriteLn(&Hex_W, Texts_Writer__typ);
		Texts_Append(Oberon_Log, Hex_W.buf);
	}
}

void Hex_GetSet (void)
{
	SET x = 0;
	BOOLEAN done = 0;
	SHORTINT i = 0, last = 0;
	Hex_SetRider(&done);
	if (done) {
		Files_ReadSet(&Hex_R, Files_Rider__typ, &x);
		Texts_WriteString(&Hex_W, Texts_Writer__typ, (CHAR*)"SET", (LONGINT)4);
		Texts_Write(&Hex_W, Texts_Writer__typ, 0x09);
		Texts_Write(&Hex_W, Texts_Writer__typ, '{');
		i = 0;
		last = -1;
		do {
			if (__IN(i, x)) {
				if (last >= 0) {
					Texts_WriteInt(&Hex_W, Texts_Writer__typ, last, 0);
					Texts_Write(&Hex_W, Texts_Writer__typ, ',');
				}
				last = i;
			}
			i += 1;
		} while (!(i == 32));
		if (last >= 0) {
			Texts_WriteInt(&Hex_W, Texts_Writer__typ, last, 0);
		}
		Texts_Write(&Hex_W, Texts_Writer__typ, '}');
		Texts_WriteLn(&Hex_W, Texts_Writer__typ);
		Texts_Append(Oberon_Log, Hex_W.buf);
	}
}

void Hex_GetBool (void)
{
	CHAR x = 0;
	BOOLEAN done = 0;
	Hex_SetRider(&done);
	if (done) {
		Files_Read(&Hex_R, Files_Rider__typ, (void*)&x);
		Texts_WriteString(&Hex_W, Texts_Writer__typ, (CHAR*)"BOOLEAN", (LONGINT)8);
		Texts_Write(&Hex_W, Texts_Writer__typ, 0x09);
		if (x == 0x01) {
			Texts_WriteString(&Hex_W, Texts_Writer__typ, (CHAR*)"TRUE", (LONGINT)5);
		} else {
			Texts_WriteString(&Hex_W, Texts_Writer__typ, (CHAR*)"FALSE", (LONGINT)6);
		}
		Texts_WriteLn(&Hex_W, Texts_Writer__typ);
		Texts_Append(Oberon_Log, Hex_W.buf);
	}
}

void Hex_Search (void)
{
	Hex_Frame F = NIL;
	CHAR name[64];
	CHAR ch = 0;
	Texts_Text T = NIL;
	LONGINT beg = 0, end = 0, time = 0;
	Texts_Reader R = {0};
	INTEGER len = 0;
	Hex_GetFrame(&F, (void*)name, 64);
	if (F != NIL) {
		Oberon_GetSelection(&T, &beg, &end, &time);
		if (time > 0) {
			Texts_OpenReader(&R, Texts_Reader__typ, T, beg);
			Texts_Read(&R, Texts_Reader__typ, &ch);
			len = 0;
			while (len <= 64 && Texts_Pos(&R, Texts_Reader__typ) <= end) {
				name[__X(len, 64)] = ch;
				len += 1;
				Texts_Read(&R, Texts_Reader__typ, &ch);
			}
			Hex_SearchPat(F, name, 64, len);
		}
	}
}

static void EnumPtrs(void (*P)(void*))
{
	P(Hex_font);
	P(Hex_hexcurs);
	P(Hex_asccurs);
	__ENUMR(&Hex_R, Files_Rider__typ, 20, 1, P);
	__ENUMR(&Hex_W, Texts_Writer__typ, 36, 1, P);
}

__TDESC(Hex_CursorCoordDesc, 1, 0) = {__TDFLDS("CursorCoordDesc", 4), {-4}};
__TDESC(Hex_ModelDesc, 1, 1) = {__TDFLDS("ModelDesc", 68), {64, -8}};
__TDESC(Hex_FrameDesc, 1, 5) = {__TDFLDS("FrameDesc", 52), {0, 4, 24, 28, 40, -24}};
__TDESC(Hex_UpdateMsg, 1, 1) = {__TDFLDS("UpdateMsg", 16), {4, -8}};
__TDESC(Hex_CursorMsg, 1, 1) = {__TDFLDS("CursorMsg", 8), {4, -8}};

export void *Hex__init(void)
{
	__DEFMOD;
	__IMPORT(Display);
	__IMPORT(Files);
	__IMPORT(Fonts);
	__IMPORT(Input);
	__IMPORT(MenuViewers);
	__IMPORT(Oberon);
	__IMPORT(TextFrames);
	__IMPORT(Texts);
	__IMPORT(Viewers);
	__REGMOD(Hex, EnumPtrs);
	__REGCMD("GetBool", Hex_GetBool);
	__REGCMD("GetInt", Hex_GetInt);
	__REGCMD("GetLInt", Hex_GetLInt);
	__REGCMD("GetLReal", Hex_GetLReal);
	__REGCMD("GetNum", Hex_GetNum);
	__REGCMD("GetReal", Hex_GetReal);
	__REGCMD("GetSInt", Hex_GetSInt);
	__REGCMD("GetSet", Hex_GetSet);
	__REGCMD("Open", Hex_Open);
	__REGCMD("OpenText", Hex_OpenText);
	__REGCMD("Search", Hex_Search);
	__REGCMD("SetFont", Hex_SetFont);
	__REGCMD("Store", Hex_Store);
	__INITYP(Hex_CursorCoordDesc, Hex_CursorCoordDesc, 0);
	__INITYP(Hex_ModelDesc, Hex_ModelDesc, 0);
	__INITYP(Hex_FrameDesc, Display_FrameDesc, 1);
	__INITYP(Hex_UpdateMsg, Display_FrameMsg, 1);
	__INITYP(Hex_CursorMsg, Display_FrameMsg, 1);
/* BEGIN */
	Texts_OpenWriter(&Hex_W, Texts_Writer__typ);
	Hex_ChangeFont((CHAR*)"Courier12.Scn.Fnt", (LONGINT)18, &Hex_res);
	if (Hex_res != 0) {
		Hex_FontLogText((CHAR*)"Courier12.Scn.Fnt", (LONGINT)18, Hex_res);
		__HALT(99);
	}
	__ENDMOD;
}
