/* Ofront 1.4 -xtspkaG */
#include "SYSTEM.h"
#include "Files.h"
#include "In.h"
#include "KeplerFrames.h"
#include "KeplerGraphs.h"
#include "KeplerPorts.h"
#include "Math.h"
#include "MenuViewers.h"
#include "Oberon.h"
#include "Printer.h"
#include "TextFrames.h"
#include "Texts.h"
#include "Viewers.h"


static Texts_Writer Kepler_W;
static MenuViewers_Viewer Kepler_AttrV;
static Texts_Text Kepler_AttrT;


export void Kepler_AlignToGrid (void);
export void Kepler_AlignX (void);
export void Kepler_AlignY (void);
static void Kepler_Backup (CHAR *name, LONGINT name__len);
export void Kepler_Constellations (void);
export void Kepler_Delete (void);
static void Kepler_InitAttrV (void);
export void Kepler_Join (void);
export void Kepler_Open (void);
export void Kepler_Print (void);
export void Kepler_Recall (void);
export void Kepler_Reset (void);
export void Kepler_RotatePoints (void);
export void Kepler_ScalePoints (void);
export void Kepler_SendBack (void);
export void Kepler_SetGrid (void);
export void Kepler_SetScale (void);
export void Kepler_Split (void);
export void Kepler_Store (void);


static struct Print__12 {
	struct Print__12 *lnk;
} *Print__12_s;

static void PrintUnit__13 (KeplerGraphs_Graph G, INTEGER nofcopies);

static void PrintUnit__13 (KeplerGraphs_Graph G, INTEGER nofcopies)
{
	KeplerPorts_PrinterPort P = NIL;
	__NEW(P, KeplerPorts_PrinterPortDesc);
	P->X = 0;
	P->Y = 0;
	P->W = 32767;
	P->H = 3300;
	P->x0 = 0;
	P->y0 = 0;
	P->scale = 1;
	__KeplerGraphs_Object_Draw((void*)G, (void*)P);
	Printer_Page(nofcopies);
}

void Kepler_Print (void)
{
	Texts_Scanner S = {0};
	KeplerGraphs_Graph source = NIL;
	Viewers_Viewer V = NIL;
	INTEGER nofcopies = 0;
	struct Print__12 _s;
	_s.lnk = Print__12_s;
	Print__12_s = &_s;
	Texts_WriteString(&Kepler_W, Texts_Writer__typ, (CHAR*)"Kepler.Print", (LONGINT)13);
	Texts_WriteLn(&Kepler_W, Texts_Writer__typ);
	Texts_Append(Oberon_Log, Kepler_W.buf);
	Texts_OpenScanner(&S, Texts_Scanner__typ, Oberon_Par->text, Oberon_Par->pos);
	Texts_Scan(&S, Texts_Scanner__typ);
	if (S.class == 1) {
		Printer_Open((void*)S.s, 256, (void*)Oberon_User, 12, Oberon_Password);
		if (Printer_res == 0) {
			Texts_Scan(&S, Texts_Scanner__typ);
			nofcopies = 1;
			if (S.class == 3) {
				nofcopies = (int)S.i;
				Texts_Scan(&S, Texts_Scanner__typ);
			}
			while (S.class == 1) {
				source = KeplerGraphs_Old(S.s, 256);
				if (source == NIL) {
					Texts_WriteString(&Kepler_W, Texts_Writer__typ, (CHAR*)" -- not found: ", (LONGINT)16);
					Texts_WriteString(&Kepler_W, Texts_Writer__typ, S.s, 256);
					Texts_WriteLn(&Kepler_W, Texts_Writer__typ);
					Texts_Append(Oberon_Log, Kepler_W.buf);
				} else {
					PrintUnit__13(source, nofcopies);
				}
				Texts_Scan(&S, Texts_Scanner__typ);
			}
			if (S.class == 6 && S.c == '*') {
				V = Oberon_MarkedViewer();
				if (__ISP(V, MenuViewers_ViewerDesc, 2) && __ISP(V->dsc->next, KeplerFrames_FrameDesc, 3)) {
					PrintUnit__13(__GUARDP(V->dsc->next, KeplerFrames_FrameDesc, 3)->G, nofcopies);
				}
			}
			Printer_Close();
		} else {
			if (Printer_res == 1) {
				Texts_WriteString(&Kepler_W, Texts_Writer__typ, (CHAR*)" no such printer", (LONGINT)17);
			} else if (Printer_res == 2) {
				Texts_WriteString(&Kepler_W, Texts_Writer__typ, (CHAR*)" no link", (LONGINT)9);
			} else if (Printer_res == 3) {
				Texts_WriteString(&Kepler_W, Texts_Writer__typ, (CHAR*)" printer not ready", (LONGINT)19);
			} else if (Printer_res == 4) {
				Texts_WriteString(&Kepler_W, Texts_Writer__typ, (CHAR*)" no permission", (LONGINT)15);
			}
			Texts_WriteLn(&Kepler_W, Texts_Writer__typ);
			Texts_Append(Oberon_Log, Kepler_W.buf);
		}
	} else {
		Texts_WriteString(&Kepler_W, Texts_Writer__typ, (CHAR*)" no printer specified", (LONGINT)22);
		Texts_WriteLn(&Kepler_W, Texts_Writer__typ);
		Texts_Append(Oberon_Log, Kepler_W.buf);
	}
	Print__12_s = _s.lnk;
}

void Kepler_Open (void)
{
	MenuViewers_Viewer V = NIL;
	INTEGER X = 0, Y = 0, grid = 0;
	KeplerGraphs_Graph G = NIL;
	KeplerFrames_Frame F = NIL;
	CHAR name[32];
	In_Open();
	In_Name((void*)name, 32);
	if (In_Done) {
		In_Int(&grid);
		if (!In_Done) {
			grid = 5;
		}
		Oberon_AllocateUserViewer(Oberon_Mouse.X, &X, &Y);
		G = KeplerGraphs_Old(name, 32);
		if (G == NIL) {
			__NEW(G, KeplerGraphs_GraphDesc);
			G->seltime = -1;
		}
		F = KeplerFrames_New(G);
		F->grid = grid;
		V = MenuViewers_New((void*)TextFrames_NewMenu(name, 32, (CHAR*)"System.Close  System.Copy  System.Grow  Kepler.Store", (LONGINT)53), (void*)F, TextFrames_menuH, X, Y);
	}
}

static void Kepler_InitAttrV (void)
{
	INTEGER X = 0, Y = 0;
	Texts_Delete(Kepler_AttrT, 0, Kepler_AttrT->len);
	if (Kepler_AttrV == NIL || Kepler_AttrV->state <= 0) {
		Oberon_AllocateSystemViewer(Oberon_Mouse.X, &X, &Y);
		Kepler_AttrV = MenuViewers_New((void*)TextFrames_NewMenu((CHAR*)"Kepler", (LONGINT)7, (CHAR*)"System.Close  System.Grow  Kepler.Delete  Kepler.SendBack  Edit.Store", (LONGINT)70), (void*)TextFrames_NewText(Kepler_AttrT, 0), TextFrames_menuH, X, Y);
	}
}

void Kepler_Constellations (void)
{
	KeplerGraphs_Constellation c = NIL;
	CHAR mod[32], class[32];
	KeplerGraphs_Graph sel = NIL;
	INTEGER minstate = 0;
	In_Open();
	In_Int(&minstate);
	if (!In_Done) {
		minstate = 1;
	}
	KeplerFrames_GetSelection(&sel);
	if (sel != NIL) {
		Kepler_InitAttrV();
		c = sel->cons;
		while (c != NIL) {
			if (__KeplerGraphs_Constellation_State(c) >= minstate) {
				Texts_WriteInt(&Kepler_W, Texts_Writer__typ, (LONGINT)c, 10);
				Texts_WriteString(&Kepler_W, Texts_Writer__typ, (CHAR*)"  ", (LONGINT)3);
				KeplerGraphs_GetType((void*)c, (void*)mod, 32, (void*)class, 32);
				Texts_WriteString(&Kepler_W, Texts_Writer__typ, mod, 32);
				Texts_Write(&Kepler_W, Texts_Writer__typ, '.');
				Texts_WriteString(&Kepler_W, Texts_Writer__typ, class, 32);
				Texts_WriteLn(&Kepler_W, Texts_Writer__typ);
			}
			Texts_Append(Kepler_AttrT, Kepler_W.buf);
			c = c->next;
		}
	}
}

void Kepler_Delete (void)
{
	Texts_Scanner S = {0};
	KeplerGraphs_Graph sel = NIL;
	TextFrames_Frame F = NIL;
	Texts_Reader R = {0};
	CHAR ch = 0;
	KeplerFrames_GetSelection(&sel);
	if (sel != NIL) {
		if (Kepler_AttrV != NIL) {
			F = __GUARDP(Kepler_AttrV->dsc->next, TextFrames_FrameDesc, 1);
			if (F->hasSel) {
				Texts_OpenScanner(&S, Texts_Scanner__typ, Kepler_AttrT, F->selbeg.org);
				Texts_Scan(&S, Texts_Scanner__typ);
				if (S.class == 3) {
					__KeplerGraphs_Graph_Delete(sel, (KeplerGraphs_Object)S.i);
					Texts_OpenReader(&R, Texts_Reader__typ, F->text, F->selbeg.org);
					Texts_Read(&R, Texts_Reader__typ, &ch);
					while (ch >= ' ' || ch == 0x09) {
						Texts_Read(&R, Texts_Reader__typ, &ch);
					}
					Texts_Delete(F->text, F->selbeg.org, Texts_Pos(&R, Texts_Reader__typ));
				}
			}
		}
	}
}

static void Kepler_Backup (CHAR *name, LONGINT name__len)
{
	INTEGER res = 0, i = 0;
	CHAR bak[64];
	i = 0;
	while (name[__X(i, name__len)] != 0x00) {
		i += 1;
	}
	if (i < 60) {
		__COPY(name, bak, 64);
		bak[__X(i, 64)] = '.';
		bak[__X(i + 1, 64)] = 'B';
		bak[__X(i + 2, 64)] = 'a';
		bak[__X(i + 3, 64)] = 'k';
		bak[__X(i + 4, 64)] = 0x00;
		Files_Rename(name, name__len, bak, 64, &res);
	}
}

void Kepler_Store (void)
{
	Oberon_ParList par = NIL;
	Viewers_Viewer V = NIL;
	Texts_Text T = NIL;
	Texts_Scanner S = {0};
	Files_File f = NIL;
	Files_Rider R = {0};
	LONGINT beg = 0, end = 0, time = 0;
	par = Oberon_Par;
	if (par->frame == par->vwr->dsc) {
		V = par->vwr;
		Texts_OpenScanner(&S, Texts_Scanner__typ, __GUARDP(V->dsc, TextFrames_FrameDesc, 1)->text, 0);
	} else {
		V = Oberon_MarkedViewer();
		Texts_OpenScanner(&S, Texts_Scanner__typ, par->text, par->pos);
	}
	Texts_Scan(&S, Texts_Scanner__typ);
	if (S.class == 6 && S.c == '^') {
		Oberon_GetSelection(&T, &beg, &end, &time);
		if (time >= 0) {
			Texts_OpenScanner(&S, Texts_Scanner__typ, T, beg);
			Texts_Scan(&S, Texts_Scanner__typ);
		}
	}
	if ((S.class == 1 && V->dsc != NIL) && __ISP(V->dsc->next, KeplerFrames_FrameDesc, 3)) {
		Texts_WriteString(&Kepler_W, Texts_Writer__typ, (CHAR*)"Kepler.Store ", (LONGINT)14);
		Texts_WriteString(&Kepler_W, Texts_Writer__typ, S.s, 256);
		Texts_WriteLn(&Kepler_W, Texts_Writer__typ);
		Texts_Append(Oberon_Log, Kepler_W.buf);
		f = Files_New(S.s, 256);
		Files_Set(&R, Files_Rider__typ, f, 0);
		KeplerGraphs_Reset();
		KeplerGraphs_WriteObj(&R, Files_Rider__typ, (void*)__GUARDP(V->dsc->next, KeplerFrames_FrameDesc, 3)->G);
		Kepler_Backup((void*)S.s, 256);
		Files_Register(f);
	}
}

void Kepler_SetGrid (void)
{
	INTEGER i = 0;
	KeplerFrames_Frame F = NIL;
	Viewers_Viewer V = NIL;
	V = Oberon_MarkedViewer();
	if (__ISP(V->dsc->next, KeplerFrames_FrameDesc, 3)) {
		F = __GUARDP(V->dsc->next, KeplerFrames_FrameDesc, 3);
		In_Open();
		In_Int(&i);
		if (In_Done) {
			F->grid = i;
			__KeplerFrames_Frame_Restore(F, F->X, F->Y, F->W, F->H);
		}
	}
}

void Kepler_SetScale (void)
{
	KeplerFrames_Frame F = NIL;
	Viewers_Viewer V = NIL;
	INTEGER X = 0, Y = 0, i = 0;
	V = Oberon_MarkedViewer();
	if (__ISP(V->dsc->next, KeplerFrames_FrameDesc, 3)) {
		F = __GUARDP(V->dsc->next, KeplerFrames_FrameDesc, 3);
		In_Open();
		In_Int(&i);
		if (In_Done && i > 0) {
			X = Oberon_Pointer.X;
			Y = Oberon_Pointer.Y;
			F->x0 = (X - F->X) * (int)((int)i) - __KeplerPorts_Port_Cx((void*)F, X);
			F->y0 = ((Y - F->Y) - F->H) * (int)((int)i) - __KeplerPorts_Port_Cy((void*)F, Y);
			F->scale = i;
			__KeplerFrames_Frame_Restore(F, F->X, F->Y, F->W, F->H);
		}
	}
}

static struct Join__8 {
	KeplerGraphs_Graph *G;
	KeplerGraphs_Star *f;
	struct Join__8 *lnk;
} *Join__8_s;

static void JoinCons__9 (KeplerGraphs_Constellation c);

static void JoinCons__9 (KeplerGraphs_Constellation c)
{
	INTEGER i = 0;
	KeplerGraphs_Star p = NIL;
	i = 0;
	while (i < c->nofpts) {
		p = c->p[__X(i, 4)];
		if ((p->sel && !__ISP(p, KeplerGraphs_PlanetDesc, 2)) && p != *Join__8_s->f) {
			__KeplerGraphs_Graph_Move(*Join__8_s->G, p, (*Join__8_s->f)->x - p->x, (*Join__8_s->f)->y - p->y);
			c->p[__X(i, 4)] = *Join__8_s->f;
			(*Join__8_s->f)->refcnt += 1;
			p->refcnt -= 1;
			if (p->refcnt == 0) {
				__KeplerGraphs_Graph_Delete(*Join__8_s->G, (void*)p);
			}
		} else if (__ISP(p, KeplerGraphs_PlanetDesc, 2)) {
			JoinCons__9(__GUARDP(p, KeplerGraphs_PlanetDesc, 2)->c);
		}
		i += 1;
	}
}

void Kepler_Join (void)
{
	KeplerGraphs_Graph G = NIL;
	KeplerGraphs_Star f = NIL, s = NIL;
	KeplerGraphs_Constellation c = NIL;
	struct Join__8 _s;
	_s.G = &G;
	_s.f = &f;
	_s.lnk = Join__8_s;
	Join__8_s = &_s;
	G = KeplerFrames_Focus;
	if (KeplerFrames_nofpts >= 1) {
		KeplerFrames_ConsumePoint(&f);
		f->refcnt -= 1;
		c = G->cons;
		while (c != NIL) {
			JoinCons__9(c);
			c = c->next;
		}
		__KeplerGraphs_Graph_SendToBack(G, (void*)f);
		s = f->next;
		while (s != NIL) {
			if (__ISP(s, KeplerGraphs_PlanetDesc, 2) && s != f) {
				JoinCons__9(__GUARDP(s, KeplerGraphs_PlanetDesc, 2)->c);
			}
			s = s->next;
		}
	}
	Join__8_s = _s.lnk;
}

static struct Split__22 {
	KeplerGraphs_Graph *G;
	struct Split__22 *lnk;
} *Split__22_s;

static void SplitCons__23 (KeplerGraphs_Constellation c);

static void SplitCons__23 (KeplerGraphs_Constellation c)
{
	INTEGER i = 0;
	KeplerGraphs_Star p = NIL, q = NIL;
	INTEGER _for__25 = 0;
	_for__25 = c->nofpts - 1;
	i = 0;
	while (i <= _for__25) {
		p = c->p[__X(i, 4)];
		if (p->sel) {
			__NEW(q, KeplerGraphs_StarDesc);
			c->p[__X(i, 4)] = q;
			__GUARDEQP(q, KeplerGraphs_StarDesc) = *p;
			q->refcnt = 1;
			q->next = (*Split__22_s->G)->stars;
			(*Split__22_s->G)->stars = q;
			p->refcnt -= 1;
			if (p->refcnt == 0 && !__ISP(p, KeplerGraphs_PlanetDesc, 2)) {
				__KeplerGraphs_Graph_Delete(*Split__22_s->G, (void*)p);
			}
		}
		i += 1;
	}
}

void Kepler_Split (void)
{
	KeplerGraphs_Graph G = NIL;
	KeplerGraphs_Constellation c = NIL;
	KeplerGraphs_Star s = NIL;
	struct Split__22 _s;
	_s.G = &G;
	_s.lnk = Split__22_s;
	Split__22_s = &_s;
	KeplerFrames_GetSelection(&G);
	if (G != NIL) {
		c = G->cons;
		while (c != NIL) {
			SplitCons__23(c);
			c = c->next;
		}
		s = G->stars;
		while (s != NIL) {
			if (__ISP(s, KeplerGraphs_PlanetDesc, 2)) {
				SplitCons__23(__GUARDP(s, KeplerGraphs_PlanetDesc, 2)->c);
			}
			s = s->next;
		}
	}
	Split__22_s = _s.lnk;
}

void Kepler_SendBack (void)
{
	Texts_Scanner S = {0};
	KeplerGraphs_Graph sel = NIL;
	TextFrames_Frame F = NIL;
	KeplerFrames_GetSelection(&sel);
	if (sel != NIL) {
		if (Kepler_AttrV != NIL) {
			F = __GUARDP(Kepler_AttrV->dsc->next, TextFrames_FrameDesc, 1);
			if (F->hasSel) {
				Texts_OpenScanner(&S, Texts_Scanner__typ, Kepler_AttrT, F->selbeg.org);
				Texts_Scan(&S, Texts_Scanner__typ);
				if (S.class == 3) {
					__KeplerGraphs_Graph_SendToBack(sel, (KeplerGraphs_Object)S.i);
				}
			}
		}
	}
}

void Kepler_AlignX (void)
{
	KeplerGraphs_Graph G = NIL;
	KeplerGraphs_Star s = NIL, p = NIL;
	if (KeplerFrames_nofpts > 0) {
		KeplerFrames_GetPoint(&p);
		KeplerFrames_GetSelection(&G);
		s = G->stars;
		while (s != NIL) {
			if (s->sel && !__ISP(s, KeplerGraphs_PlanetDesc, 2)) {
				__KeplerGraphs_Graph_Move(G, s, p->x - s->x, 0);
			}
			s = s->next;
		}
	}
}

void Kepler_AlignY (void)
{
	KeplerGraphs_Graph G = NIL;
	KeplerGraphs_Star s = NIL, p = NIL;
	if (KeplerFrames_nofpts > 0) {
		KeplerFrames_GetPoint(&p);
		KeplerFrames_GetSelection(&G);
		s = G->stars;
		while (s != NIL) {
			if (s->sel && !__ISP(s, KeplerGraphs_PlanetDesc, 2)) {
				__KeplerGraphs_Graph_Move(G, s, 0, p->y - s->y);
			}
			s = s->next;
		}
	}
}

void Kepler_AlignToGrid (void)
{
	Viewers_Viewer V = NIL;
	KeplerFrames_Frame F = NIL;
	KeplerGraphs_Star s = NIL;
	INTEGER X = 0, Y = 0;
	V = Oberon_MarkedViewer();
	if (__ISP(V->dsc->next, KeplerFrames_FrameDesc, 3)) {
		F = __GUARDP(V->dsc->next, KeplerFrames_FrameDesc, 3);
		if (F->grid > 0) {
			s = F->G->stars;
			while (s != NIL) {
				if (s->sel && !__ISP(s, KeplerGraphs_PlanetDesc, 2)) {
					X = __KeplerPorts_Port_CX((void*)F, s->x);
					Y = __KeplerPorts_Port_CY((void*)F, s->y);
					KeplerFrames_AlignToGrid(F, &X, &Y);
					__KeplerGraphs_Graph_Move(F->G, s, __KeplerPorts_Port_Cx((void*)F, X) - s->x, __KeplerPorts_Port_Cy((void*)F, Y) - s->y);
				}
				s = s->next;
			}
		}
	}
}

void Kepler_Reset (void)
{
	Viewers_Viewer V = NIL;
	KeplerFrames_Frame F = NIL;
	V = Oberon_MarkedViewer();
	if (__ISP(V->dsc->next, KeplerFrames_FrameDesc, 3)) {
		F = __GUARDP(V->dsc->next, KeplerFrames_FrameDesc, 3);
		F->x0 = 0;
		F->y0 = 0;
		F->scale = 4;
		__KeplerFrames_Frame_Restore(F, F->X, F->Y, F->W, F->H);
	}
}

void Kepler_Recall (void)
{
	KeplerGraphs_Recall();
}

void Kepler_ScalePoints (void)
{
	KeplerGraphs_Graph sel = NIL;
	KeplerGraphs_Star p0 = NIL, p1 = NIL, p2 = NIL, s = NIL;
	REAL cx = 0, cy = 0, dx = 0, dy = 0;
	KeplerFrames_GetSelection(&sel);
	if (sel != NIL && KeplerFrames_nofpts >= 3) {
		KeplerFrames_GetPoint(&p0);
		KeplerFrames_GetPoint(&p1);
		KeplerFrames_GetPoint(&p2);
		if (p0->x == p1->x) {
			cx = (REAL)1;
		} else {
			cx = (p0->x - p2->x) / (REAL)(p0->x - p1->x);
		}
		dx = p0->x - p0->x * cx;
		if (p0->y == p1->y) {
			cy = (REAL)1;
		} else {
			cy = (p0->y - p2->y) / (REAL)(p0->y - p1->y);
		}
		dy = p0->y - p0->y * cy;
		s = sel->stars;
		while (s != NIL) {
			if (s->sel && !__ISP(s, KeplerGraphs_PlanetDesc, 2)) {
				__KeplerGraphs_Graph_Move(sel, s, (int)__ENTIER((s->x * cx + dx) - s->x), (int)__ENTIER((s->y * cy + dy) - s->y));
			}
			s = s->next;
		}
	}
}

void Kepler_RotatePoints (void)
{
	KeplerGraphs_Graph sel = NIL;
	KeplerGraphs_Star p0 = NIL, p1 = NIL, p2 = NIL, s = NIL;
	KeplerGraphs_StarDesc p1r = {0}, p2r = {0};
	REAL p1rAbs = 0, p2rAbs = 0, cos = 0, sin = 0, factor = 0;
	KeplerFrames_GetSelection(&sel);
	if (sel != NIL && KeplerFrames_nofpts >= 3) {
		KeplerFrames_GetPoint(&p0);
		KeplerFrames_GetPoint(&p1);
		KeplerFrames_GetPoint(&p2);
		p1r.x = p1->x - p0->x;
		p1r.y = p1->y - p0->y;
		p2r.x = p2->x - p0->x;
		p2r.y = p2->y - p0->y;
		if (p1r.x == 0 && p1r.y == 0 || p2r.x == 0 && p2r.y == 0) {
			return;
		}
		p1rAbs = Math_sqrt(p1r.x * p1r.x + p1r.y * p1r.y);
		p2rAbs = Math_sqrt(p2r.x * p2r.x + p2r.y * p2r.y);
		cos = ((p1r.x * p2r.x + p1r.y * p2r.y) / p1rAbs) / p2rAbs;
		sin = Math_sqrt((REAL)1 - cos * cos);
		factor = p2rAbs / p1rAbs;
		if (((p0->x * (p1->y - p2->y) - p0->y * (p1->x - p2->x)) + p1->x * p2->y) - p1->y * p2->x < 0) {
			sin = -sin;
		}
		s = sel->stars;
		while (s != NIL) {
			if (s->sel && !__ISP(s, KeplerGraphs_PlanetDesc, 2)) {
				__KeplerGraphs_Graph_Move(sel, s, ((int)__ENTIER(((s->x - p0->x) * cos - (s->y - p0->y) * sin) * factor) - s->x) + p0->x, ((int)__ENTIER(((s->x - p0->x) * sin + (s->y - p0->y) * cos) * factor) - s->y) + p0->y);
			}
			s = s->next;
		}
	}
}

static void EnumPtrs(void (*P)(void*))
{
	__ENUMR(&Kepler_W, Texts_Writer__typ, 36, 1, P);
	P(Kepler_AttrV);
	P(Kepler_AttrT);
}


export void *Kepler__init(void)
{
	__DEFMOD;
	__IMPORT(Files);
	__IMPORT(In);
	__IMPORT(KeplerFrames);
	__IMPORT(KeplerGraphs);
	__IMPORT(KeplerPorts);
	__IMPORT(Math);
	__IMPORT(MenuViewers);
	__IMPORT(Oberon);
	__IMPORT(Printer);
	__IMPORT(TextFrames);
	__IMPORT(Texts);
	__IMPORT(Viewers);
	__REGMOD(Kepler, EnumPtrs);
	__REGCMD("AlignToGrid", Kepler_AlignToGrid);
	__REGCMD("AlignX", Kepler_AlignX);
	__REGCMD("AlignY", Kepler_AlignY);
	__REGCMD("Constellations", Kepler_Constellations);
	__REGCMD("Delete", Kepler_Delete);
	__REGCMD("Join", Kepler_Join);
	__REGCMD("Open", Kepler_Open);
	__REGCMD("Print", Kepler_Print);
	__REGCMD("Recall", Kepler_Recall);
	__REGCMD("Reset", Kepler_Reset);
	__REGCMD("RotatePoints", Kepler_RotatePoints);
	__REGCMD("ScalePoints", Kepler_ScalePoints);
	__REGCMD("SendBack", Kepler_SendBack);
	__REGCMD("SetGrid", Kepler_SetGrid);
	__REGCMD("SetScale", Kepler_SetScale);
	__REGCMD("Split", Kepler_Split);
	__REGCMD("Store", Kepler_Store);
/* BEGIN */
	Texts_OpenWriter(&Kepler_W, Texts_Writer__typ);
	Kepler_AttrT = TextFrames_Text((CHAR*)"", (LONGINT)1);
	__ENDMOD;
}
