/* Ofront 1.4 -xtspkaG */
#include "SYSTEM.h"
#include "Display.h"
#include "Files.h"
#include "Fonts.h"
#include "KeplerFrames.h"
#include "KeplerGraphs.h"
#include "Math.h"
#include "Oberon.h"
#include "Texts.h"

typedef
	struct Kepler2_FractionDesc *Kepler2_Fraction;

typedef
	struct Kepler2_FractionDesc { /* KeplerGraphs_PlanetDesc */
		INTEGER x, y, refcnt;
		char _prvt0[2];
		BOOLEAN sel;
		KeplerGraphs_Star next;
		KeplerGraphs_Constellation c;
		LONGINT cnt, dn;
	} Kepler2_FractionDesc;

export void Kepler2_Fraction_Calc (Kepler2_Fraction self);
export void Kepler2_Fraction_Read (Kepler2_Fraction self, Files_Rider *R, LONGINT *R__typ);
export void Kepler2_Fraction_Write (Kepler2_Fraction self, Files_Rider *R, LONGINT *R__typ);

typedef
	struct Kepler2_OffsetDesc *Kepler2_Offset;

typedef
	struct Kepler2_OffsetDesc { /* KeplerGraphs_PlanetDesc */
		INTEGER x, y, refcnt;
		char _prvt0[2];
		BOOLEAN sel;
		KeplerGraphs_Star next;
		KeplerGraphs_Constellation c;
		INTEGER dx, dy;
	} Kepler2_OffsetDesc;

export void Kepler2_Offset_Calc (Kepler2_Offset self);
export void Kepler2_Offset_Read (Kepler2_Offset self, Files_Rider *R, LONGINT *R__typ);
export void Kepler2_Offset_Write (Kepler2_Offset self, Files_Rider *R, LONGINT *R__typ);

typedef
	struct Kepler2_XYDesc *Kepler2_XY;

typedef
	struct Kepler2_XYDesc { /* KeplerGraphs_PlanetDesc */
		INTEGER x, y, refcnt;
		char _prvt0[2];
		BOOLEAN sel;
		KeplerGraphs_Star next;
		KeplerGraphs_Constellation c;
	} Kepler2_XYDesc;

export void Kepler2_XY_Calc (Kepler2_XY self);



export LONGINT *Kepler2_FractionDesc__typ;
export LONGINT *Kepler2_XYDesc__typ;
export LONGINT *Kepler2_OffsetDesc__typ;

export void Kepler2_NewFractions (void);
export void Kepler2_NewOffset (void);
export void Kepler2_NewXY (void);


void Kepler2_Fraction_Calc (Kepler2_Fraction self)
{
	INTEGER x0 = 0, x1 = 0, y0 = 0, y1 = 0;
	x0 = self->c->p[0]->x;
	x1 = self->c->p[1]->x;
	y0 = self->c->p[0]->y;
	y1 = self->c->p[1]->y;
	self->x = (int)((LONGINT)x0 + __DIV((LONGINT)(x1 - x0) * self->cnt, self->dn));
	self->y = (int)((LONGINT)y0 + __DIV((LONGINT)(y1 - y0) * self->cnt, self->dn));
}

void Kepler2_Fraction_Write (Kepler2_Fraction self, Files_Rider *R, LONGINT *R__typ)
{
	Files_WriteNum(&*R, R__typ, self->cnt);
	Files_WriteNum(&*R, R__typ, self->dn);
	KeplerGraphs_Planet_Write((void*)self, &*R, R__typ);
}

void Kepler2_Fraction_Read (Kepler2_Fraction self, Files_Rider *R, LONGINT *R__typ)
{
	Files_ReadNum(&*R, R__typ, &self->cnt);
	Files_ReadNum(&*R, R__typ, &self->dn);
	KeplerGraphs_Planet_Read((void*)self, &*R, R__typ);
}

void Kepler2_NewFractions (void)
{
	Kepler2_Fraction f = NIL;
	KeplerGraphs_Constellation c = NIL;
	Texts_Scanner S = {0};
	KeplerGraphs_Star p0 = NIL, p1 = NIL;
	LONGINT i = 0;
	if (KeplerFrames_nofpts >= 2) {
		Texts_OpenScanner(&S, Texts_Scanner__typ, Oberon_Par->text, Oberon_Par->pos);
		Texts_Scan(&S, Texts_Scanner__typ);
		if (S.class == 3 && S.i > 0) {
			KeplerFrames_ConsumePoint(&p0);
			KeplerFrames_ConsumePoint(&p1);
			i = S.i - 1;
			while (i > 0) {
				__NEW(f, Kepler2_FractionDesc);
				__NEW(c, KeplerGraphs_ConsDesc);
				c->nofpts = 2;
				f->c = c;
				c->p[0] = p0;
				c->p[1] = p1;
				p0->refcnt += 1;
				p1->refcnt += 1;
				f->cnt = (int)i;
				f->dn = (int)S.i;
				__KeplerGraphs_Planet_Calc((void*)f);
				__KeplerGraphs_Graph_Append(KeplerFrames_Focus, (void*)f);
				__KeplerGraphs_Graph_FlipSelection(KeplerFrames_Focus, (void*)f);
				i -= 1;
			}
			p0->refcnt -= 1;
			p1->refcnt -= 1;
		}
	}
}

void Kepler2_XY_Calc (Kepler2_XY self)
{
	self->x = self->c->p[0]->x;
	self->y = self->c->p[1]->y;
}

void Kepler2_NewXY (void)
{
	Kepler2_XY cp = NIL;
	KeplerGraphs_Constellation c = NIL;
	if (KeplerFrames_nofpts >= 2) {
		__NEW(cp, Kepler2_XYDesc);
		__NEW(c, KeplerGraphs_ConsDesc);
		c->nofpts = 2;
		KeplerFrames_ConsumePoint(&c->p[0]);
		KeplerFrames_ConsumePoint(&c->p[1]);
		cp->c = c;
		__KeplerGraphs_Planet_Calc((void*)cp);
		__KeplerGraphs_Graph_Append(KeplerFrames_Focus, (void*)cp);
		__KeplerGraphs_Graph_FlipSelection(KeplerFrames_Focus, (void*)cp);
	}
}

void Kepler2_Offset_Calc (Kepler2_Offset self)
{
	self->x = self->c->p[0]->x + self->dx;
	self->y = self->c->p[0]->y + self->dy;
}

void Kepler2_Offset_Read (Kepler2_Offset self, Files_Rider *R, LONGINT *R__typ)
{
	LONGINT t = 0;
	Files_ReadNum(&*R, R__typ, &t);
	self->dx = (int)t;
	Files_ReadNum(&*R, R__typ, &t);
	self->dy = (int)t;
	KeplerGraphs_Planet_Read((void*)self, &*R, R__typ);
}

void Kepler2_Offset_Write (Kepler2_Offset self, Files_Rider *R, LONGINT *R__typ)
{
	Files_WriteNum(&*R, R__typ, self->dx);
	Files_WriteNum(&*R, R__typ, self->dy);
	KeplerGraphs_Planet_Write((void*)self, &*R, R__typ);
}

void Kepler2_NewOffset (void)
{
	KeplerGraphs_Star p0 = NIL, p1 = NIL;
	Kepler2_Offset offset = NIL;
	if (KeplerFrames_nofpts > 1) {
		KeplerFrames_ConsumePoint(&p0);
		KeplerFrames_GetPoint(&p1);
		__NEW(offset, Kepler2_OffsetDesc);
		__NEW(offset->c, KeplerGraphs_ConsDesc);
		offset->c->p[0] = p0;
		offset->c->nofpts = 1;
		offset->dx = p1->x - p0->x;
		offset->dy = p1->y - p0->y;
		__KeplerGraphs_Planet_Calc((void*)offset);
		__KeplerGraphs_Graph_Append(KeplerFrames_Focus, (void*)offset);
		__KeplerGraphs_Graph_FlipSelection(KeplerFrames_Focus, (void*)offset);
	}
}

__TDESC(Kepler2_FractionDesc, 5, 2) = {__TDFLDS("FractionDesc", 28), {12, 16, -12}};
__TDESC(Kepler2_XYDesc, 5, 2) = {__TDFLDS("XYDesc", 20), {12, 16, -12}};
__TDESC(Kepler2_OffsetDesc, 5, 2) = {__TDFLDS("OffsetDesc", 24), {12, 16, -12}};

export void *Kepler2__init(void)
{
	__DEFMOD;
	__IMPORT(Display);
	__IMPORT(Files);
	__IMPORT(Fonts);
	__IMPORT(KeplerFrames);
	__IMPORT(KeplerGraphs);
	__IMPORT(Math);
	__IMPORT(Oberon);
	__IMPORT(Texts);
	__REGMOD(Kepler2, 0);
	__REGCMD("NewFractions", Kepler2_NewFractions);
	__REGCMD("NewOffset", Kepler2_NewOffset);
	__REGCMD("NewXY", Kepler2_NewXY);
	__INITYP(Kepler2_FractionDesc, KeplerGraphs_PlanetDesc, 3);
	__INITBP(Kepler2_FractionDesc, Kepler2_Fraction_Calc, 3);
	__INITBP(Kepler2_FractionDesc, Kepler2_Fraction_Read, 1);
	__INITBP(Kepler2_FractionDesc, Kepler2_Fraction_Write, 2);
	__INITYP(Kepler2_XYDesc, KeplerGraphs_PlanetDesc, 3);
	__INITBP(Kepler2_XYDesc, Kepler2_XY_Calc, 3);
	__INITYP(Kepler2_OffsetDesc, KeplerGraphs_PlanetDesc, 3);
	__INITBP(Kepler2_OffsetDesc, Kepler2_Offset_Calc, 3);
	__INITBP(Kepler2_OffsetDesc, Kepler2_Offset_Read, 1);
	__INITBP(Kepler2_OffsetDesc, Kepler2_Offset_Write, 2);
/* BEGIN */
	__ENDMOD;
}
