/* Ofront 1.4 -xtspkaG */
#include "SYSTEM.h"
#include "Display.h"
#include "KeplerFrames.h"
#include "KeplerGraphs.h"
#include "KeplerPorts.h"
#include "Math.h"

typedef
	struct Kepler6_BezierDesc *Kepler6_Bezier;

typedef
	struct Kepler6_BezierDesc { /* KeplerGraphs_ConsDesc */
		INTEGER nofpts;
		KeplerGraphs_Star p[4];
		KeplerGraphs_Constellation next;
	} Kepler6_BezierDesc;

export void Kepler6_Bezier_Draw (Kepler6_Bezier s, KeplerPorts_Port f);

typedef
	struct Kepler6_CRSplineDesc *Kepler6_CRSpline;

typedef
	struct Kepler6_CRSplineDesc { /* KeplerGraphs_ConsDesc */
		INTEGER nofpts;
		KeplerGraphs_Star p[4];
		KeplerGraphs_Constellation next;
	} Kepler6_CRSplineDesc;

export void Kepler6_CRSpline_Draw (Kepler6_CRSpline s, KeplerPorts_Port f);



export LONGINT *Kepler6_CRSplineDesc__typ;
export LONGINT *Kepler6_BezierDesc__typ;

static void Kepler6_DrawCurve (KeplerPorts_Port f, REAL a3, REAL a2, REAL a1, REAL a0, REAL b3, REAL b2, REAL b1, REAL b0);
static void Kepler6_GetBoundingBox (REAL a3, REAL a2, REAL a1, REAL a0, REAL b3, REAL b2, REAL b1, REAL b0, INTEGER x1, INTEGER y1, INTEGER x2, INTEGER y2, INTEGER *x, INTEGER *y, INTEGER *w, INTEGER *h);
static BOOLEAN Kepler6_Intersect (KeplerPorts_Port f, INTEGER x, INTEGER y, INTEGER w, INTEGER h);
static INTEGER Kepler6_Max (INTEGER a, INTEGER b);
static INTEGER Kepler6_Min (INTEGER a, INTEGER b);
export void Kepler6_NewBezier (void);
export void Kepler6_NewCRSpline (void);
export void Kepler6_NewClosedBezier (void);
export void Kepler6_NewClosedCRSpline (void);


static INTEGER Kepler6_Min (INTEGER a, INTEGER b)
{
	if (a < b) {
		return a;
	} else {
		return b;
	}
	__RETCHK;
}

static INTEGER Kepler6_Max (INTEGER a, INTEGER b)
{
	if (a < b) {
		return b;
	} else {
		return a;
	}
	__RETCHK;
}

static void Kepler6_GetBoundingBox (REAL a3, REAL a2, REAL a1, REAL a0, REAL b3, REAL b2, REAL b1, REAL b0, INTEGER x1, INTEGER y1, INTEGER x2, INTEGER y2, INTEGER *x, INTEGER *y, INTEGER *w, INTEGER *h)
{
	REAL t = 0, rt = 0, temp = 0;
	INTEGER x3 = 0, x4 = 0, y3 = 0, y4 = 0;
	if (__ABS(a3) <   1.0000000e-006) {
		if (__ABS(a2) <   1.0000000e-006) {
			*x = Kepler6_Min(x1, x2);
			*w = __ABS(x2 - x1);
		} else {
			t = (  5.0000000e-001 * a1) / a2;
			temp = t * (a2 + t * a3);
			x3 = (int)__ENTIER(a0 + t * (a1 + temp));
			*x = Kepler6_Min(x1, Kepler6_Min(x2, x3 - 1));
			*w = Kepler6_Max(x1, Kepler6_Max(x2, x3 + 1)) - *x;
		}
	} else {
		rt = a2 * a2 - ((REAL)3 * a1) * a3;
		if (rt < (REAL)0) {
			*x = Kepler6_Min(x1, x2);
			*w = __ABS(x2 - x1);
		} else {
			rt = Math_sqrt(rt);
			t = ((-a2 - rt) / (REAL)(REAL)3) / a3;
			if (t > (REAL)0 && t < (REAL)1) {
				temp = t * (a2 + t * a3);
				x3 = (int)__ENTIER(a0 + t * (a1 + temp));
			} else {
				x3 = x1;
			}
			t = ((-a2 + rt) / (REAL)(REAL)3) / a3;
			if (t > (REAL)0 && t < (REAL)1) {
				temp = t * (a2 + t * a3);
				x4 = (int)__ENTIER(a0 + t * (a1 + temp));
			} else {
				x4 = x1;
			}
			*x = Kepler6_Min(x1, Kepler6_Min(x2, Kepler6_Min(x3, x4) - 1));
			*w = Kepler6_Max(x1, Kepler6_Max(x2, Kepler6_Max(x3, x4) + 1)) - *x;
		}
	}
	if (__ABS(b3) <   1.0000000e-006) {
		if (__ABS(b2) <   1.0000000e-006) {
			*y = Kepler6_Min(y1, y2);
			*h = __ABS(y2 - y1);
		} else {
			t = (  5.0000000e-001 * b1) / b2;
			temp = t * (b2 + t * b3);
			y3 = (int)__ENTIER(b0 + t * (b1 + temp));
			*y = Kepler6_Min(y1, Kepler6_Min(y2, x3 - 1));
			*h = Kepler6_Max(y1, Kepler6_Max(y2, y3 + 1)) - *y;
		}
	} else {
		rt = b2 * b2 - ((REAL)3 * b1) * b3;
		if (rt < (REAL)0) {
			*y = Kepler6_Min(y1, y2);
			*h = __ABS(y2 - y1);
		} else {
			rt = Math_sqrt(rt);
			t = ((-b2 - rt) / (REAL)(REAL)3) / b3;
			if (t > (REAL)0 && t < (REAL)1) {
				temp = t * (b2 + t * b3);
				y3 = (int)__ENTIER(b0 + t * (b1 + temp));
			} else {
				y3 = y1;
			}
			t = ((-b2 + rt) / (REAL)(REAL)3) / b3;
			if (t > (REAL)0 && t < (REAL)1) {
				temp = t * (b2 + t * b3);
				y4 = (int)__ENTIER(b0 + t * (b1 + temp));
			} else {
				y4 = y1;
			}
			*y = Kepler6_Min(y1, Kepler6_Min(y2, Kepler6_Min(y3, y4) - 1));
			*h = Kepler6_Max(y1, Kepler6_Max(y2, Kepler6_Max(y3, y4) + 1)) - *y;
		}
	}
	*x -= 2;
	*y -= 2;
	*w += 4;
	*h += 4;
}

static BOOLEAN Kepler6_Intersect (KeplerPorts_Port f, INTEGER x, INTEGER y, INTEGER w, INTEGER h)
{
	INTEGER t = 0;
	x = __KeplerPorts_Port_CX(f, x);
	y = __KeplerPorts_Port_CY(f, y);
	w = __DIV(w, f->scale);
	h = __DIV(h, f->scale);
	t = x + w;
	if (f->X > x) {
		x = f->X;
	}
	if (f->X + f->W < t) {
		w = (f->X + f->W) - x;
	} else {
		w = t - x;
	}
	if (w <= 0) {
		return 0;
	}
	t = y + h;
	if (f->Y > y) {
		y = f->Y;
	}
	if (f->Y + f->H < t) {
		h = (f->Y + f->H) - y;
	} else {
		h = t - y;
	}
	return h > 0;
}

static struct DrawCurve__3 {
	KeplerPorts_Port *f;
	REAL *a3, *a2, *a1, *a0, *b3, *b2, *b1, *b0;
	struct DrawCurve__3 *lnk;
} *DrawCurve__3_s;

static void DrawRec__4 (REAL lo, REAL hi);

static void DrawRec__4 (REAL lo, REAL hi)
{
	INTEGER xlo = 0, xhi = 0, ylo = 0, yhi = 0;
	REAL med = 0;
	xlo = (int)__ENTIER(*DrawCurve__3_s->a0 + lo * (*DrawCurve__3_s->a1 + lo * (*DrawCurve__3_s->a2 + lo * *DrawCurve__3_s->a3)));
	xhi = (int)__ENTIER(*DrawCurve__3_s->a0 + hi * (*DrawCurve__3_s->a1 + hi * (*DrawCurve__3_s->a2 + hi * *DrawCurve__3_s->a3)));
	ylo = (int)__ENTIER(*DrawCurve__3_s->b0 + lo * (*DrawCurve__3_s->b1 + lo * (*DrawCurve__3_s->b2 + lo * *DrawCurve__3_s->b3)));
	yhi = (int)__ENTIER(*DrawCurve__3_s->b0 + hi * (*DrawCurve__3_s->b1 + hi * (*DrawCurve__3_s->b2 + hi * *DrawCurve__3_s->b3)));
	if (__ABS(xhi - xlo) + __ABS(yhi - ylo) <= __ASHL((*DrawCurve__3_s->f)->scale, 1)) {
		__KeplerPorts_Port_DrawLine(*DrawCurve__3_s->f, xlo, ylo, xhi, yhi, 15, 0);
	} else {
		med = (lo + hi) / (REAL)(REAL)2;
		DrawRec__4(lo, med);
		DrawRec__4(med, hi);
	}
}

static void Kepler6_DrawCurve (KeplerPorts_Port f, REAL a3, REAL a2, REAL a1, REAL a0, REAL b3, REAL b2, REAL b1, REAL b0)
{
	struct DrawCurve__3 _s;
	_s.f = &f;
	_s.a3 = &a3;
	_s.a2 = &a2;
	_s.a1 = &a1;
	_s.a0 = &a0;
	_s.b3 = &b3;
	_s.b2 = &b2;
	_s.b1 = &b1;
	_s.b0 = &b0;
	_s.lnk = DrawCurve__3_s;
	DrawCurve__3_s = &_s;
	DrawRec__4((REAL)0, (REAL)1);
	DrawCurve__3_s = _s.lnk;
}

void Kepler6_CRSpline_Draw (Kepler6_CRSpline s, KeplerPorts_Port f)
{
	REAL a3 = 0, a2 = 0, a1 = 0, a0 = 0, b3 = 0, b2 = 0, b1 = 0, b0 = 0;
	INTEGER x = 0, y = 0, w = 0, h = 0, t = 0;
	t = s->p[3]->x - 3 * s->p[2]->x;
	a3 = ((t + 3 * s->p[1]->x) - s->p[0]->x) / (REAL)(REAL)2;
	t = -s->p[3]->x + __ASHL(s->p[2]->x, 2);
	a2 = ((t - 5 * s->p[1]->x) + __ASHL(s->p[0]->x, 1)) / (REAL)(REAL)2;
	a1 = (s->p[2]->x - s->p[0]->x) / (REAL)(REAL)2;
	a0 = s->p[1]->x;
	t = s->p[3]->y - 3 * s->p[2]->y;
	b3 = ((t + 3 * s->p[1]->y) - s->p[0]->y) / (REAL)(REAL)2;
	t = -s->p[3]->y + __ASHL(s->p[2]->y, 2);
	b2 = ((t - 5 * s->p[1]->y) + __ASHL(s->p[0]->y, 1)) / (REAL)(REAL)2;
	b1 = (s->p[2]->y - s->p[0]->y) / (REAL)(REAL)2;
	b0 = s->p[1]->y;
	Kepler6_GetBoundingBox(a3, a2, a1, a0, b3, b2, b1, b0, s->p[1]->x, s->p[1]->y, s->p[2]->x, s->p[2]->y, &x, &y, &w, &h);
	if (__ISP(f, KeplerPorts_BalloonPortDesc, 2)) {
		__KeplerPorts_Port_DrawRect(f, x, y, w, h, 0, 0);
	} else if (Kepler6_Intersect(f, x, y, w, h)) {
		Kepler6_DrawCurve(f, a3, a2, a1, a0, b3, b2, b1, b0);
	}
}

void Kepler6_NewCRSpline (void)
{
	Kepler6_CRSpline s = NIL, s1 = NIL;
	if (KeplerFrames_nofpts >= 4) {
		__NEW(s, Kepler6_CRSplineDesc);
		s->nofpts = 4;
		KeplerFrames_ConsumePoint(&s->p[0]);
		KeplerFrames_ConsumePoint(&s->p[1]);
		KeplerFrames_ConsumePoint(&s->p[2]);
		KeplerFrames_ConsumePoint(&s->p[3]);
		__KeplerGraphs_Graph_Append(KeplerFrames_Focus, (void*)s);
		while (KeplerFrames_nofpts > 0) {
			__NEW(s1, Kepler6_CRSplineDesc);
			s1->nofpts = 4;
			s1->p[0] = s->p[1];
			s1->p[0]->refcnt += 1;
			s1->p[1] = s->p[2];
			s1->p[1]->refcnt += 1;
			s1->p[2] = s->p[3];
			s1->p[2]->refcnt += 1;
			KeplerFrames_ConsumePoint(&s1->p[3]);
			s = s1;
			__KeplerGraphs_Graph_Append(KeplerFrames_Focus, (void*)s);
		}
	}
}

void Kepler6_NewClosedCRSpline (void)
{
	Kepler6_CRSpline s = NIL, s1 = NIL;
	KeplerGraphs_Star point[3];
	INTEGER i = 0;
	if (KeplerFrames_nofpts >= 4) {
		__NEW(s, Kepler6_CRSplineDesc);
		s->nofpts = 4;
		KeplerFrames_ConsumePoint(&s->p[0]);
		point[0] = s->p[0];
		KeplerFrames_ConsumePoint(&s->p[1]);
		point[1] = s->p[1];
		KeplerFrames_ConsumePoint(&s->p[2]);
		point[2] = s->p[2];
		KeplerFrames_ConsumePoint(&s->p[3]);
		__KeplerGraphs_Graph_Append(KeplerFrames_Focus, (void*)s);
		while (KeplerFrames_nofpts > 0) {
			__NEW(s1, Kepler6_CRSplineDesc);
			s1->nofpts = 4;
			s1->p[0] = s->p[1];
			s1->p[0]->refcnt += 1;
			s1->p[1] = s->p[2];
			s1->p[1]->refcnt += 1;
			s1->p[2] = s->p[3];
			s1->p[2]->refcnt += 1;
			KeplerFrames_ConsumePoint(&s1->p[3]);
			s = s1;
			__KeplerGraphs_Graph_Append(KeplerFrames_Focus, (void*)s);
		}
		i = 0;
		while (i <= 2) {
			__NEW(s1, Kepler6_CRSplineDesc);
			s1->nofpts = 4;
			s1->p[0] = s->p[1];
			s1->p[0]->refcnt += 1;
			s1->p[1] = s->p[2];
			s1->p[1]->refcnt += 1;
			s1->p[2] = s->p[3];
			s1->p[2]->refcnt += 1;
			s1->p[3] = point[__X(i, 3)];
			s1->p[3]->refcnt += 1;
			s = s1;
			__KeplerGraphs_Graph_Append(KeplerFrames_Focus, (void*)s);
			i += 1;
		}
	}
}

void Kepler6_Bezier_Draw (Kepler6_Bezier s, KeplerPorts_Port f)
{
	INTEGER a3 = 0, a2 = 0, a1 = 0, a0 = 0, b3 = 0, b2 = 0, b1 = 0, b0 = 0, x = 0, y = 0, w = 0, h = 0, t = 0;
	t = 3 * s->p[3]->x - 5 * s->p[2]->x;
	a3 = (t + 3 * s->p[1]->x) - s->p[0]->x;
	t = -(3 * s->p[3]->x) + 6 * s->p[2]->x;
	a2 = (t - 6 * s->p[1]->x) + 3 * s->p[0]->x;
	a1 = (s->p[1]->x - s->p[0]->x) * 3;
	a0 = s->p[0]->x;
	t = 3 * s->p[3]->y - 5 * s->p[2]->y;
	b3 = (t + 3 * s->p[1]->y) - s->p[0]->y;
	t = -(3 * s->p[3]->y) + 6 * s->p[2]->y;
	b2 = (t - 6 * s->p[1]->y) + 3 * s->p[0]->y;
	b1 = (s->p[1]->y - s->p[0]->y) * 3;
	b0 = s->p[0]->y;
	Kepler6_GetBoundingBox(a3, a2, a1, a0, b3, b2, b1, b0, s->p[0]->x, s->p[0]->y, ((a3 + a2) + a1) + a0, ((b3 + b2) + b1) + b0, &x, &y, &w, &h);
	if (__ISP(f, KeplerPorts_BalloonPortDesc, 2)) {
		__KeplerPorts_Port_DrawRect(f, x, y, w, h, 0, 0);
	} else if (Kepler6_Intersect(f, x, y, w, h)) {
		Kepler6_DrawCurve(f, a3, a2, a1, a0, b3, b2, b1, b0);
	}
}

void Kepler6_NewBezier (void)
{
	Kepler6_Bezier s = NIL, s1 = NIL;
	if (KeplerFrames_nofpts >= 4) {
		__NEW(s, Kepler6_BezierDesc);
		s->nofpts = 4;
		KeplerFrames_ConsumePoint(&s->p[0]);
		KeplerFrames_ConsumePoint(&s->p[1]);
		KeplerFrames_ConsumePoint(&s->p[2]);
		KeplerFrames_ConsumePoint(&s->p[3]);
		__KeplerGraphs_Graph_Append(KeplerFrames_Focus, (void*)s);
		while (KeplerFrames_nofpts > 1) {
			__NEW(s1, Kepler6_BezierDesc);
			s1->nofpts = 4;
			s1->p[0] = s->p[2];
			s1->p[0]->refcnt += 1;
			s1->p[1] = s->p[3];
			s1->p[1]->refcnt += 1;
			KeplerFrames_ConsumePoint(&s1->p[2]);
			KeplerFrames_ConsumePoint(&s1->p[3]);
			s = s1;
			__KeplerGraphs_Graph_Append(KeplerFrames_Focus, (void*)s);
		}
	}
}

void Kepler6_NewClosedBezier (void)
{
	Kepler6_Bezier s = NIL, s1 = NIL, s0 = NIL;
	if (KeplerFrames_nofpts >= 4) {
		__NEW(s, Kepler6_BezierDesc);
		s0 = s;
		s1 = s;
		s->nofpts = 4;
		KeplerFrames_ConsumePoint(&s->p[0]);
		KeplerFrames_ConsumePoint(&s->p[1]);
		KeplerFrames_ConsumePoint(&s->p[2]);
		KeplerFrames_ConsumePoint(&s->p[3]);
		__KeplerGraphs_Graph_Append(KeplerFrames_Focus, (void*)s);
		while (KeplerFrames_nofpts > 1) {
			__NEW(s1, Kepler6_BezierDesc);
			s1->nofpts = 4;
			s1->p[0] = s->p[2];
			s1->p[0]->refcnt += 1;
			s1->p[1] = s->p[3];
			s1->p[1]->refcnt += 1;
			KeplerFrames_ConsumePoint(&s1->p[2]);
			KeplerFrames_ConsumePoint(&s1->p[3]);
			s = s1;
			__KeplerGraphs_Graph_Append(KeplerFrames_Focus, (void*)s);
		}
		__NEW(s, Kepler6_BezierDesc);
		s->nofpts = 4;
		s->p[0] = s1->p[2];
		s->p[0]->refcnt += 1;
		s->p[1] = s1->p[3];
		s->p[1]->refcnt += 1;
		s->p[2] = s0->p[0];
		s->p[2]->refcnt += 1;
		s->p[3] = s0->p[1];
		s->p[3]->refcnt += 1;
		__KeplerGraphs_Graph_Append(KeplerFrames_Focus, (void*)s);
	}
}

__TDESC(Kepler6_CRSplineDesc, 5, 5) = {__TDFLDS("CRSplineDesc", 24), {4, 8, 12, 16, 20, -24}};
__TDESC(Kepler6_BezierDesc, 5, 5) = {__TDFLDS("BezierDesc", 24), {4, 8, 12, 16, 20, -24}};

export void *Kepler6__init(void)
{
	__DEFMOD;
	__IMPORT(Display);
	__IMPORT(KeplerFrames);
	__IMPORT(KeplerGraphs);
	__IMPORT(KeplerPorts);
	__IMPORT(Math);
	__REGMOD(Kepler6, 0);
	__REGCMD("NewBezier", Kepler6_NewBezier);
	__REGCMD("NewCRSpline", Kepler6_NewCRSpline);
	__REGCMD("NewClosedBezier", Kepler6_NewClosedBezier);
	__REGCMD("NewClosedCRSpline", Kepler6_NewClosedCRSpline);
	__INITYP(Kepler6_CRSplineDesc, KeplerGraphs_ConsDesc, 2);
	__INITBP(Kepler6_CRSplineDesc, Kepler6_CRSpline_Draw, 0);
	__INITYP(Kepler6_BezierDesc, KeplerGraphs_ConsDesc, 2);
	__INITBP(Kepler6_BezierDesc, Kepler6_Bezier_Draw, 0);
/* BEGIN */
	__ENDMOD;
}
